/*
 * IBM Technical Support Appliance Version 2 Release 0 Modification 0
 *
 * (C) Copyright IBM Corp. 2017
 */

/*
 * System Files Screen Post-Processor
 */
define(
  
  [ 'jquery' ]
  
  , function( $ ) {
    
    /**
     * @constructor
     * Expected properties:
     *   controllerPath  : {string}        The controller path.
     *   sysFiles        : {Array<string>} The known system files.
     *   sysFileSelector : {HTMLSelect}    The system file HTML Selector.
     *   sysFileTextArea : {HTMLTextArea}  The system file edition HTML Text Area.
     *   waitMessage     : {HTMLDiv}       The HTML Element containing the "Please wait..." message.
     *   loadButton      : {HTMLButton}    The load HTML Button.
     *   saveButton      : {HTMLButton}    The save HTML Button.
     *   undoButton      : {HTMLButton}    The undo HTML Button.
     * @param {Object} properties Property map.
     */
    tsa.SysFilesScreen = function(properties) {
      
      tsa.logger.info('System Files Screen Post-Processor created.');
      
      // Handling the received properties:
      
      if( ( ( typeof properties ) !== 'object' ) || ( properties === null ) ) properties = {};
      
      this.controllerPath  = ( ( typeof properties.controllerPath  ) === 'string' ) ? properties.controllerPath  : '/unknown';
      this.sysFiles        = ( ( typeof properties.sysFiles        ) === 'object' ) ? properties.sysFiles        : []; 
      this.sysFileSelector = ( ( typeof properties.sysFileSelector ) === 'object' ) ? properties.sysFileSelector : null;
      this.sysFileTextArea = ( ( typeof properties.sysFileTextArea ) === 'object' ) ? properties.sysFileTextArea : null;
      this.waitMessage     = ( ( typeof properties.waitMessage     ) === 'object' ) ? properties.waitMessage     : null;
      this.loadButton      = ( ( typeof properties.loadButton      ) === 'object' ) ? properties.loadButton      : null;
      this.saveButton      = ( ( typeof properties.saveButton      ) === 'object' ) ? properties.saveButton      : null;
      this.undoButton      = ( ( typeof properties.undoButton      ) === 'object' ) ? properties.undoButton      : null;
      
      // Populating the system file selector...
      
      if( this.sysFiles instanceof Array ) {
        $( this.sysFiles ).each( $.proxy( function( sysFileIndex, sysFile ) {
          $( sysFileSelector ).append( '<option value="' + sysFile + '">' + sysFile + '</option>' );
        }
        , this ) )
      };
      
      // Ensure the text area is cleared...
      $( this.sysFileTextArea ).val( '' );
      
      // The load button will be initially disabled...
      
      this.loadButton.disabled = true;
      
      // Event handlers...
      
      $( this.sysFileSelector ).on( 'change'       , $.proxy( this.onSysFileSelectorChanged , this ) );
      $( this.sysFileTextArea ).on( 'change'       , $.proxy( this.onSysFileTextAreaChanged , this ) );
      $( this.loadButton      ).on( 'click'        , $.proxy( this.onLoadButtonClicked      , this ) );
      $( this.saveButton      ).on( 'click'        , $.proxy( this.onSaveButtonClicked      , this ) );
      $( this.undoButton      ).on( 'click'        , $.proxy( this.onUndoButtonClicked      , this ) );
      $( window               ).on( 'beforeunload' , $.proxy( this.onBeforeUnload           , this ) );
    };
    
    /*
     * I18N User Messages.
     */
    tsa.SysFilesScreen.I18N_ERROR_OCCURRED      = 'ERROR_OCCURRED';
    tsa.SysFilesScreen.I18N_COMM_FAILED         = 'COMM_FAILED';
    tsa.SysFilesScreen.I18N_LOAD_BEFORE_SAVING  = 'LOAD_BEFORE_SAVING';
    tsa.SysFilesScreen.I18N_CONFIRM_CHG_DISCARD = 'CONFIRM_CHG_DISCARD';
    
    /**
     * The controller path.
     * @type {string}
     */
    tsa.SysFilesScreen.prototype.controllerPath = '';
    
    /**
     * The current file has been modified.
     *  @type {boolean}
     */
    tsa.SysFilesScreen.prototype.sysFileHasBeenModified = false;
    
    /**
     * The known system files.
     * @type {Array<string>}
     */
    tsa.SysFilesScreen.prototype.sysFiles = null;
    
    /** The current loaded file (as an index of this.sysFiles). No file loaded: -1.
     *  @type {number}
     */
    tsa.SysFilesScreen.prototype.sysFile = -1;
    
    /**
     * The system file selector.
     * @type {HTMLSelect}
     */
    tsa.SysFilesScreen.prototype.sysFileSelector = null;
    
    /**
     * The system file text area.
     * @type {HTMLTextArea}
     */
    tsa.SysFilesScreen.prototype.sysFileTextArea = null;
    
    /**
     * The wait message HTML Element (it visually replaces the Text Area
     * while waiting for a response from server).
     * @type {HTMLDiv}
     */
    tsa.SysFilesScreen.prototype.waitMessage = null;
    
    /**
     * The load system file button.
     * @type {HTMLButton}
     */
    tsa.SysFilesScreen.prototype.loadButton = null;
    
    /**
     * The save system file button.
     * @type {HTMLButton}
     */
    tsa.SysFilesScreen.prototype.saveButton = null;
    
    /**
     * The undo (reload system file) button.
     * @type {HTMLButton}
     */
    tsa.SysFilesScreen.prototype.undoButton = null;
    
    /**
     * Event handler: file selector changed.
     * @param {Event} event The event.
     */
    tsa.SysFilesScreen.prototype.onSysFileSelectorChanged = function( event ) {
      
      // Removing the first (informative) option, if it has not been removed before...
      if( this.sysFileSelector.options.length > this.sysFiles.length ) {
        $( this.sysFileSelector.options[0] ).remove();
      }
      
      this.loadButton.disabled = false;
    };
    
    /**
     * Event handler: text area changed.
     * @param {Event} event The event.
     */
    tsa.SysFilesScreen.prototype.onSysFileTextAreaChanged = function( event ) {
      
      this.sysFileHasBeenModified = true;
    };
    
    /**
     * Event handler: load button clicked.
     * @param {Event} event The event.
     */
    tsa.SysFilesScreen.prototype.onLoadButtonClicked = function( event ) {
      
      if( !this.sysFileHasBeenModified || this.confirmChangesDiscardment() ) {
        
        $( this.sysFileTextArea ).val( '' );
        this.sysFile = -1;
        this.sysFileHasBeenModified = false;
        this.disableControls( true );
        
        $.ajax( { url      : this.controllerPath + '/read'
                , data     : { sysFile : $( this.sysFileSelector ).val() }
                , dataType : 'json'
                , type     : 'post'
                , complete : $.proxy( function() {
                               this.disableControls( false );
                             }
                             , this )
                , error    : $.proxy( function() {
                               alert( tsa.SysFilesScreen.I18N_ERROR_OCCURRED + ': ' + tsa.SysFilesScreen.I18N_COMM_FAILED );
                             }
                             , this )
                , success  : $.proxy( function( data ) {
                               if( ( typeof data.text ) === 'string' ) {
                                 $( this.sysFileTextArea ).val( data.text );
                                 this.sysFile = this.sysFileSelector.selectedIndex;
                               }
                               else {
                                 alert( tsa.SysFilesScreen.I18N_ERROR_OCCURRED + ( ( ( typeof data.error ) === 'string' ) ? ': ' + data.error : '.' ) );
                               }
                             }
                             , this )
                } );
      }
    };
    
    /**
     * Event handler: save button clicked.
     * @param {Event} event The event.
     */
    tsa.SysFilesScreen.prototype.onSaveButtonClicked = function( event ) {
      
      if( this.sysFile >= 0 ) {
        this.disableControls( true );
        
        $.ajax( { url      : this.controllerPath + '/write'
                , data     : { sysFile : $( this.sysFileSelector ).val()
                             , text    : $( this.sysFileTextArea ).val() }
                , dataType : 'json'
                , type     : 'post'
                , complete : $.proxy( function() {
                               this.disableControls( false );
                             }
                             , this )
                , error    : $.proxy( function() {
                               alert( tsa.SysFilesScreen.I18N_ERROR_OCCURRED + ': ' + tsa.SysFilesScreen.I18N_COMM_FAILED );
                             }
                             , this )
                , success  : $.proxy( function( data ) {
                               if( ( typeof data.info ) === 'string' ) {
                                 this.sysFileHasBeenModified = false;
                                 alert( data.info );
                               }
                               else {
                                 alert( tsa.SysFilesScreen.I18N_ERROR_OCCURRED + ( ( ( typeof data.error ) === 'string' ) ? ': ' + data.error : '.' ) );
                               }
                             }
                             , this )
                } );
      }
      else {
        alert( tsa.SysFilesScreen.I18N_LOAD_BEFORE_SAVING );
      }
    };
    
    /**
     * Event handler: undo button clicked.
     * @param {Event} event The event.
     */
    tsa.SysFilesScreen.prototype.onUndoButtonClicked = function( event ) {
      
      if( this.sysFile >= 0 ) {
        this.sysFileSelector.selectedIndex = this.sysFile;
        this.onLoadButtonClicked( event );
      }
    };
    
    /**
     * Event handler: user wants to leave the page.
     * @param {Event} event The event.
     */
    tsa.SysFilesScreen.prototype.onBeforeUnload = function( event ) {
      
      if( this.sysFileHasBeenModified ) {
        return tsa.SysFilesScreen.I18N_CONFIRM_CHG_DISCARD;
      }
    };
    
    /**
     * Verify the user agrees to discard the changes.
     * @return {boolean}
     */
    tsa.SysFilesScreen.prototype.confirmChangesDiscardment = function() {
      
      return confirm( tsa.SysFilesScreen.I18N_CONFIRM_CHG_DISCARD );
    };
    
    /**
     * Disable controls (while waiting for an AJAX response from server).
     * @param {boolean} flag Use <code>true</code> to disable or <code>false</code> to enable.
     */
    tsa.SysFilesScreen.prototype.disableControls = function( flag ) {
      
      if( ( typeof flag ) !== 'boolean' ) {
        flag = true;
      }
      
      this.loadButton.disabled = flag;
      
      $( this.waitMessage      ).css( { 'visibility' : ( flag ? 'visible' : 'hidden' ) } );
      $( this.sysFileTextArea  ).css( { 'visibility' : ( flag ? 'hidden' : 'visible' ) } );
      $( this.saveButton       ).css( { 'visibility' : ( flag ? 'hidden' : 'visible' ) } );
      $( this.undoButton       ).css( { 'visibility' : ( flag ? 'hidden' : 'visible' ) } );
    };
    
    
    return tsa.SysFilesScreen;
  }
);