/*
 * IBM Technical Support Appliance Version 2 Release 4 Modification 0
 *
 * (C) Copyright IBM Corp. 2018
 */

//-----------------------------------------------------------------------------
// This command calls CommunicationCommand (if the request is an Ajax request)
// or simply reloads the current document using a traditional HTTP request.
//-----------------------------------------------------------------------------

define(
  [ 'jquery'
  , 'tsa/cmd/AjaxResponse'
  , 'tsa/cmd/CommunicationCommand'
  , 'tsa/TypeVerificator'
  ],
  
  function( $
          , AjaxResponse
          , CommunicationCommand
          , TV )
  {
    //-----------------------------------------------------------
    
    var logger = tsa.getLogger('Command');
    
    //-----------------------------------------------------------
    
    /** @constructor
     *  Parameters:
     *    action              : Action to be performed ("controller/command").
     *    baseURL             : The base URL (action prefix). Optional. Default: Command.URL_CONTROLLER_BASE.
     *    ajax                : AJAX Request required? (optional; default: true).
     *    blockUI             : Block UI during request (optional; default: true).
     *    clearSystemMessages : Clear the system messages when executed (optional; default: true).
     *    printSystemMessages : Print the system messages received from server (optional; default: true).
     *    parameters          : HTTP Request Parameters
     *    cancelable          : The cancel button is shown while the UI is blocked because of an AJAX request (optional; default: true).
     *    eventHandlers       : Map (by event name) containing event handlers to be attached
     *                          to this command at creation time.
     *                          a) executionSucceeded
     *                          b) executionFailed
     *                          c) executionFinished
     *    secureToken         : Generate and set secure token to execution(optional; default true);
     *  
     *  @param {{action              : (undefined|string)
     *          ,baseURL             : (undefined|string)
     *          ,ajax                : (undefined|boolean|string)
     *          ,blockUI             : (undefined|boolean)
     *          ,clearSystemMessages : (undefined|boolean)
     *          ,printSystemMessages : (undefined|boolean)
     *          ,parameters          : (undefined|Object)
     *          ,cancelable          : (undefined|boolean)
     *          ,eventHandlers       : (undefined|Object)
     *          ,secureToken         : (undefined|boolean) }} properties The properties.  */
    
    tsa.cmd.Command = function( properties )
    {
      /** @type {boolean} */ var ajax          = true;
      /** @type {Object}  */ var eventHandlers = null;
      
      if( !TV.typeIsObject( typeof properties ) || ( properties === null ) ) properties = /** @type {?} */ ( {} );
      
      this.executionStarted  = 0;
      this.executionFinished = 0;
      this.parameters        = {};
      this.uploadFileInputs  = [];
      
      this.setBlockUI             ( !TV.typeIsBoolean( typeof properties.blockUI             ) || properties.blockUI             );
      this.setClearSystemMessages ( !TV.typeIsBoolean( typeof properties.clearSystemMessages ) || properties.clearSystemMessages );
      this.setPrintSystemMessages ( !TV.typeIsBoolean( typeof properties.printSystemMessages ) || properties.printSystemMessages );
      this.setCancelable          ( !TV.typeIsBoolean( typeof properties.cancelable          ) || properties.cancelable          );
      this.setSecureToken         (  TV.typeIsUndefined( typeof properties.secureToken       ) ? true : properties.secureToken   );
      
      // HTTP Request Parameters (as a constructor argument, the parameters can contain file inputs)...
      if( TV.typeIsObject( typeof properties.parameters ) && ( properties.parameters !== null ) )
      {
        for (/** @type {string} */ var name in properties.parameters)
        {
          this.setParameter( name, properties.parameters[ name ] );
        }
      }
      
      // Base URL:
      this.setBaseURL( TV.typeIsString( typeof properties.baseURL )
                       ? properties.baseURL
                       : tsa.cmd.Command.URL_CONTROLLER_BASE );
      
      // Is Ajax?
      if      ( TV.typeIsString  ( typeof properties.ajax ) ) ajax = ( properties.ajax === 'true' );
      else if ( TV.typeIsBoolean ( typeof properties.ajax ) ) ajax = properties.ajax;
      
      this.setAjax( ajax );
      
      // Action...
      if ( TV.typeIsString( typeof properties.action ) ) this.setAction( properties.action );
      
      // Event handlers...
      if( TV.typeIsObject( typeof properties.eventHandlers ) && ( properties.eventHandlers !== null ) )
      {
        eventHandlers = properties.eventHandlers;
      }
      
      if( eventHandlers !== null )
      {
        for( /** @type {string} */ var eventName in eventHandlers )
        {
          /** @type {Function} */ var eventHandler = eventHandlers[ eventName ];
          
          if( TV.typeIsFunction( typeof eventHandler ) && ( eventHandler !== null ) )
          {
            $( this ).on( eventName, eventHandler );
          }
        }
      }
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Servlet Context and base URL for the legacy model requests.
     * @type {string}
     */
    tsa.cmd.Command.URL_SERVLET_CONTEXT = '/appliance/';
    
    /**
     * Base URL for the new UI model requests.
     * @type {string}
     */
    tsa.cmd.Command.URL_CONTROLLER_BASE = tsa.cmd.Command.URL_SERVLET_CONTEXT + 'v2/';
    /**
     * String to handle the CSRF secure tokens.
     * @type {string}
     */
    tsa.cmd.Command.CSRFTOKEN = 'CSRFToken';
    /**
     * String to handle the CSRF secure tokens.
     * @type {string}
     */
    tsa.cmd.Command.SECURE_ACTION = 'main/ajaxSecurity'
    /**
     * String to identify the heart beat.
     * @type {string}
     */
    tsa.cmd.Command.SECURE_HEARTBEAT = 'main/ajaxHeartbeat'
    /**
     * String to request the secure token.
     * @type {string}
     */
    tsa.cmd.Command.CSRF_REQUIRED = 'CSRFTokenRequired';
    /**
     * Last execution time: the last time a command finished its execution.
     * @private
     * @type {Date}
     */
    tsa.cmd.Command.lastExecutionTime = null;
    
    //-------------------------------------------------------------------------
    
    /** The time the execution started. 0 means this command has not been executed yet.
     *  @private
     *  @type {number}
     */
    tsa.cmd.Command.prototype.executionStarted = 0;
    
    /** The time the execution finished. 0 means this command has not finished its execution.
     *  @private
     *  @type {number}
     */
    tsa.cmd.Command.prototype.executionFinished = 0;
    
    /** The action (controller/command) to be executed on server.
     *  @private
     *  @type {string}
     */
    tsa.cmd.Command.prototype.action = '';
    
    /** The base URL to be used as the action prefix.
     *  @private
     *  @type {string}
     */
    tsa.cmd.Command.prototype.baseURL = '';
    
    /** Is an ajax request required to complete this command?
     *  @private
     *  @type {boolean} */
    
    tsa.cmd.Command.prototype.ajax = false;
    
    /** Parameters of the command (HTTP parameters in case this command
     *  triggers an HTTP request).
     *  @private
     *  @type {Object} */
    
    tsa.cmd.Command.prototype.parameters = null;
    
    /** HTML File Inputs for file uploading.
     *  @private
     *  @type {Array.<Element>} */
    
    tsa.cmd.Command.prototype.uploadFileInputs = null;
    
    /**
     * Clear the system messages when executed.
     * @private
     * @type {boolean}
     */
    tsa.cmd.Command.prototype.clearSystemMessages = true;
    
    /**
     * Should the UI be blocked during command execution?
     * @private
     * @type {boolean}
     */
    tsa.cmd.Command.prototype.blockUI = true;
    
    /**
     * Generate secure token before executing command.
     * @private
     * @type {boolean}
     */
    tsa.cmd.Command.prototype.secureToken = true;
    
    /**
     * Print the system messages received from server.
     * @private
     * @type {boolean}
     */
    tsa.cmd.Command.prototype.printSystemMessages = true;
    
    /**
     * A cancel button is shown while the UI is blocked because of an AJAX request.
     * @private
     * @type {boolean}
     */
    tsa.cmd.Command.prototype.cancelable = true;
    
    //-------------------------------------------------------------------------
    
    /**
     * @return {Date} The last time a command has been executed, or <code>null</code>
     *                if no command has been executed so far.
     */
    tsa.cmd.Command.getLastExecutionTime = function()
    {
      return tsa.cmd.Command.lastExecutionTime;
    };
    
    //-------------------------------------------------------------------------
    
    /** Clear system messages just before command execution.
     *  @param {boolean} b */
    
    tsa.cmd.Command.prototype.setClearSystemMessages = function( b )
    {
      this.clearSystemMessages = ( TV.typeIsBoolean( typeof b ) && b );
    };
    
    //-------------------------------------------------------------------------
    
    /** Should system messages be cleared just before command execution?
     *  @return {boolean} */
    
    tsa.cmd.Command.prototype.shouldSystemMessagesBeCleared = function()
    {
      return this.clearSystemMessages;
    };
    
    //-------------------------------------------------------------------------
    
    /** Block the UI during command execution.
     *  @param {boolean} b */
    
    tsa.cmd.Command.prototype.setBlockUI = function( b )
    {
      this.blockUI = ( TV.typeIsBoolean( typeof b ) && b );
    };
    
    //-------------------------------------------------------------------------
    
    /** Should the UI be blocked during command execution?
     *  @return {boolean} */
    
    tsa.cmd.Command.prototype.shouldUIBeBlocked = function()
    {
      return this.blockUI;
    };
    //-------------------------------------------------------------------------
    
    /** Enable/Disable secure token for this command execution.
     *  @param {boolean} b */
    tsa.cmd.Command.prototype.setSecureToken = function( b )
    {
      this.secureToken = ( TV.typeIsBoolean( typeof b ) && b );
    };
    /** Set secure token to http parameter.
     *  @param {void}  */
    tsa.cmd.Command.prototype.setCsrfToken = function(token)
    {
      if(token!=null)
      {
        this.setParameter( tsa.cmd.Command.CSRFTOKEN, token );
      }
      else{
        this.setParameter( tsa.cmd.Command.CSRFTOKEN, $("#" + tsa.cmd.Command.CSRFTOKEN).val() );
      }
    };
    //-------------------------------------------------------------------------
    
    /** Verify if secure token is required
     *  @return {boolean} */
    tsa.cmd.Command.prototype.isSecureToken = function()
    {
      return this.secureToken;
    };
    //-------------------------------------------------------------------------
    
    /** Print system messages received from server as part of command execution.
     *  @param {boolean} b */
    
    tsa.cmd.Command.prototype.setPrintSystemMessages = function( b )
    {
      this.printSystemMessages = ( TV.typeIsBoolean( typeof b ) && b );
    };
    
    //-------------------------------------------------------------------------
    
    /** Should system messages be printed as part of command execution?
     *  @return {boolean} */
    
    tsa.cmd.Command.prototype.shouldSystemMessagesBePrinted = function()
    {
      return this.printSystemMessages;
    };
    
    //-------------------------------------------------------------------------
    
    /** Show a cancel button while the UI is blocked because of an AJAX request.
     *  @param {boolean} b */
    
    tsa.cmd.Command.prototype.setCancelable = function( b )
    {
      this.cancelable = ( TV.typeIsBoolean( typeof b ) && b );
    };
    
    //-------------------------------------------------------------------------
    
    /** Is a cancel button shown while the UI is blocked because of an AJAX request?
     *  @return {boolean} */
    
    tsa.cmd.Command.prototype.isCancelable = function()
    {
      return this.cancelable;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * @param {string} baseURL The base URL, to be used as the action prefix.
     */
    tsa.cmd.Command.prototype.setBaseURL = function( baseURL )
    {
      this.baseURL = ( TV.typeIsString( typeof baseURL ) ? baseURL : tsa.cmd.Command.URL_CONTROLLER_BASE );
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * @return {string} The base URL, to be used as the action prefix.
     */
    tsa.cmd.Command.prototype.getBaseURL = function()
    {
      return this.baseURL;
    };
    
    //-------------------------------------------------------------------------
    
    /** @return {number} The execution time (in milliseconds). */
    
    tsa.cmd.Command.prototype.getExecutionTime = function()
    {
      return ( ( this.executionStarted > 0 ) ? ( ( this.executionFinished >= this.executionStarted ) ? this.executionFinished : ( new Date() ).getTime() ) - this.executionStarted : 0 );
    };
    
    //-------------------------------------------------------------------------
    
    /** @param {boolean} b Is an AJAX request required to complete this command? */
    
    tsa.cmd.Command.prototype.setAjax = function( b )
    {
      this.ajax = ( TV.typeIsBoolean( typeof b ) ? b : false );
    };
    
    //-------------------------------------------------------------------------
    
    /** @return {boolean} Is an AJAX request required to complete this command? */
    
    tsa.cmd.Command.prototype.isAjax = function()
    {
      return this.ajax;
    };
    
    //-------------------------------------------------------------------------
    
    /** Set a request parameter.
     *  @param {string} name  Parameter name.
     *  @param {?}      value Parameter value. A <code>null</code> value
     *                        will cause the parameter to be cleared. */
    
    tsa.cmd.Command.prototype.setParameter = function( name, value )
    {
      /** @type {function(?):string}  */ var valueToString  = function( v ) { return ( v + '' ); };
      /** @type {function(?):boolean} */ var valueIsNull    = function( v ) { return ( TV.typeIsUndefined( typeof v ) || ( TV.typeIsObject( typeof v ) && ( v === null ) ) ); };
      /** @type {boolean}             */ var fileInputAdded = false;
      
      if( TV.typeIsString( typeof name ) )
      {
        if( valueIsNull( value ) )
        {
          if( !TV.typeIsUndefined( typeof this.parameters[ name ] ) )
          {
            delete this.parameters[ name ];
          }
        }
        else
        {
          if( TV.typeIsObject( typeof value         ) && ( value                        !== null    ) &&
              TV.typeIsString( typeof value.tagName ) && ( value.tagName .toLowerCase() === 'input' ) &&
              TV.typeIsString( typeof value.type    ) && ( value.type    .toLowerCase() === 'file'  ) )
          {
            // Is the user trying to add a File Input? He should have been used addUploadFileInput;
            // as a courtesy, we will try to add the File Input here...
            
            fileInputAdded = this.addUploadFileInput( name, value );
          }
          
          if( !fileInputAdded )
          {
            value = valueToString( value );
            this.parameters[ name ] = value;
          }
        }
      }
      else
      {
        logger.error( 'Command: Unable to set request parameter (invalid name: [' + ( typeof name ) + ']).' );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /** A copy of the parameters to be sent to the server.
     *  @return {Object.<string,string>} */
    
    tsa.cmd.Command.prototype.getParameters = function()
    {
      /** @type {Object.<string,string>} */ var parameters = {};
      
      for( /** @type {string} */ var name in this.parameters )
      {
        /** @type {?string} */ var value = this.parameters[ name ];
        
        parameters[ name ] = ( TV.typeIsString( typeof value ) ? value : '' + value );
      }
      
      return parameters;
    };
    
    //-------------------------------------------------------------------------
    
    /** @return {Array.<Element>} The HTML File Input Elements for uploading. */
    
    tsa.cmd.Command.prototype.getUploadFileInputs = function()
    {
      return this.uploadFileInputs;
    };
    
    //-------------------------------------------------------------------------
    
    /** Add an HTML File Input Element for uploading.
     *  @param {string}  name  Input name will be overridden.
     *  @param {Element} input Input (use null to remove a previously added input having the given name).
     *  @return {boolean}      Return <code>true</code> if successfully added. */
    
    tsa.cmd.Command.prototype.addUploadFileInput = function( name, input )
    {
      /** @type {boolean} */ var added = false;
      
      if( TV.typeIsString( typeof name ) )
      {
        $( this.uploadFileInputs ).each( $.proxy( function( index, fi )
        {
          /** @type {boolean} */ var duplicate = false;
          
          if( duplicate = ( name === fi.name ) ) this.uploadFileInputs.splice( index, 1 );
          
          return !duplicate;
        }
        , this ) );
        
        if( TV.typeIsObject( typeof input ) && ( input === null ) )
        {
          // Nothing to do, the input has already been deleted.
        }
        else if( TV.typeIsObject( typeof input         ) && ( input                        !== null    ) &&
                 TV.typeIsString( typeof input.tagName ) && ( input.tagName .toLowerCase() === 'input' ) &&
                 TV.typeIsString( typeof input.type    ) && ( input.type    .toLowerCase() === 'file'  ) )
        {
          this.uploadFileInputs.push( input );
          added = true;
        }
        else logger.error( 'Unable to add HTML File Input to Command: input is invalid.' );
      }
      else logger.error( 'Unable to add HTML File Input to Command: invalid name (type: ' + ( typeof name ) + ').' );
      
      return added;
    };
    
    //-------------------------------------------------------------------------
    
    /** Get an HTML File Input Element for uploading.
     *  @param  {string}  name  Input name.
     *  @return {Element}       Input. */
    
    tsa.cmd.Command.prototype.getUploadFileInput = function( name )
    {
      /** @type {Element} */ var result = null;
      
      if( TV.typeIsString( typeof name ) )
      {
        $( this.uploadFileInputs ).each( function( index, fi )
        {
          if( name === fi.name ) result = fi;
          
          return ( result === null );
        } );
      }
      
      return result;
    };
    
    //-------------------------------------------------------------------------
    
    /** @param  {string}   name The name of the parameter.
     *  @return {?string}       The value (or <code>null</code> if the parameter is not found). */
    
    tsa.cmd.Command.prototype.getParameter = function( name )
    {
      /** @type {?string} */ var value = null;
      
      if( TV.typeIsString( typeof name ) && !TV.typeIsUndefined( typeof this.parameters[ name ] ) )
      {
        value = this.parameters[ name ];
      }
      
      return value;
    };
    
    //-------------------------------------------------------------------------
    
    /** Set the action (controller/command ID associated to this request).
     *  @param {string} action The action. */
    
    tsa.cmd.Command.prototype.setAction = function( action )
    {
      if( !TV.typeIsString( typeof action ) ) action = '';
      
      this.action = action;
    };
    
    //-------------------------------------------------------------------------
    
    /** Get the action (controller/command ID associated to this request).
     *  @return {string} The action. */
    
    tsa.cmd.Command.prototype.getAction = function()
    {
      return this.action;
    };
    
    //-------------------------------------------------------------------------
    
    /** @return {boolean} Has this command been executed? */
    
    tsa.cmd.Command.prototype.hasBeenExecuted = function()
    {
      return ( this.executionStarted > 0 );
    };
    
    //-------------------------------------------------------------------------
    
    /** Execute the command. */
    
    tsa.cmd.Command.prototype.execute = function()
    {
      if( !this.hasBeenExecuted() )
      {
        this.executionStarted = ( new Date() ).getTime();
        
        logger.info( 'A command has started its execution (' + this.getAction() + ')...' );
        
        // Storing the last time a command has been executed:
        $( this ).on( 'executionFinished', function( event )
        {
          tsa.cmd.Command.lastExecutionTime = new Date();
        } );
        
        $( this ).trigger( { 'type'    : 'executionStarted'
                           , 'command' : this  } );
        
        // Should the system messages be cleared?
        if( this.shouldSystemMessagesBeCleared() )
        {
          tsa.clearSysMessages();
        }
        
        try
        {
          // We must not do secure token for action  of security and heartbeat
          if( this.secureToken  && this.action!== tsa.cmd.Command.SECURE_ACTION && this.action !== tsa.cmd.Command.SECURE_HEARTBEAT )
          {
            var securityCmd = new tsa.cmd.Command({ secureToken         : false 
                                                  , clearSystemMessages : false });
            securityCmd.generateSecureToken();
            $( securityCmd ).on( 'executionSucceeded', $.proxy(function( event )
            {
              this.setCsrfToken(event.response.csrfToken);
              this.executeCommand();
            },this));
          }
          else
          {
            this.executeCommand();
          }
        }
        catch( e )
        {
          logger.error( e );
        }
      }
      else
      {
        logger.error( 'Command can\'t be executed: a command isn\'t allowed to be executed more than once (' + this.getAction() + ').' );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /** a) For commands requiring AJAX request:      A communication command is built and executed,
     *                                               by using the AJAX request parameters stored into
     *                                               this command.
     *  b) For commands not requiring AJAX requests: HTTP Request, causing a full-page reload.
     *  
     *  Any other type of command that does not require an HTTP Request, or that requires
     *  special handling of the HTTP request, must override this method, and the overrider
     *  method must trigger the "executionSucceeded", "executionFailed" and "executionFinished"
     *  events; the event handlers must be removed subsequently using <code>$(this).off()</code>. */
    
    tsa.cmd.Command.prototype.executeCommand = function()
    {
      if( this.isAjax() )
      {
        /** @type {tsa.cmd.CommunicationCommand} */ var commCmd = null;
        
        /** Processing for both successful and failed AJAX requests, to be
         *  performed just before triggering any event.
         *  @param {tsa.cmd.AjaxResponse} response
         */
        var commonProcessing = $.proxy( function( response )
        {
          // Setting the time in which the execution finished.
          this.executionFinished = ( new Date() ).getTime();
          
          // Print system messages, if any...
          if( this.shouldSystemMessagesBePrinted() )
          {
            /** @type {boolean} */ var sysMsgsAvailable = ( ( response                         !== null ) &&
                                                            ( response.getSysMessages()        !== null ) &&
                                                            ( response.getSysMessages().length >   0    ) );
            if( sysMsgsAvailable )
            {
              tsa.printSysMessages( response.getSysMessages() );
            }
          }
          
          // Update the application attribute map with the one received from the server:
          /** @type {Object} */ var payload = ( ( response !== null ) ? response.payload   : null );
          /** @type {Object} */ var attr    = ( ( payload  !== null ) ? payload.attributes : null );
          
          tsa.updateAttributes( attr );
        }
        , this );
        
        /** @param {tsa.cmd.AjaxResponse} response */
        var commonExecutionFinished = $.proxy( function( response )
        {
          // Trigger the execution finished event (for both successful and failed executions):
          $( this ).trigger( { 'type'     : 'executionFinished'
                             , 'response' : response
                             , 'command'  : this } );
          
          // All events have been triggered; as this command will not be executed
          // anymore, we can safely remove the event handler attachments, just in
          // case they prevent this object to be taken by the garbage collector.
          $( this ).off();
        }
        , this );
        
        /** @param {tsa.cmd.AjaxResponse} response */
        var successCallback = $.proxy( function( response )
        {
          commonProcessing( response );
          logger.info( '...command execution succeeded (' + this.getExecutionTime() + ' ms; ' + this.getAction() + ').' );
          $( this ).trigger( { 'type'     : 'executionSucceeded'
                             , 'response' : response
                             , 'command'  : this } );
          commonExecutionFinished();
        }
        , this );
        
        /** @param {tsa.cmd.AjaxResponse} response */
        var failureCallback = $.proxy( function( response )
        {
          commonProcessing( response );
          logger.warn( '...command execution failed (' + this.getExecutionTime() + ' ms; ' + this.getAction() + ').' );
          $( this ).trigger( { 'type'     : 'executionFailed'
                             , 'response' : response
                             , 'command'  : this } );
          commonExecutionFinished();
        }
        , this );
        
        commCmd = new CommunicationCommand( { url             : this.getBaseURL() + this.getAction()
                                            , cancelable      : this.isCancelable()
                                            , blockUI         : this.shouldUIBeBlocked()
                                            , requestData     : this.getParameters()
                                            , successCallback : successCallback
                                            , failureCallback : failureCallback } );
        
        if( ( this.getUploadFileInputs()        != null ) &&
            ( this.getUploadFileInputs().length >  0    ) )
        {
          commCmd.setUploadFileInputs( this.getUploadFileInputs() );
        }
        
        commCmd.execute();
      }
      else if( this.getAction().length > 0 )
      {
        logger.info( 'The page will be reloaded: new content will be retrieved from server now.' );
        
        /** @type {Object} */  var parameters = this.getParameters();
        /** @type {Element} */ var form       = document.createElement( 'form' );
        
        form.action  = this.getBaseURL() + this.getAction();
        form.enctype = 'multipart/form-data';
        form.method  = 'post';
        
        $( 'body' ).append( form );
        
        for( /** @type {string} */ paramName in parameters )
        {
          /** @type {Element} */ var input      = document.createElement( 'input' );
          /** @type {string}  */ var paramValue = parameters[ paramName ];
          
          $( input ).attr( 'type'  , 'hidden'   );
          $( input ).attr( 'name'  , paramName  );
          $( input ).attr( 'value' , paramValue );
          
          $( form ).append( input );
        }
        
        $.each( this.getUploadFileInputs(), function( inputIndex, input )
        {
          $( form ).append( input );
        } );
        
        if( this.shouldUIBeBlocked() )
        {
          window.setTimeout( function() { form.submit(); }, 100 );
          tsa.blockUI();
        }
        else
        {
          form.submit();
        }
      }
      else
      {
        logger.error( 'Command\'s action not set!' );
      }
    };
    
    //-------------------------------------------------------------------------
    /** 
     * Generate Secure token
     **/
    tsa.cmd.Command.prototype.generateSecureToken = function()
    {
      this.setAction(tsa.cmd.Command.SECURE_ACTION);
      this.setParameter(tsa.cmd.Command.CSRF_REQUIRED, true);
      this.execute();
      
    }
    //-------------------------------------------------------------------------
    return tsa.cmd.Command;
  }
);