/*
 * IBM Technical Support Appliance Version 2 Release 0 Modification 0
 *
 * (C) Copyright IBM Corp. 2017
 */

//-----------------------------------------------------------------------------

define(
  
  [ 'jquery'
  , 'tsa/cmd/AjaxResponse'
  , 'tsa/TypeVerificator' ],
  
  function( $
          , AjaxResponse
          , TV )
  {
    //-------------------------------------------------------------------------
    
    var logger = tsa.getLogger( 'CommunicationCommand' );
    
    //-------------------------------------------------------------------------
    
    /** @constructor
     *  Avoid using this class directly unless necessary.
     *  Use the Command class (wrapper) instead.
     *  
     *  Parameters:
     *    url               The URL to which the request will be sent.
     *    method            HTTP METHOD (GET, POST, PUT, DELETE)
     *    requestData       Request data.
     *    uploadFileInputs  Input element(s) referencing the file(s) to be uploaded (only supported by POST method).
     *    timeout           Milliseconds. 0 means no timeout will be set.
     *    successCallback   Request and communication succeeded / Use: function( {tsa.cmd.AjaxResponse} response ).
     *    failureCallback   Either the request or the communication failed / Use: function( {tsa.cmd.AjaxResponse} response ).
     *    commonCallback    Called no matter if the execution succeeded or failed.
     *    blockUI           Block UI during request. Default: true.
     *    cancelable        The UI blockage will show a cancel button, that cancels this command (response from server will be ignored). Default: true.
     *   
     *  @param {{url              : (string|undefined),
     *           method           : (string|undefined),
     *           requestData      : (Object|undefined),
     *           uploadFileInputs : (Array.<Element>|undefined),
     *           timeout          : (number|undefined),
     *           successCallback  : (Function|undefined),
     *           failureCallback  : (Function|undefined),
     *           commonCallback   : (Function|undefined),
     *           blockUI          : (boolean|undefined),
     *           cancelable       : (boolean|undefined)}} parameters */
    
    tsa.cmd.CommunicationCommand = function( parameters )
    {
      if( !TV.typeIsObject( typeof parameters ) || ( parameters === null ) ) parameters = /** @type {?} */ ( {} );
      
      if( TV.typeIsString  ( typeof parameters.url              ) ) this.url              = parameters.url;
      if( TV.typeIsString  ( typeof parameters.method           ) ) this.method           = parameters.method;
      if( TV.typeIsObject  ( typeof parameters.requestData      ) ) this.requestData      = parameters.requestData;
      if( TV.typeIsObject  ( typeof parameters.uploadFileInputs ) ) this.uploadFileInputs = parameters.uploadFileInputs;
      if( TV.typeIsNumber  ( typeof parameters.timeout          ) ) this.timeout          = parameters.timeout;
      if( TV.typeIsFunction( typeof parameters.successCallback  ) ) this.successCallback  = parameters.successCallback;
      if( TV.typeIsFunction( typeof parameters.failureCallback  ) ) this.failureCallback  = parameters.failureCallback;
      if( TV.typeIsFunction( typeof parameters.commonCallback   ) ) this.commonCallback   = parameters.commonCallback;
      if( TV.typeIsBoolean ( typeof parameters.blockUI          ) ) this.blockUI          = parameters.blockUI;
      if( TV.typeIsBoolean ( typeof parameters.cancelable       ) ) this.cancelable       = parameters.cancelable;
      
      // Using default values if necessary...
      if( this.method.length === 0 ) this.method = 'POST';
      
      // Initialize file input values...
      this.fileInputValues = [];
    };
    
    //-----------------------------------------------------------------------------
    
    // Public properties (to be set at construction time):
    
    /** @type {string}          */ tsa.cmd.CommunicationCommand.prototype.url              = '';
    /** @type {string}          */ tsa.cmd.CommunicationCommand.prototype.method           = 'POST'; // HTTP METHOD (GET, POST, PUT, DELETE)
    /** @type {Object}          */ tsa.cmd.CommunicationCommand.prototype.requestData      = null;   // Request data.
    /** @type {Array.<Element>} */ tsa.cmd.CommunicationCommand.prototype.uploadFileInputs = null;   // Input element(s) referencing the file(s) to be uploaded (only supported by POST method).
    /** @type {number}          */ tsa.cmd.CommunicationCommand.prototype.timeout          = 0;      // Milliseconds. 0 means no timeout will be set.
    /** @type {Function}        */ tsa.cmd.CommunicationCommand.prototype.successCallback  = null;   // Request and communication succeeded / Use: function( {tsa.cmd.AjaxResponse} response ).
    /** @type {Function}        */ tsa.cmd.CommunicationCommand.prototype.failureCallback  = null;   // Either the request or the communication failed / Use: function( {tsa.cmd.AjaxResponse} response ).
    /** @type {Function}        */ tsa.cmd.CommunicationCommand.prototype.commonCallback   = null;   // Callback executed after success/failure callback, no matter what / Use: function().
    /** @type {boolean}         */ tsa.cmd.CommunicationCommand.prototype.blockUI          = true;   // Block UI while waiting for response. Default: true.
    /** @type {boolean}         */ tsa.cmd.CommunicationCommand.prototype.cancelable       = true;   // If UI blocked, a Cancel button is displayed. Default: true.
    
    // Private properties:
    
    /** @type {Date}           ** @private */ tsa.cmd.CommunicationCommand.prototype.sendingTime       = null;
    /** @type {Date}           ** @private */ tsa.cmd.CommunicationCommand.prototype.receptionTime     = null;
    /** @type {string}         ** @private */ tsa.cmd.CommunicationCommand.prototype.status            = 'NEW'; // NEW, INVALID, WAITING, SUCCEED, COMM_FAILED, RQST_FAILED, CANCELED
    /** @type {Object}         ** @private */ tsa.cmd.CommunicationCommand.prototype.responseData      = null;
    /** @type {string}         ** @private */ tsa.cmd.CommunicationCommand.prototype.commErrorTechInfo = '';    // Technical information about a communication error (status: COMM_FAILED).
    /** @type {string}         ** @private */ tsa.cmd.CommunicationCommand.prototype.contentType       = 'application/x-www-form-urlencoded';
    /** @type {Array.<string>} ** @private */ tsa.cmd.CommunicationCommand.prototype.fileInputValues   = null;  // Original File Input Element values.
    
    //-----------------------------------------------------------------------
    
    /** @param {string}          s */ tsa.cmd.CommunicationCommand.prototype.setUrl              = function( s ) { this.url              = s; };
    /** @param {string}          s */ tsa.cmd.CommunicationCommand.prototype.setMethod           = function( s ) { this.method           = s; };
    /** @param {Object}          o */ tsa.cmd.CommunicationCommand.prototype.setRequestData      = function( o ) { this.requestData      = o; };
    /** @param {Array.<Element>} a */ tsa.cmd.CommunicationCommand.prototype.setUploadFileInputs = function( a ) { this.uploadFileInputs = a; };
    /** @param {number}          n */ tsa.cmd.CommunicationCommand.prototype.setTimeout          = function( n ) { this.timeout          = n; };
    /** @param {Function}        f */ tsa.cmd.CommunicationCommand.prototype.setSuccessCallback  = function( f ) { this.successCallback  = f; };
    /** @param {Function}        f */ tsa.cmd.CommunicationCommand.prototype.setFailureCallback  = function( f ) { this.failureCallback  = f; };
    /** @param {Function}        f */ tsa.cmd.CommunicationCommand.prototype.setCommonCallback   = function( f ) { this.commonCallback   = f; };
    /** @param {boolean}         b */ tsa.cmd.CommunicationCommand.prototype.setBlockUI          = function( b ) { this.blockUI          = b; };
    /** @param {boolean}         b */ tsa.cmd.CommunicationCommand.prototype.setCancelable       = function( b ) { this.cancelable       = b; };
    
    //-----------------------------------------------------------------------
    
    /** @type {boolean} Is a message dialog being shown? */
    
    tsa.cmd.CommunicationCommand.dialogMsgIsBeingShown = false;
    
    //-----------------------------------------------------------------------
    
    /** Execute the communication command. */
    
    tsa.cmd.CommunicationCommand.prototype.execute = function()
    {
      /** @type {boolean} */ var success      = false;
      /** @type {boolean} */ var canceled     = false;
      /** @type {string}  */ var warningMsg   = 'Communication Command: ';
      /** @type {Object}  */ var ajaxSettings = null;
      
      /** @this {tsa.cmd.CommunicationCommand} */ var communicationSucceededHandler;
      /** @this {tsa.cmd.CommunicationCommand} */ var communicationFailedHandler;
      /** @this {tsa.cmd.CommunicationCommand} */ var communicationCompletedHandler;
      /** @this {tsa.cmd.CommunicationCommand} */ var cancelHandler;
      
      // The command can only be executed once...
      
      success = ( this.status === 'NEW' );
      
      if( !success ) warningMsg += 'This has already been executed.';
      
      // Adding command to history...
      
      if( success ) {
        tsa.cmd.CommunicationCommand.addCommandToHistory( this );
      }
      
      // Validating and sanitizing command...
      
      if( success )
      {
        success = ( ( typeof this.url ) === 'string' ) && ( this.url.length > 0 );
        
        if( !success ) warningMsg += 'URL not set.';
        
        if( success )
        {
          success = ( ( typeof this.method ) === 'string' );
          
          if( success )
          {
            this.method = this.method.toUpperCase();
            success = ( ( this.method === 'GET' ) || ( this.method === 'POST'   ) ||
                        ( this.method === 'PUT' ) || ( this.method === 'DELETE' ) );
          }
          
          if( !success ) warningMsg += 'Invalid HTTP method (' + this.method + ').';
        }
        
        if( success )
        {
          success = ( ( typeof this.requestData ) === 'object' );
          
          if( !success ) warningMsg += 'Request data is not valid (' + ( typeof this.requestData ) + ').';
        }
        
        // Verifying file upload input elements...
        if( success ) {
          success = ( ( typeof this.uploadFileInputs ) === 'object' ) &&
                    ( ( this.uploadFileInputs === null ) || Array.isArray( this.uploadFileInputs ) );
          
          if( success && Array.isArray( this.uploadFileInputs ) ) {
            if( this.uploadFileInputs.length > 0 ) {
              $( this.uploadFileInputs ).each( function( index, element )
              {
                success = success && ( ( typeof element ) === 'object' ) &&
                                     ( element !== null ) &&
                                     ( ( typeof element.tagName ) === 'string'  ) &&
                                     ( element.tagName.toLowerCase() === 'input' ) &&
                                     ( element.getAttribute( 'type' ).toLowerCase() === 'file' );
              } );
            }
            else {
              this.uploadFileInputs = null;
            }
          }
          
          if( !success ) warningMsg += 'Invalid HTML File Input Element Array. Please provide an array containing valid input elements (type must be \'file\').';
          
          // Old browsers do not support XMLHttpRequest Level 2's FormData, required
          // by this command to upload files to the server. Also, the browser has to
          // support the 'files' property for File Input Elements...
          
          if( success && Array.isArray( this.uploadFileInputs ) ) {
            success = ( ( typeof FormData                       ) !== 'undefined' ) &&
                      ( ( typeof this.uploadFileInputs[0].files ) !== 'undefined' );
            
            if( !success ) warningMsg += 'Your browser does not support the file uploading method required by this application '
                                      +  '(XMLHttpRequest Level 2 / FormData). Please use a more modern browser.';
          }
        }
        
        if( success )
        {
          // Discarding invalid combinations...
          
          if( this.uploadFileInputs !== null )
          {
            if( this.method !== 'POST' )
            {
              success = false;
              warningMsg += 'Only the HTTP POST method supports file uploading. You tried to use: ' + this.method + '.';
            }
            else
            {
              // File uploading requires content type to be multipart/form-data...
              this.contentType = 'multipart/form-data';
            }
          }
        }
        
        if( !success ) this.status = 'INVALID';
      }
      
      // Building handlers...
      
      if( success )
      {
        communicationSucceededHandler = $.proxy( function( response )
        {
          if( ( this.status === 'WAITING' ) && !canceled )
          {
            /** @type {string} */ var errPrefix = 'The response received from server could not be converted to a AjaxResponse instance';
            
            if( ( typeof response ) !== 'object' )
            {
              logger.error( errPrefix + ' (unexpected type: ' + ( typeof response ) + '; ' + response + ').' );
              response = null;
            }
            else if( response === null ) logger.error( errPrefix + ' (null).' );
            
            // Converting from a plain object to an AjaxResponse instance...
            response = tsa.cmd.AjaxResponse.buildFromJSON( response );
            
            tsa.cmd.CommunicationCommand.defaultSuccessHandler( this, response );
            
            if( response !== null )
            {
              if( response.success && ( response.payload === null ) ) response.payload = {};
              
              // Set the successCallback and failureCallback members of this
              // command, at construction time.
              
              try
              {
                if( response.success ) { if( this.successCallback ) this.successCallback( response ); }
                else                   { if( this.failureCallback ) this.failureCallback( response ); }
              }
              catch( error )
              {
                logger.error( 'An unhandled error has been catched while handling a Communication Command response (using: ' + ( response.success ? 'success' : 'failure' ) + 'Callback).' );
                logger.error( error );
              }
              
              try
              {
                if( this.commonCallback ) this.commonCallback( response );
              }
              catch( error )
              {
                logger.error( 'An unhandled error has been catched while handling a Communication Command response (using: commonCallback).' );
                logger.error( error );
              }
            }
            else
            {
              /** @type {string} */ var failureMsg = 'Invalid response received from server';
              
              logger.error( failureMsg );
              
              if( this.failureCallback )
              {
                /** @type {tsa.cmd.AjaxResponse} */ var response = null;
                
                response = new AjaxResponse( { success     : false
                                             , payload     : {}
                                             , sysMessages : [ new tsa.cmd.SystemMessage( { level : 'ERROR', message : failureMsg /* TODO i18n */ } ) ]
                                             , exception   : null } );
                
                this.failureCallback( response );
              }
            }
          }
        }
        , this );
        
        communicationFailedHandler = $.proxy( function( error )
        {
          if( ( this.status === 'WAITING' ) && !canceled )
          {
            /** @type {string} */ var commErrorTechInfo = '';
            
            if( TV.typeIsObject( typeof error         ) && ( error !== null ) &&
                TV.typeIsString( typeof error.message ) ) {
              commErrorTechInfo = error.message;
            }
            
            tsa.cmd.CommunicationCommand.defaultErrorHandler( this, commErrorTechInfo );
            
            if( this.failureCallback )
            {
              /** @type {tsa.cmd.AjaxResponse} */ var response = null;
              
              response = new AjaxResponse( { success     : false
                                           , payload     : {}
                                           , sysMessages : [ new tsa.cmd.SystemMessage( { level : 'ERROR', message : 'Communication error' /* TODO i18n */ } ) ]
                                           , exception   : null } );
              
              this.failureCallback( response );
            }
          }
        }
        , this );
        
        communicationCompletedHandler = $.proxy( function( jqXHR, textStatus )
        {
          if( this.blockUI )
          {
            // If the communication was canceled, the UI has been already unlocked...
            if( !canceled )
            {
              tsa.unblockUI();
            }
          }
        }
        , this );
        
        cancelHandler = $.proxy( function()
        {
          logger.warn( 'Execution canceled: the response from server will be ignored.' );
          canceled = true;
        }
        , this );
      }
      
      // Sending request to servlet...
      
      if( success )
      {
        ajaxSettings = { 'dataType' : 'json' /* response */
                       , 'cache'    : false
                       , 'error'    : communicationFailedHandler
                       , 'success'  : communicationSucceededHandler
                       , 'complete' : communicationCompletedHandler
                       , 'type'     : this.method
                       , 'url'      : this.url };
        
        if( ( ( typeof this.timeout ) === 'number' ) && !isNaN( this.timeout ) && ( this.timeout > 0 ) ) {
          ajaxSettings[ 'timeout' ] = this.timeout;
        }
        
        this.sendingTime = new Date();
        this.status      = 'WAITING';
        
        if( this.requestData === null ) this.requestData = {};
        
        if( this.method === 'POST' )
        {
          switch( this.contentType )
          {
            case 'multipart/form-data':
              ajaxSettings[ 'contentType' ] = false;
              ajaxSettings[ 'processData' ] = false;
              ajaxSettings[ 'data'        ] = /** @type {?} */ ( new FormData() );
              
              for( /** @type {string} */ var parameterName in this.requestData )
              {
                ajaxSettings[ 'data' ].append( parameterName, this.requestData[ parameterName ] );
              }
              
              if( this.uploadFileInputs !== null )
              {
                $( this.uploadFileInputs ).each( $.proxy( function( uploadFileInputIndex, uploadFileInput )
                {
                  /** @type {number} */ var fileCounter = 0;
                  
                  if( ( ( typeof uploadFileInput.name ) !== 'string' ) || ( uploadFileInput.name.length === 0 ) )
                  {
                    uploadFileInput.name = 'file' + uploadFileInputIndex;
                  }
                  
                  $( uploadFileInput.files ).each( $.proxy( function( fileIndex, file )
                  {
                    /** @type {string} */ var parameterName = uploadFileInput.name + '@' + fileCounter;
                    /** @type {string} */ var fileName      = file.name;
                    
                    ajaxSettings[ 'data' ].append( parameterName, file, fileName );
                    
                    this.fileInputValues.push( '(' + parameterName + ') ' + fileName );
                    
                    fileCounter++;
                  }
                  , this ) );
                }
                , this ) );
              }
              break;
            
            case 'application/json':
              ajaxSettings[ 'contentType' ] = this.contentType;
              ajaxSettings[ 'data'        ] = ( TV.typeIsUndefined( typeof JSON ) ? this.requestData : JSON.stringify( this.requestData ) );
              break;
            
            default:
              ajaxSettings[ 'contentType' ] = this.contentType;
              ajaxSettings[ 'data'        ] = this.requestData;
              break;
          }
        }
        else
        {
          ajaxSettings[ 'data' ] = tsa.cmd.CommunicationCommand.createHttpQuery( this.requestData );
        }
        
        // Block UI...
        
        if( this.blockUI )
        {
          tsa.blockUI( this.cancelable ? cancelHandler : null );
        }
        
        // Sending ajax request...
        
        $.ajax( ajaxSettings );
      }
      
      if( !success )
      {
        logger.warn( warningMsg );
      }
    };
    
    /** Cancel command execution (response will be discarded, and no handlers will be called). */
    
    tsa.cmd.CommunicationCommand.prototype.cancel = function()
    {
      this.status = 'CANCELED';
    };
    
    //-------------------------------------------------------------------------
    
    /** @type {number}                                   */ tsa.cmd.CommunicationCommand.HISTORY_SIZE = 24;
    /** @type {Array.<tsa.cmd.CommunicationCommand>} */ tsa.cmd.CommunicationCommand.history      = [];
    /** @type {string}                                   */ tsa.cmd.CommunicationCommand.defaultURL   = '';
    
    /** Add command to history.
     *  @param {tsa.cmd.CommunicationCommand} commCmd The command. */
    
    tsa.cmd.CommunicationCommand.addCommandToHistory = function( commCmd )
    {
      if( ( ( typeof commCmd ) === 'object' ) && ( commCmd !== null ) )
      {
        if( tsa.cmd.CommunicationCommand.history.length >= tsa.cmd.CommunicationCommand.HISTORY_SIZE )
        {
          tsa.cmd.CommunicationCommand.history.shift();
        }
        
        tsa.cmd.CommunicationCommand.history.push( commCmd );
      }
    };
    
    /** Creates an HTTP parameter query, using the parameter names and values
     *  provided as arguments.
     *  @param {Object.<string,string>} parameters The parameters.
     *  @return                                    The HTTP Query String.
     */
    
    tsa.cmd.CommunicationCommand.createHttpQuery = function( parameters )
    {
      /** @type {string}         */ var result         = '';
      /** @type {Array.<string>} */ var nameValuePairs = [];
      
      if( TV.typeIsObject( typeof parameters  ) && ( parameters !== null ) )
      {
          for( /** @type {string} */ var parameterName in parameters )
          {
            if( ( parameterName.length > 0 ) && !TV.typeIsUndefined( typeof parameters[ parameterName ] ) )
            {
              nameValuePairs.push( encodeURIComponent( parameterName )             + "="
                                 + encodeURIComponent( parameters[ parameterName ] + '' ) );
            }
          }
          
          if( nameValuePairs.length > 0 ) result = nameValuePairs.join( '&' );
      }
      
      return result;
    };
    
    /** Default success handler (to be executed before custom success callback, if provided).
     *  @param {tsa.cmd.CommunicationCommand} commCmd  The Communication Command object invoking the function
     *  @param {tsa.cmd.AjaxResponse}         response Response (from server).
     */
    
    tsa.cmd.CommunicationCommand.defaultSuccessHandler = function( commCmd, response )
    {
      if( commCmd.status === 'WAITING' )
      {
        if( !TV.typeIsObject( typeof response ) ||
            !TV.isInstanceOf( response, AjaxResponse ) )
        {
          response = null;
        }
        
        commCmd.receptionTime = new Date();
        commCmd.responseData  = response;
        
        // Even when the communication operation succeeded, the request processing may have failed...
        if( ( response === null ) || !response.getSuccess() )
        {
          commCmd.status = 'RQST_FAILED';
        }
        else
        {
          commCmd.status = 'SUCCEED';
        }
        
        logger.info( 'Communication succeeded. Status: ' + commCmd.status + '.' );
      }
    };
    
    /** Default error handler (to be executed before custom error callback, if provided).
     *  @param {tsa.cmd.CommunicationCommand} commCmd           The Communication Command object invoking the function
     *  @param {string=}                      commErrorTechInfo Optional. Technical information (if any) about the communication error.
     */
    
    tsa.cmd.CommunicationCommand.defaultErrorHandler = function( commCmd, commErrorTechInfo )
    {
      if( commCmd.status === 'WAITING' )
      {
        /** @type {string} */ var errorMsg = 'Communication error.';
        
        if( ( typeof commErrorTechInfo ) !== 'string' ) commErrorTechInfo = '';
        
        commCmd.receptionTime     = new Date();
        commCmd.status            = 'COMM_FAILED';
        commCmd.responseData      = null;
        commCmd.commErrorTechInfo = commErrorTechInfo;
        
        logger.warn( 'Communication failed: ' + commCmd.status + ( ( commErrorTechInfo.length > 0 ) ? ' (' + commErrorTechInfo + ')' : '' ) + '.' );
      }
    };
    
    /** Set the default URL.
     *  @param {string} url The default URL. */
    
    tsa.cmd.CommunicationCommand.setDefaultURL = function( url )
    {
      tsa.cmd.CommunicationCommand.defaultURL = ( TV.typeIsString( typeof url ) ? url : '' );
    };
    
    /** Get the default URL.
     *  @return {string} The default URL. */
    
    tsa.cmd.CommunicationCommand.getDefaultURL = function()
    {
      return tsa.cmd.CommunicationCommand.defaultURL;
    };
    
    /** Test communication.
     *  To execute this test by using the console, use the following statement
     *  (change test number depending on the test you want to run):
     *  
     *  CommunicationCommand.test( 1 );
     *  
     *  @param {number=} testNumber Optional. Default: 0.
     *                              0: Server will process a request, reporting it has been processed flawlessly.
     *                              1: Server will process a request, reporting an error was found during such processing.
     *                              2: Server will generate and report an innocuous exception.
     */
    
    tsa.cmd.CommunicationCommand.test = function( testNumber )
    {
      
      if( ( ( typeof testNumber ) !== 'number' ) || isNaN( testNumber ) || ( testNumber < 0 ) || ( testNumber > 2 ) )
      {
        testNumber = 0;
      }
      
      var commCmd = new tsa.cmd.CommunicationCommand(
      {
        method          : 'GET',
        requestData     : { 'action' : 'test', 'testNumber' : testNumber },
        successCallback : function( response ) { logger.info( '...Communication Command Test finished. Response processed by the success-callback.' ); logger.info( response ); },
        failureCallback : function( response ) { logger.info( '...Communication Command Test finished. Response processed by the failure-callback.' ); logger.info( response ); }
      } );
      
      logger.info( "Communication Command Test is about to be executed..." );
      commCmd.execute();
    };
    
    return tsa.cmd.CommunicationCommand;
  }
);