/*
 * IBM Technical Support Appliance Version 2 Release 4 Modification 0
 *
 * (C) Copyright IBM Corp. 2018
 */

/*
 * Entry point for HTML post-processing using modular JavaScript.
 * Tasks performed:
 * a) Set-up the console logger.
 * b) Print third-party libraries versions.
 * c) Execute any required HTML post-processing scripts.
 */
(function() {
  
  //***************************************************************************
  // TSA Module Folders:
  //***************************************************************************
  
  tsa.cmd     = { handlers : {} }; // Commands and Handlers.
  tsa.widgets = {};                // UI Widgets.
  tsa.data    = {};                // Static, general purpose data.
  
  //***************************************************************************
  // TSA Global constants & variables:
  //***************************************************************************
  
  /** @const ** @type {string} */ tsa.NBSP = String.fromCharCode( 0x00A0 );
  
  /** @type {number}                        */ tsa.uiBlockCount          = 0;
  /** @type {number}                        */ tsa.uiBlockDelay          = 2000; /* milliseconds */
  /** @type {number}                        */ tsa.uiBlockTimeout        = 0;
  /** @type {?number}                       */ tsa.autoRefreshIntervalId = null;
  /** @type {boolean}                       */ tsa.changesCanBeLostFlag  = false;
  /** @type {Array.<tsa.cmd.SystemMessage>} */ tsa.systemMessages        = [];
  
  //***************************************************************************
  /** Basic date/time formatter.
   *  @param {Date}     date        The date object.
   *  @param {boolean}  dateFlag    Format date.
   *  @param {boolean}  timeFlag    Format time.
   *  @param {boolean=} timeMsFlag  Include milliseconds (optional; default: false).
   *                                Ignored if <code>timeFlag</code> is not set.
   *  @return {string}
   */
  tsa.formatDate = function(date, dateFlag, timeFlag, timeMsFlag) {
    
    /** @type {string} */ var result = '';
    
    if (date instanceof Date) {
      
      if (dateFlag) {
        
        result +=     date.getFullYear() + '-'
               +  ( ( date.getMonth() < 9 ) ? '0' : '' )
               +    ( date.getMonth() + 1 ) + '-'
               +  ( ( date.getDate() < 10 ) ? '0' : '' )
               +      date.getDate()
               + ( timeFlag ? ' ' : '' );
      }
      
      if (timeFlag) {
        
        result += ( ( date.getHours() < 10 ) ? '0' : '' )
               +      date.getHours() + ':'
               +  ( ( date.getMinutes() < 10 ) ? '0' : '' )
               +      date.getMinutes() + ':'
               +  ( ( date.getSeconds() < 10 ) ? '0' : '' )
               +      date.getSeconds();
        
        if (timeMsFlag) {
          
          result += '.' + ( ( date.getMilliseconds() < 100 ) ? '0' : '' )
                 +        ( ( date.getMilliseconds() <  10 ) ? '0' : '' )
                 +            date.getMilliseconds();
          
        }
      }
    }
    
    return result;
  };
  
  //***************************************************************************
  /** Safe log function (to console).
   *  @private
   *  @param {string}          level          Level: 'info', 'warn' or 'error'.
   *  @param {string}          moduleId       Module ID.
   *  @param {(string|Object)} errorOrMessage Error or message to be logged.
   */
  tsa._log = function( level, moduleId, errorOrMessage ) {
    
    /** @type {Date}   */ var now            = new Date();
    /** @type {string} */ var formattedDate  = tsa.formatDate( now, false, true, true );
    /** @type {string} */ var msgPrefix      = '';
    /** @type {string} */ var levelForPrefix = '';
    /** @type {string} */ var msg            = '';
    
    if ((typeof level) !== 'string') {
      
      level = 'info';
    }
    
    if ((typeof moduleId) !== 'string') {
      moduleId = '';
    }
    
    if (((typeof errorOrMessage) !== 'string') &&
        ((typeof errorOrMessage) !== 'object')) {
      
      errorOrMessage = 'Message unavailable (type: ' + (typeof errorOrMessage) + ')';
    }
    
    if ( errorOrMessage === null ) {
      
      errorOrMessage = 'Message unavailable (null)';
    }
    
    if ((level !== 'info' ) &&
        (level !== 'warn' ) &&
        (level !== 'error')) {
    
      level = 'info';
    }
    
    if (( typeof errorOrMessage) === 'string') {
    
      msg = /** @type {string} */ ( errorOrMessage );
    }
    else {
    
           if ((typeof errorOrMessage.name       ) === 'string'  ) msg += errorOrMessage.name + ': ';
           if ((typeof errorOrMessage.message    ) === 'string'  ) msg += errorOrMessage.message;
      else if ((typeof errorOrMessage.description) === 'string'  ) msg += errorOrMessage.description;
      else if ((typeof errorOrMessage.toString   ) === 'function') msg += errorOrMessage.toString();
      else                                                         msg += errorOrMessage + '';
      
      if ((typeof errorOrMessage.stack ) !== 'undefined') msg += '\n' + errorOrMessage.stack;
    }
    
    levelForPrefix = level.toUpperCase();
    
    while (levelForPrefix.length < 5) levelForPrefix += ' ';
    
    msgPrefix = '[TSA][' + formattedDate + '][' + levelForPrefix + '][' + moduleId + '] ';
    
    try {
      
      if (window.console) {
        
        if      ( window.console[ level ] ) window.console[ level ]( msgPrefix + msg );
        else if ( window.console[ 'log' ] ) window.console[ 'log' ]( msgPrefix + msg );
      }
    }
    catch( error ) { /* Nothing to do. */ }
  };
  
  //***************************************************************************
  //* Build a logger for a module/class.
  //***************************************************************************
  
  tsa.getLogger = function(moduleName) {
    return { info  : function(errorOrMsg) { tsa._log('info'  , moduleName , errorOrMsg); }
           , warn  : function(errorOrMsg) { tsa._log('warn'  , moduleName , errorOrMsg); }
           , error : function(errorOrMsg) { tsa._log('error' , moduleName , errorOrMsg); } };
  };
  
  //***************************************************************************
  //* Base logger.
  //***************************************************************************
  
  tsa.logger = tsa.getLogger('TSA');
  
  //***************************************************************************
  //* First thing we will log: the jQuery versions.
  //***************************************************************************
  
  tsa.logger.info('jQuery versions: { IBM Template: "' + jQuery.fn.jquery + '", TSA Modules: "' + tsa.jQuery.fn.jquery + '" }.' );
  
  //***************************************************************************
  //* Handling module loader errors:
  //***************************************************************************
  
  require.on( 'error', function( error )
  {
    /** @type {string} */ var module = '?';
    
    if( ( ( typeof error.info        ) === 'object' ) && ( error.info        !== null ) &&
        ( ( typeof error.info.length ) === 'number' ) && ( error.info.length >   0    ) &&
        ( ( typeof error.info[0]     ) === 'string' ) )
    {
      module = error.info[0];
    }
    
    tsa.logger.error( 'The requested module doesn\'t exist: "' + module +'".' );
  } );
  
  //***************************************************************************
  /**
   * Call this method if at least a change (in a form, for example) has been made.
   * Actions will be taken to prevent the user from accidentally discard the changes.
   * @public
   * @param {boolean=} flag Optional. Default: true. Use <code>false</code> to reset the
   *                        change loss prevention mechanism (until the flag is set again).
   */
  tsa.changesCanBeLost = function( flag )
  {
    /** @type {boolean} */ var oldFlag = tsa.changesCanBeLostFlag;
    
    tsa.changesCanBeLostFlag = !( flag === false );
    
    if( oldFlag !== tsa.changesCanBeLostFlag )
    {
      tsa.logger.info( 'Change loss prevention mechanism ' + ( tsa.changesCanBeLostFlag ? 'activated' : 'deactivated' ) + '.' );
      
      // Alert the user if the tab is closed...
      try
      {
        if( tsa.changesCanBeLostFlag )
        {
          if( !window.onbeforeunload )
          {
            window.onbeforeunload = function( event )
            {
              return tsa.i18n.labelFormUnsavedChanges;
            };
          }
        }
        else
        {
          window.onbeforeunload = null;
        }
      }
      catch( error )
      {
        // Ignore: unsupported.
      }
    }
  };
  
  //***************************************************************************
  /**
   * Ask the user to confirm that any change can be discarded.
   * If no changes can be lost, the user will not be bothered,
   * instead this method will simply return <code>true</code>.
   * @public
   * @return {boolean}
   */
  tsa.confirmChangesCanBeDiscarded = function()
  {
    /** @type {boolean} */ var result = false;
    
    if( tsa.changesCanBeLostFlag )
    {
      try
      {
        result = window.confirm( tsa.i18n.labelFormUnsavedChanges );
      }
      catch( error )
      {
        tsa.logger.error( 'Couldn\'t ask the user for confirmation: message unavailable.' );
        result = true;
      }
    }
    else
    {
      result = true;
    }
    
    return result;
  };
  
  //***************************************************************************
  /**
   * Bind (strings into i18n descriptors).
   * @param {string}                  base       Base string. Format: "abc {0} def {1} ghi {2} ...".
   * @param {(string|Array.<string>)} insertions Insertion(s).
   * @return {string}                            Resulting string.
   */
  tsa.bind = function( base, insertions )
  {
    var $  = tsa.$;
    var TV = tsa.TypeVerificator;
    
    /** @type {string} */ var result = '';
    
    if( TV.typeIsString( typeof base ) )
    {
      result = base;
      
      if( TV.typeIsString  ( typeof insertions ) ||
          TV.typeIsBoolean ( typeof insertions ) ||
          TV.typeIsNumber  ( typeof insertions ) )
      {
        insertions = [ insertions ];
      }
      
      if( TV.typeIsObject( typeof insertions ) &&
          TV.isInstanceOf( insertions, Array ) )
      {
        $.each( insertions, function( index, insertion )
        {
          if( TV.typeIsString  ( typeof insertion ) ||
              TV.typeIsBoolean ( typeof insertion ) ||
              TV.typeIsNumber  ( typeof insertion ) )
          {
            var pattern = new RegExp( '\\{' + index + '\\}', 'g' );
            result = result.replace( pattern, '' + insertion );
          }
        } );
      }
    }
    
    return result;
  };
  
  //***************************************************************************
  //* Build left navigation menu, and append it to the document.
  //* The i18n labels have to be already available.
  //***************************************************************************
  
  tsa.buildLeftNav = function()
  {
    require(
      [ 'jquery'
      , 'tsa/widgets/LeftNav'
      ],
      function( $
              , LeftNav )
      {
        var divIbmPcon       = document.getElementById( 'ibm-pcon' );
        var leftNav          = new LeftNav( {} );
        var divIbmNavigation = leftNav.getDomNode();
        
        tsa.updateLeftNav();
        
        $( divIbmPcon ).append( divIbmNavigation );
      }
    ); 
  };
  
  //***************************************************************************
  /**
   * Show or hide the Java logo.
   * @param {boolean} show Default: true. Use false to hide.
   */
  
  tsa.showJavaLogo = function( show )
  {
    show = ( ( typeof show ) !== 'boolean' ) || show;
    
    tsa.$( '#ibm-java-compatible-logo' ).css( { 'display' : ( show ? 'block' : 'none' ) } );
  };
  
  //***************************************************************************
  /**
   * Print a system message.
   * @param {(tsa.cmd.SystemMessage|Object)} sysMsg The system message or the parameters to construct it.
   */
  tsa.printSysMessage = function( sysMsg )
  {
    var $  = tsa.$;
    var TV = tsa.TypeVerificator;
    
    if( TV.typeIsObject( typeof sysMsg ) && !TV.isInstanceOf( sysMsg, tsa.cmd.SystemMessage ) )
    {
      sysMsg = new tsa.cmd.SystemMessage( sysMsg );
    }
    
    if( TV.typeIsObject( typeof sysMsg ) && TV.isInstanceOf( sysMsg, tsa.cmd.SystemMessage ) )
    {
      /** @type {boolean} */ var duplicated = false;
      
      $.each( tsa.systemMessages, function( existingSysMsgIndex, existingSysMsg )
      {
        duplicated = ( existingSysMsg.getMessage() === sysMsg.getMessage() );
        return !duplicated;
      } );
      
      if( !duplicated )
      {
        /** @type {Element} */ var list         = null;
        /** @type {Element} */ var listItem     = document.createElement( 'li' );
        /** @type {Element} */ var icon         = document.createElement( 'img' );
        /** @type {string}  */ var iconSrc      = tsa.attributes.imagesPath + '/';
        /** @type {string}  */ var iconAlt      = '';
        /** @type {Element} */ var msgContainer = document.createElement( 'em' );
        /** @type {Object}  */ var i18n         = tsa.i18n;
        
        if( $( '#flash' ).children( 'ul' ).length === 0 )
        {
          list = document.createElement( 'ul' );
          $( list ).css( { 'margin'          : '0px'
                         , 'padding'         : '8px'
                         , 'list-style-type' : 'none' } );
          $( '#flash' ).append( list );
        }
        else
        {
          list = $( '#flash' ).children( 'ul' ).get(0);
        }
        
        switch( sysMsg.getLevel().toLowerCase() )
        {
          case 'info'  : iconSrc += i18n.imageStatusInfo16    ; alt = i18n.labelInfo    ; break;
          case 'warn'  : iconSrc += i18n.imageStatusWarn16    ; alt = i18n.labelWarn    ; break;
          case 'error' : iconSrc += i18n.imageStatusError16   ; alt = i18n.labelError   ; break;
          default      : iconSrc += i18n.imageStatusUnknown16 ; alt = i18n.labelUnknown ; break;
        }
        
        $( icon ).attr( 'src', iconSrc );
        $( icon ).attr( 'alt', iconAlt );
        $( icon ).css( { 'vertical-align' : 'top'
                       , 'border-width'   : '0px'
                       , 'margin-right'   : '1em' } );
        
        $( msgContainer ).text( sysMsg.getMessage() );
        $( listItem ).append( icon );
        $( listItem ).append( msgContainer );
        $( list ).append( listItem );
        $( '#flash' ).show();
        window.scrollTo( 0, 0 );
        tsa.systemMessages.push( sysMsg );
      }
      else
      {
        tsa.logger.warn( 'Discarding duplicated system message (' + sysMsg.getMessage() + ').' );
      }
    }
    else
    {
      tsa.logger.warn( 'Couldn\'t print system message: invalid system message received.' );
    }
  };
  
  //***************************************************************************
  /**
   * Print multiple system messages.
   * @param {Array.<tsa.cmd.SystemMessage} sysMsgs
   */
  tsa.printSysMessages = function( sysMsgs )
  {
    var $  = tsa.$;
    var TV = tsa.TypeVerificator;
    
    if( TV.typeIsObject( typeof sysMsgs ) && TV.isInstanceOf( sysMsgs, Array ) )
    {
      $.each( sysMsgs, function( sysMsgIndex, sysMsg )
      {
        tsa.printSysMessage( sysMsg );
      } );
    }
  };
  
  //***************************************************************************
  /**
   * Clear the system messages.
   */
  tsa.clearSysMessages = function()
  {
    var $ = tsa.$;
    
    $( '#flash' ).empty();
    $( '#flash' ).hide();
    tsa.systemMessages = [];
  };
  
  //***************************************************************************
  /**
   * Block the UI.
   * @param {Function} cancelHandler Optional. A handler that will be called if
   *                                 the user cancels the blockade. Use <code>null</code>
   *                                 to avoid displaying a cancel button, making this
   *                                 blockade uncancelable.
   */
  tsa.blockUI = function( cancelHandler )
  {
    var $  = tsa.$;
    var TV = tsa.TypeVerificator;
    
    /** @type {Element} */ var cancelButton = document.createElement( 'input' );
    /** @type {boolean} */ var uncancelable = TV.typeIsObject( typeof cancelHandler ) && ( cancelHandler === null );
    
    tsa.uiBlockCount++;
    
    $( cancelButton ).addClass( 'ibm-btn-cancel-pri' );
    $( cancelButton ).attr( 'type', 'button' );
    $( cancelButton ).attr( 'name', 'ibm-cancel' );
    $( cancelButton ).attr( 'value', tsa.i18n.buttonCancel );
    $( cancelButton ).css( { 'visibility' : ( uncancelable ? 'hidden' : 'visible' ) } );
    
    $( cancelButton ).on( 'click', function( event )
    {
      if( !uncancelable && TV.typeIsFunction( typeof cancelHandler ) )
      {
        cancelHandler();
      }
      
      tsa.unblockUI();
    } );
    
    // Some cancel buttons are being shown? Hide them:
    $( '#ajax-wait-btn-container' ).children( 'input' ).hide();
    
    // Append the newly created cancel button:
    $( '#ajax-wait-btn-container' ).append( cancelButton );
    
    if( tsa.uiBlockCount === 1 )
    {
      /** @type {boolean} */ var invisibleCoverExists = ( $( '#invisible-cover' ).length > 0 );
      /** @type {Element} */ var invisibleCover       = ( invisibleCoverExists ? $( '#invisible-cover' ).get( 0 ) : null );
      
      if( invisibleCover === null )
      {
        invisibleCover    = document.createElement( 'div' );
        invisibleCover.id = 'invisible-cover';
        
        $( invisibleCover ).css( { 'opacity'          : '0.33'
                                 , 'cursor'           : 'wait'
                                 , 'background-color' : 'white'
                                 , 'position'         : 'fixed'
                                 , 'width'            : '100%'
                                 , 'height'           : '100%'
                                 , 'top'              : '0px'
                                 , 'left'             : '0px'
                                 , 'z-index'          : '1000' } );
        
        $( 'body' ).append( invisibleCover );
      }
      
      // Do nothing but block screen during tsa.uiBlockDelay; then show a Please Wait screen...
      tsa.uiBlockTimeout = window.setTimeout( function()
      {
        $( '#invisible-cover'     ).remove();
        $( '#ajax-wait'           ).css( { 'display' : ''     } );
        $( '#dynamic-content'     ).css( { 'display' : 'none' } );
        $( '#ibm-welcome-message' ).css( { 'display' : 'none' } );
        
        tsa.uiBlockTimeout = 0;
        
        // Update cursor (webkit bug: cursor doesn't get updated if not moved).
        window.scrollBy(0, 1);
      }
      , tsa.uiBlockDelay );
      
      // Block the left navigation menu, only if it's already loaded:
      if( !TV.typeIsUndefined( typeof tsa.widgets.LeftNav ) && ( tsa.widgets.LeftNav.getInstance() !== null ) )
      {
        tsa.widgets.LeftNav.getInstance().block();
      }
    }
  };
  
  //***************************************************************************
  /**
   * Unblock the UI.
   */
  tsa.unblockUI = function()
  {
    var $  = tsa.$;
    var TV = tsa.TypeVerificator;
    
    tsa.uiBlockCount--;
    
    // Remove the last cancel button:
    $( '#ajax-wait-btn-container' ).children( 'input' ).last().remove();
    
    // Some cancel buttons remain? Show the last one:
    $( '#ajax-wait-btn-container' ).children( 'input' ).last().show();
    
    if( tsa.uiBlockCount <= 0 )
    {
      tsa.uiBlockCount = 0;
      
      // Cancel block-timeout...
      if( tsa.uiBlockTimeout !== 0 )
      {
        window.clearTimeout( tsa.uiBlockTimeout );
        tsa.uiBlockTimeout = 0;
      }
      
      $( '#invisible-cover'     ).remove();
      $( '#ajax-wait'           ).css( { 'display' : 'none' } );
      $( '#dynamic-content'     ).css( { 'display' : ''     } );
      $( '#ibm-welcome-message' ).css( { 'display' : ''     } );
      
      // Update cursor (webkit bug: cursor doesn't get updated if not moved).
      window.scrollBy(0, 1);
      
      // Unblock the left navigation menu, only if it's already loaded:
      if( !TV.typeIsUndefined( typeof tsa.widgets.LeftNav ) && ( tsa.widgets.LeftNav.getInstance() !== null ) )
      {
        tsa.widgets.LeftNav.getInstance().unblock();
      }
    }
  };
  
  //***************************************************************************
  /**
   * User wants to log out.
   */
  tsa.logout = function()
  {
    ( new tsa.cmd.ShowCommand( { action : 'admin/logout' } ) ).execute();
  };
  
  //***************************************************************************
  /**
   * Layout mode:
   * a) 'normal' : It has room for the left navigation menu, the main content section and a right panel.
   * b) 'wide'   : It has room for the left navigation menu and the main content section.
   * c) 'wider'  : The main content section will take all the width. The left navigation menu will be collapsed.
   * @param {string} value The value.
   */
  tsa.setLayout = function( value )
  {
    var $  = tsa.$;
    var TV = tsa.TypeVerificator;
    
    if( TV.typeIsString( typeof value ) )
    {
      /** @type {boolean} */ var minifyLeftNav = false;
      
      $( '#ibm-top' ).removeClass( 'ibm-content-expand' );
      $( '#ibm-top' ).removeClass( 'tsa-wider' );
      
      switch( value )
      {
        case 'wider':
          $( '#ibm-top' ).addClass( 'ibm-content-expand' );
          $( '#ibm-top' ).addClass( 'tsa-wider' );
          minifyLeftNav = true;
          break;
        case 'wide':
          $( '#ibm-top' ).addClass( 'ibm-content-expand' );
          break;
        case 'normal':
        default :
          value = 'normal';
      }
      
      if( !TV.typeIsUndefined( typeof tsa.widgets.LeftNav ) )
      {
        var LeftNav = tsa.widgets.LeftNav;
        
        if( LeftNav.getInstance() !== null )
        {
          LeftNav.getInstance().minify( minifyLeftNav );
        }
      }
      
      tsa.logger.info( 'Layout mode set: "' + value + '".' );
    }
  };
  

  
  //***************************************************************************
  /**
   * Update the global attributes with new ones received from server.
   * @param {Object.<string,string>} attr The new attribute map.
   */
  tsa.updateAttributes = function( attr )
  {
    var TV = tsa.TypeVerificator;
    
    if( TV.typeIsObject( typeof attr ) && ( attr !== null ) )
    {
      // Basic i18n labels must remain...
      if( TV.typeIsObject( typeof tsa.attributes.i18nBasic ) )
      {
        attr.i18nBasic = tsa.attributes.i18nBasic;
      }
      
      tsa.attributes = attr;
      
      // Update session message...
      tsa.updateSessionMessage();
      
      // Update menu...
      tsa.updateLeftNav();
      
      // Log the inactivity time we are allowed to take from now:
      if( TV.typeIsNumber( typeof tsa.attributes.remainingInactivityTimeAllowed ) )
      {
        tsa.logger.info( 'Inactivity time allowed before session expires: ' + tsa.attributes.remainingInactivityTimeAllowed + ' second(s).' );
      }
      
      // Is the server asking to perform a redirection?
      var pathname = window.location.pathname;
      //if we are on the redirect window, then remove redirect attribute, to avoid infinite loop redirections
      if( TV.typeIsString( typeof attr.redirect ) && TV.typeIsString( typeof attr.redirect ) && pathname.indexOf(attr.redirect) > -1){
        delete attr.redirect;
      }
          
          
      if( TV.typeIsString( typeof attr.redirect ) && ( attr.redirect.length > 0 ) )
      {
        tsa.logger.warn( 'The server is asking to perform a redirection (' + attr.redirect + ').' );
        
        window.setTimeout(function()
        {
          require( [ 'tsa/cmd/ShowCommand' ], function( ShowCommand ) {
            ( new ShowCommand( { action : attr.redirect } ) ).execute();
          } );
        }, 250);
      }
    }
  };
  
  //***************************************************************************
  /**
   * Update session message by using application attributes set by the server.
   * This method should be called only when those attributes are updated.
   * Options:
   * a) No user logged in? blank.
   * b) User logged in? welcome message and logout link.
   */
  tsa.updateSessionMessage = function()
  {
    var $  = tsa.$;
    var TV = tsa.TypeVerificator;
    
    if( TV.typeIsString( typeof tsa.attributes.userPrincipal ) && ( tsa.attributes.userPrincipal.length > 0 ) )
    {
      /** @type {Element} */ var logoutLink = document.createElement( 'a' );
      
      $( logoutLink ).text( tsa.i18n.labelBannerLogout );
      $( logoutLink ).attr( 'href', 'javascript:tsa.logout()' );
      
      $( '#ibm-welcome-message' ).text( tsa.i18n.labelBannerWelcome + ' ' + tsa.attributes.userPrincipal + ' [ ' );
      $( '#ibm-welcome-message' ).append( logoutLink );
      $( '#ibm-welcome-message' ).append( document.createTextNode( ' ]' ) );
    }
    else
    {
      $( '#ibm-welcome-message' ).text( '' );
    }
  };
  
  //***************************************************************************
  
  /**
   * Get boolean value of attribute flag.
   * @param  {string}  flagName Attribute flag name.
   * @return {boolean}
   */
  tsa.isAttrFlagOn = function( flagName )
  {
    var TV = tsa.TypeVerificator;
    
    /** @type {boolean} */ var on = false;
    
    if( TV.typeIsString( typeof flagName ) )
    {
      if( TV.typeIsString( typeof tsa.attributes[ flagName ] ) )
      {
        on = ( tsa.attributes[ flagName ].toLowerCase() !== 'false' ) &&
             ( tsa.attributes[ flagName ].toLowerCase() !== 'off'   );
      }
      else if( TV.typeIsBoolean( typeof tsa.attributes[ flagName ] ) )
      {
        on = tsa.attributes[ flagName ];
      }
    }
    
    return on;
  };
  
  //***************************************************************************
  
  /**
   * Is debug mode on?
   * @return {boolean}
   */
  tsa.isDebugModeOn = function()
  {
    return tsa.isAttrFlagOn( 'DEBUG' );
  };
  
  //***************************************************************************
  
  /**
   * Is development mode on?
   * @return {boolean}
   */
  tsa.isDevModeOn = function()
  {
    return tsa.isAttrFlagOn( 'enableDevMenu' );
  };
  
  //***************************************************************************
  /**
   * Update the left navigation menu elements depending on the application
   * attributes set by server. This method should be called only after such
   * attributes are updated.
   * @private
   */
  tsa.updateLeftNav = function()
  {
    require( [ 'tsa/TypeVerificator'
             , 'tsa/widgets/LeftNav' ], function( TV, LeftNav )
    {
      if( LeftNav.getInstance() !== null )
      {
        LeftNav.getInstance().setFlags( { debug : tsa.isDebugModeOn()
                                        , dev   : tsa.isDevModeOn() } );
      }
    } )
  };
  
  //***************************************************************************
  /**
   * Update the auto-refresh settings.
   * @param {?number} interval The auto-refresh interval, in seconds. Use
   *                           <code>null</code> to disable auto-refresh.
   */
  tsa.updateAutoRefreshSettings = function( interval )
  {
    require(
      [ 'jquery'
      , 'tsa/TypeVerificator'
      , 'tsa/cmd/ShowCommand'
      ],
      function( $
              , TV
              , ShowCommand )
      {
        /** @type {tsa.cmd.ShowCommand} */ var lastShowCmd = tsa.cmd.ShowCommand.getLastInstanceSuccessfullyExecuted();
        /** @type {number}              */ var MIN_INTERVAL = 5; // seconds
        
        var clearAutoRefreshInterval = function()
        {
          window.clearInterval( tsa.autoRefreshIntervalId );
          tsa.autoRefreshIntervalId = null;
          $( '#tsa-auto-refresh-info' ).text( tsa.NBSP );
          tsa.logger.info( 'Content auto-refresh settings cleared.' );
        };
        
        if( TV.typeIsNumber( typeof interval ) && !isNaN( interval ) )
        {
          if( lastShowCmd !== null )
          {
            /** @type {Date} */ var start = new Date();
            
            if( interval < MIN_INTERVAL )
            {
              interval = MIN_INTERVAL;
              tsa.logger.warn( 'Automatic refresh interval can\'t be less than the minimum allowed (' + MIN_INTERVAL + ' seconds); adjusted.' );
            }
            
            tsa.logger.info( 'The content will be automatically refreshed in ' + interval + ' seconds.' );
            
            tsa.autoRefreshIntervalId = window.setInterval( function()
            {
              /** @type {Date}   */ var now  = new Date();
              /** @type {number} */ var diff = interval - Math.floor( ( now.getTime() - start.getTime() ) / 1000 );
              
              if( diff <= 0 )
              {
                tsa.logger.info( 'Content will be automatically refreshed now.' );
                clearAutoRefreshInterval();
                lastShowCmd.clone( true ).execute();
              }
              else
              {
                $( '#tsa-auto-refresh-info' ).text( tsa.bind( tsa.i18n.auto_refresh_info, diff ) );
              }
            }
            , 1000 );
          }
          else
          {
            tsa.logger.warn( 'Automatic refresh request will be ignored (the action to be refreshed is unknown).' );
          }
        }
        else if( !TV.typeIsObject( typeof interval ) || ( interval !== null ) )
        {
          tsa.logger.warn( 'Invalid automatic refresh delay will be ignored (not a number).' );
        }
        else if( !TV.typeIsObject( typeof tsa.autoRefreshIntervalId ) || ( tsa.autoRefreshIntervalId !== null ) )
        {
          clearAutoRefreshInterval();
        }
      }
    );
  };
  
  //***************************************************************************
  /**
   * Build a ShowCommand instance by using the original attributes set by the
   * server (what is the server requesting the client to show?)
   */
  
  tsa.buildShowCommandFromAttributes = function()
  {
    var $           = tsa.$;
    var TV          = tsa.TypeVerificator;
    var ShowCommand = tsa.cmd.ShowCommand;
    
    /** @type {tsa.cmd.ShowCommand} */ var showCmd = null;
    
    // Executing the required show command...
    if( ( ( typeof tsa.attributes.action  ) === 'string' ) && ( tsa.attributes.action  .length > 0 ) &&
        ( ( typeof tsa.attributes.command ) === 'string' ) && ( tsa.attributes.command .length > 0 ) )
    {
      /** @type {tsa.cmd.ShowCommand} */ var showCmd    = null;
      /** @type {string}              */ var action     = tsa.attributes.action + '/' + tsa.attributes.command;
      /** @type {Object}              */ var parameters = ( TV.typeIsObject( typeof tsa.attributes.parameters )
                                                            ? tsa.attributes.parameters : null );
      
      if( TV.typeIsString( typeof tsa.attributes.entityId ) && ( tsa.attributes.entityId.length > 0 ) )
      {
        action += '/' + tsa.attributes.entityId;
      }
      
      showCmd = new ShowCommand( { action     : action
                                 , parameters : parameters } );
      
      if( ( action === 'admin/login' ) && TV.typeIsString( typeof tsa.attributes.duration ) )
      {
        showCmd.setParameter( 'loginLockDuration', tsa.attributes.duration );
      }
    }
    else
    {
      tsa.logger.warn( 'The server didn\'t requested any ShowCommand to run.' );
    }
    
    return showCmd;
  };
  
  //***************************************************************************
  //* Post-processing the loaded document...
  //***************************************************************************
  
  tsa.jQuery(document).ready(function() {
    
    require(
        
      [ 'jquery'
      , 'dojo/_base/kernel'
      , 'tsa/FormCustomizer'
      , 'tsa/PopStateHandler'
      , 'tsa/TypeVerificator'
      , 'tsa/cmd/AjaxResponse'
      , 'tsa/cmd/Command'
      , 'tsa/cmd/ShowCommand'
      ],
      
      function( $
              , DojoKernel
              , FormCustomizer
              , PopStateHandler
              , TV
              , AjaxResponse // required to load the SystemMessage inner class.
              , Command
              , ShowCommand
              )
      {
        // Printing the Dojo (used exclusively as Module Loader) version:
        if (DojoKernel && DojoKernel.version) {
          
          tsa.logger.info('Dojo (Module Loader) version: ' + DojoKernel.version.major
                         +                             '.' + DojoKernel.version.minor
                         +                             '.' + DojoKernel.version.patch
                         +                   ' (revision ' + DojoKernel.version.revision + ').');
        }
        
        /*
         * An 'attributes' object is defined for UI v2.
         * This object contains some UI properties that has to be processed now.
         */
        if( ( typeof tsa.attributes ) !== 'undefined' )
        {
          // What should be shown, according to the server?
          /** @type {tsa.cmd.ShowCommand} */ var showCmd = tsa.buildShowCommandFromAttributes();
          
          // Setting the i18n labels into the template...
          try
          {
            /** @type {string} */ var printVersion = TV.typeIsString( typeof tsa.attributes.tsaProductVersion );
            
            for( /** @type {string} */ var i18nLabelName in tsa.attributes.i18nBasic )
            {
              tsa.i18n[ i18nLabelName ] = tsa.attributes.i18nBasic[ i18nLabelName ];
            }
            
            tsa.logger.info( 'Setting site title: "' + tsa.i18n.labelBannerTitle + '".' );
            tsa.logger.info( 'Setting site name (header): "' + tsa.i18n.labelBannerSiteName + '".' );
            tsa.logger.info( 'Setting product version: ' + ( printVersion ? '"' + tsa.attributes.tsaProductVersion + '"' : 'unavailable' ) + '.' );
            
            $( 'title' ).text( tsa.i18n.labelBannerTitle );
            $( '.site-name' ).text( tsa.i18n.labelBannerSiteName + ( printVersion ? ' ' + tsa.attributes.tsaProductVersion : '' ) );
            $( '#ibm-tsa-icon' ).attr( 'alt', tsa.i18n.labelTSA );
            $( '#ajax-wait-msg' ).text( tsa.i18n.labelPleaseWait );
            $( '#ajax-wait-img' ).attr( 'alt', tsa.i18n.labelPleaseWait );
          }
          catch( error )
          {
            tsa.logger.error(error);
            tsa.logger.error('Some i18n labels could not be applied over the HTML Document.');
          }
          
          /*
           * The complete collection of i18n labels will be retrieved if the user is logged in;
           * basic i18n labels for the login and logout screens have been already set.
           */
          if( ( ( typeof tsa.attributes.userPrincipal ) === 'string' ) && ( tsa.attributes.userPrincipal.length > 0 ) )
          {
            var onSucceeded = function( event )
            {
              tsa.i18n = event.response.payload.i18nLabels;
              tsa.logger.info( 'I18n labels retrieved.' );
              tsa.buildLeftNav();
              
              if( showCmd !== null )
              {
                showCmd.execute();
              }
            };
            
            var onFailed = function( failureMessage ) {
              tsa.logger.error( 'I18n labels couldn\'t be retrieved.' );
            };
            
            var cmd = new Command( { action        : 'i18n/getI18nLabels'
                                   , eventHandlers : { executionSucceeded : onSucceeded
                                                     , executionFailed    : onFailed }
                                   , secureToken   : false } );
            cmd.execute();
          }
          else
          {
            /*
             * The content to be show doesn't need the i18n labels. It could
             * be, for example, the login or logout screen.
             */
            if( showCmd !== null )
            {
              showCmd.execute();
            }
          }
          
          // Enable Pop State Event Handling...
          PopStateHandler.enable();
        }
        else
        {
          // Some forms might need to be customised:
          FormCustomizer.getInstance().run();
        }
      } /* end of require function */ );
  } /* end of ready function*/ );
  
  //***************************************************************************
  
} )();
