/*
 * IBM Technical Support Appliance Version 2 Release 0 Modification 0
 *
 * (C) Copyright IBM Corp. 2017
 */

define(
  
  [ 'jquery'
  , 'tsa/TypeVerificator'
  , 'tsa/cmd/Command'
  ],
  
  function( $
          , TV
          , Command )
  {
    //-------------------------------------------------------------------------
    
    var logger = tsa.getLogger( 'ContentArea' );
    
    //-------------------------------------------------------------------------
    /**
     * @constructor
     * Note: remember to replicate this documentation in DevWidgetsHandler.
     * Parameters:
     *   title          The title.
     *   helpDoc        The name of the help document. Optional.
     *   showBackToTop  Show a back-to-top link at the end of the area. Optional; default: false.
     * 
     * @param {{title         : string
     *         ,helpDoc       : (string|undefined)
     *         ,showBackToTop : (boolean|undefined)}} parameters.
     */
    tsa.widgets.ContentArea = function( parameters )
    {
      if( !TV.typeIsObject( typeof parameters ) || ( parameters === null ) ) parameters = {};
      
      this.domNode        = document.createElement( 'div' );
      this.headElement    = document.createElement( 'div' );
      this.bodyElement    = document.createElement( 'div' );
      this.mainElement    = document.createElement( 'div' );
      this.sidebarElement = document.createElement( 'div' );
      
      $( this.domNode        ).attr( 'role' , 'main' );
      $( this.domNode        ).attr( 'aria-label' , 'Content for ContentArea' );
      $( this.headElement    ).attr( 'id'   , 'ibm-content-head' );
      $( this.bodyElement    ).attr( 'id'   , 'ibm-content-body' );
      $( this.mainElement    ).attr( 'id'   , 'ibm-content-main' );
      $( this.sidebarElement ).attr( 'id'   , 'ibm-content-sidebar' );
      
      $( this.headElement ).css( { 'position' : 'relative' } );
      
      $( this.bodyElement ).append( this.mainElement    );
      $( this.bodyElement ).append( this.sidebarElement );
      $( this.domNode     ).append( this.headElement    );
      $( this.domNode     ).append( this.bodyElement    );
      
      this.setTitle         ( TV.typeIsString  ( typeof parameters.title         ) ? parameters.title         : ''    );
      this.setHelpDoc       ( TV.typeIsString  ( typeof parameters.helpDoc       ) ? parameters.helpDoc       : ''    );
      this.setShowBackToTop ( TV.typeIsBoolean ( typeof parameters.showBackToTop ) ? parameters.showBackToTop : false );
      
      logger.info('ContentArea instance created.');
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * The help window.
     * @private
     * @type {Window}
     */
    tsa.widgets.ContentArea.helpWindow = null;
    
    //-------------------------------------------------------------------------
    
    /**
     * The HTML DOM Node.
     * @private
     * @type {Element}
     */
    tsa.widgets.ContentArea.prototype.domNode = null;
    
    /**
     * The show back to top flag.
     * @private
     * @type {boolean}
     */
    tsa.widgets.ContentArea.prototype.showBackToTop = false;
    
    /**
     * The name of the associated help document.
     * @private
     * @type {string}
     */
    tsa.widgets.ContentArea.prototype.helpDoc = '';
    
    /**
     * Head HTML Element.
     * @private
     * @type {Element}
     */
    tsa.widgets.ContentArea.prototype.headElement = null;
    
    /**
     * Body HTML Element.
     * @private
     * @type {Element}
     */
    tsa.widgets.ContentArea.prototype.bodyElement = null;
    
    /**
     * Title HTML Element.
     * @private
     * @type {Element}
     */
    tsa.widgets.ContentArea.prototype.titleElement = null;
    
    /**
     * Help link HTML Element.
     * @private
     * @type {Element}
     */
    tsa.widgets.ContentArea.prototype.helpElement = null;
    
    /**
     * Body's Main HTML Element (main body's child).
     * @private
     * @type {Element}
     */
    tsa.widgets.ContentArea.prototype.mainElement = null;
    
    /**
     * Back to top HTML Element.
     * @private
     * @type {Element}
     */
    tsa.widgets.ContentArea.prototype.backToTopElement = null;
    
    //-------------------------------------------------------------------------
    
    /**
     * Set title. Use an empty string to hide the title.
     * @param {string} title Title.
     */
    tsa.widgets.ContentArea.prototype.setTitle = function( title )
    {
      if( !TV.typeIsString( typeof title ) )
      {
        logger.error( 'Invalid title (' + ( typeof title ) + ').' );
        title = '';
      }
      
      if( this.titleElement === null )
      {
        this.titleElement = document.createElement( 'h1' );
        $( this.headElement ).append( this.titleElement );
      }
      
      if( ( title = $.trim( title ) ).length > 0 )
      {
        $( this.titleElement ).show();
        $( this.titleElement ).text( title );
      }
      else
      {
        $( this.titleElement ).hide();
        $( this.titleElement ).text( '' );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Set the show-back-to-top-link flag.
     * @param {boolean} flag The flag.
     */
    tsa.widgets.ContentArea.prototype.setShowBackToTop = function( flag )
    {
      this.showBackToTop = !!flag;
      
      logger.info( 'Back to top link ' + ( this.showBackToTop ? 'will' : 'won\'t' ) + ' be displayed.' );
      
      if( this.showBackToTop )
      {
        if( this.backToTopElement === null )
        {
          /** @type {Element} */ var link = document.createElement( 'a' );
          
          this.backToTopElement = document.createElement( 'p' );
          $( this.backToTopElement ).append( link );
          $( this.backToTopElement ).css( { 'margin-top' : '1.2em' } );
          $( this.backToTopElement ).addClass( 'ibm-back-link' );
          $( link ).addClass( 'ibm-anchor-up-link' );
          $( link ).attr( 'href', '#ibm-content' );
          $( link ).text( tsa.i18n.buttonBackToTop );
          
          link.onclick = function( event ) { window.scrollTo( 0, 0 ); return false; };
          
          $( this.mainElement ).append( this.backToTopElement );
        }
        
        $( this.backToTopElement ).show();
      }
      else if( this.backToTopElement !== null )
      {
        $( this.backToTopElement ).hide();
      }
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Set the name of the help document.
     * Base help URL will be prepended.
     * @param {string} helpDoc The document (to be retrieved from the base documents URL).
     */
    tsa.widgets.ContentArea.prototype.setHelpDoc = function( helpDoc )
    {
      if( TV.typeIsString( typeof helpDoc ) )
      {
        this.helpDoc = $.trim( helpDoc );
      }
      else
      {
        logger.error( 'Invalid help URL (' + ( typeof helpDoc ) + ').' );
        this.helpDoc = '';
      }
      
      // Don't modify it if already an URL:
      if( ( this.helpDoc.length              >   0   ) &&
          ( this.helpDoc.charAt( 0 )         !== '/' ) &&
          ( this.helpDoc.indexOf( 'http:'  ) !== 0   ) &&
          ( this.helpDoc.indexOf( 'https:' ) !== 0   ) )
      {
        this.helpDoc = Command.URL_SERVLET_CONTEXT + 'help/' + helpDoc;
      }
      
      if( this.helpElement === null )
      {
        this.helpElement = document.createElement( 'img' );
        
        $( this.helpElement ).attr( 'alt'   , tsa.i18n.labelHelp );
        $( this.helpElement ).attr( 'title' , tsa.i18n.tooltipPageHelp );
        $( this.helpElement ).attr( 'src'   , tsa.attributes.imagesPath + '/' + 'help16.gif' );
        $( this.helpElement ).attr( 'role'  , 'link' );
        
        $( this.helpElement ).css( { 'width'        : '16px'
                                   , 'height'       : '16px'
                                   , 'border-width' : '0px'
                                   , 'position'     : 'absolute'
                                   , 'right'        : '0px'
                                   , 'top'          : '0px'
                                   , 'cursor'       : 'pointer' } );
        
        $( this.helpElement ).on( 'click', $.proxy( function( event )
        {
          tsa.widgets.ContentArea.showHelp( this.helpDoc );
        }
        , this ) );
        
        $( this.headElement ).append( this.helpElement );
      }
      
      $( this.helpElement ).css( { 'display' : ( ( this.helpDoc.length > 0 ) ? '' : 'none' ) } );
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Append message. Parameters:
     *   level      : Optional. Values: 'HELP', 'INFO', 'WARN', 'ERROR'. Default: no level.
     *   border     : Optional. Default: false.
     *   text       : The message text.
     *   textIsHtml : Handle text as HTML. Optional; default: false.
     * 
     * @param {{ level      : (string|undefined)
     *         , border     : (boolean|undefined)
     *         , text       : string
     *         , textIsHtml : (boolean|undefined) }} parameters
     * @return {Element} The HTML Element that has just been added.
     */
    tsa.widgets.ContentArea.prototype.appendMessage = function( parameters )
    {
      /** @type {Element} */ var result = null;
      
      if( TV.typeIsObject( typeof parameters ) && ( parameters !== null ) )
      {
        if( TV.typeIsString( typeof parameters.text ) )
        {
          /** @type {Element} */ var container     = document.createElement( 'div' );
          /** @type {Element} */ var containerBody = document.createElement( 'div' );
          /** @type {Element} */ var paragraph     = document.createElement( 'p' );
          /** @type {boolean} */ var textIsHtml    = TV.typeIsBoolean( typeof parameters.textIsHtml ) && parameters.textIsHtml;
          /** @type {boolean} */ var border        = TV.typeIsBoolean( typeof parameters.border     ) && parameters.border;
          
          $( container     ).append( containerBody );
          $( containerBody ).append( paragraph );
          $( container     ).addClass( 'ibm-container' );
          $( containerBody ).addClass( 'ibm-container-body' );
          
          this.append( container );
          result = container;
          
          $( container ).addClass( border ? 'ibm-alternate-two' : 'ibm-alternate' );
          $( paragraph )[ textIsHtml ? 'html' : 'text' ]( parameters.text );
          $( paragraph ).css( { 'white-space' : ( textIsHtml ? 'normal' : 'pre-line' ) } );
          
          if( TV.typeIsString( typeof parameters.level ) )
          {
            /** @type {string} */ var className = '';
            
            switch( parameters.level.toUpperCase() )
            {
              case 'HELP' : className = 'ibm-ind-help'        ; break;
              case 'INFO' : className = 'ibm-ind-information' ; break;
              case 'WARN' : className = 'ibm-ind-caution'     ; break;
              case 'ERROR': className = 'ibm-ind-error'       ; break;
            }
            
            if( className.length > 0 )
            {
              $( paragraph ).addClass( className );
            }
          }
        }
        else
        {
          logger.error( 'Couldn\'t append message: invalid text (' + ( typeof parameters.text ) + ').' );
        }
      }
      else
      {
        logger.error( 'Couldn\'t append message: invalid parameters.' );
      }
      
      return result;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Append HTML Element.
     * @param {Element} element
     */
    tsa.widgets.ContentArea.prototype.append = function( element )
    {
      if( TV.typeIsObject( typeof element ) && ( element !== null ) && !TV.typeIsUndefined( typeof element.nodeType ) )
      {
        $( this.mainElement ).append( element );
        
        // Ensure the back-to-top link is at the end of this content area:
        if( this.backToTopElement !== null )
        {
          $( this.backToTopElement ).remove();
          $( this.mainElement ).append( this.backToTopElement );
        }
      }
      else
      {
        logger.error( 'Invalid element.' );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Append sidebar panel.
     * @param {{title : string
     *         ,body  : Element}} data
     * @return {Element} The panel element.
     */
    tsa.widgets.ContentArea.prototype.appendSidebarPanel = function( panelData )
    {
      /** @type {Element} */ var panelElement = null;
      
      if( TV.typeIsObject( typeof panelData       ) && ( panelData !== null ) &&
          TV.typeIsString( typeof panelData.title ) &&
          TV.typeIsObject( typeof panelData.body  ) && ( panelData.body !== null ) &&
          !TV.typeIsUndefined( typeof panelData.body.nodeType ) )
      {
        /** @type {Element} */ var headerElement = document.createElement( 'h2' );
        /** @type {Element} */ var bodyElement   = document.createElement( 'div' );
        
        panelElement = document.createElement( 'div' );
        
        $( panelElement ).addClass( 'ibm-container'      );
        $( bodyElement  ).addClass( 'ibm-container-body' );
        
        $( headerElement ).text( panelData.title );
        $( bodyElement   ).append( panelData.body );
        $( panelElement  ).append( headerElement );
        $( panelElement  ).append( bodyElement );
        
        $( this.sidebarElement ).append( panelElement );
      }
      else
      {
        logger.error( 'Invalid panel data.' );
      }
      
      return panelElement;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Append side-bar accordion panel. The first element is the one initially expanded.
     * @param {Array.<{title : string
     *                ,body  : Element>}} panelData
     * @return {Element} The panel element.
     */
    tsa.widgets.ContentArea.prototype.appendSidebarAccordionPanel = function( panelData )
    {
      /** @type {Element} */ var panelElement = null;
      
      if( TV.typeIsObject( typeof panelData ) && TV.isInstanceOf( panelData, Array ) )
      {
        panelElement = document.createElement( 'div' );
        
        $( panelElement ).addClass( 'ibm-container' );
        $( panelElement ).addClass( 'ibm-show-hide' );
        
        var showSubPanel = function( panelIndex )
        {
          $( panelElement ).children( 'h2' ).children( 'a' ).removeClass( 'ibm-show-active' );
          $( panelElement ).children( 'div' ).css( { 'display' : 'none' } );
          
          $( $( panelElement ).children( 'h2' ).get( panelIndex ) ).children( 'a' ).addClass( 'ibm-show-active' );
          $( $( panelElement ).children( 'div' ).get( panelIndex ) ).css( { 'display' : 'block' } );
        };
        
        $.each( panelData, $.proxy( function( panelDataEntryIndex, panelDataEntry )
        {
          if( TV.typeIsString( typeof panelDataEntry.title ) &&
              TV.typeIsObject( typeof panelDataEntry.body  ) && ( panelDataEntry.body !== null ) &&
              !TV.typeIsUndefined( typeof panelDataEntry.body.nodeType ) )
          {
            /** @type {boolean} */ var isFirst       = ( panelDataEntryIndex === 0 );
            /** @type {Element} */ var headerElement = document.createElement( 'h2' );
            /** @type {Element} */ var headerLink    = document.createElement( 'a' );
            /** @type {Element} */ var bodyElement   = document.createElement( 'div' );
            
            $( headerLink ).text( panelDataEntry.title );
            $( headerLink ).on( 'click', function( event )
            {
              try
              {
                event.stopPropagation();
              }
              catch( error ) { /* Ignore. */ }
              
              showSubPanel( panelDataEntryIndex );
              return false;
            } );
            
            headerLink.href = '#';
            
            $( bodyElement ).addClass( 'ibm-container-body' );
            $( bodyElement ).attr( 'role', 'region' );
            $( bodyElement ).attr( 'aria-live', 'polite' );
            $( bodyElement ).attr( 'aria-label', panelDataEntry.title );
            
            if( !isFirst )
            {
              $( headerElement ).css( { 'border-top-style' : 'none' } );
            }
            
            $( headerElement ).append( headerLink );
            $( bodyElement   ).append( panelDataEntry.body );
            $( panelElement  ).append( headerElement );
            $( panelElement  ).append( bodyElement );
            
            $( this.sidebarElement ).append( panelElement );
          }
          else
          {
            logger.error( 'Invalid entry in panel data (index: ' + panelDataEntryIndex + ').' );
          }
        }
        , this ) );
        
        showSubPanel( 0 );
      }
      else
      {
        logger.error( 'Invalid panel data.' );
      }
      
      return panelElement;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Get the DOM Node.
     * @return {Element}
     */
    tsa.widgets.ContentArea.prototype.getDomNode = function()
    {
      return this.domNode;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Show help by opening a dialog window.
     * @param {string} url The URL containing the help document.
     */
    tsa.widgets.ContentArea.showHelp = function( url )
    {
      if( ( tsa.widgets.ContentArea.helpWindow !== null ) &&
          !tsa.widgets.ContentArea.helpWindow.closed )
      {
        tsa.widgets.ContentArea.helpWindow.close();
      } 
      
      if( TV.typeIsString( typeof url ) )
      {
        tsa.widgets.ContentArea.helpWindow = window.open( url, 'tsa_help', 'width=600,height=400,resizable=yes,scrollbars=yes,menubar=no,status=no' );
      }
      else
      {
        logger.error( 'Couldn\'t show help URL: invalid URL (' + ( typeof url ) + ').' );
      }
    };
    
    //-------------------------------------------------------------------------
    
    return tsa.widgets.ContentArea;
  }
);