/*
 * IBM Technical Support Appliance Version 2 Release 6 Modification 1
 *
 * (C) Copyright IBM Corp. 2018, 2020
 */

//-----------------------------------------------------------------------------

define(
  [ 'jquery'
  , 'tsa/TypeVerificator'
  , 'tsa/VerificationToolkit'
  , 'tsa/cmd/Command'
  , 'tsa/widgets/ButtonFactory'
  ],
  
  function( $
          , TV
          , VT
          , Command
          , ButtonFactory
          )
  {
    //-------------------------------------------------------------------------
    
    var logger = tsa.getLogger( 'FormPanel' );
    
    //-------------------------------------------------------------------------
    
    /** @constructor
     *  A form panel. You can link multiple form panels (as a chain), sharing a single button set.
     *  Parameters:
     *    title         The title. Optional. Default: none (no title will be shown).
     *    subtitle      The subtitle. Optional. Default: none (no subtitle will be shown).
     *    description   The description. Optional. Default: none.
     *    confirmMsg    The submission confirmation message. Optional. Default: none.
     *    width         The width, in pixels (optional; default: 100%).
     *    border        Show border. Optional. Default: true.
     *    ajax          The AJAX flag. Optional. Default: true.
     *    cancelable    If the UI is blocked because of an AJAX request, a cancel button is shown. Default: true.
     *    action        The action ("controller_or_handler/method").
     *    linkTo        Other (slave) FormPanel to be linked by this FormPanel.
     *    showReqMarks  Show marks for the required field. Optional. Default: true.
     *    showReqHeader Show the required fields header. Optional. Default: true.
     *    csrfRequired  Generate and set CSRF secure token to the form before submitting. Optional. Default: true.
     *  
     *  @param {{title        	: string
     *          ,subtitle     	: string
     *          ,description  	: string
     *          ,confirmMsg   	: string
     *          ,border       	: boolean
     *          ,ajax         	: boolean
     *          ,cancelable   	: boolean
     *          ,action       	: string
     *          ,linkTo       	: tsa.widgets.FormPanel
     *          ,showReqMarks 	: boolean
     *          ,showReqHeader 	: boolean}} parameters The parameters object. */
    
    tsa.widgets.FormPanel = function( parameters )
    {
      if( !TV.typeIsObject( typeof parameters ) || ( parameters === null ) ) parameters = {};
      
      /** @type {string} */ var DEFAULT_METHOD   = 'POST';
      /** @type {string} */ var DEFAULT_ACTION   = 'UnknownAction';
      /** @type {string} */ var DEFAULT_ENC_TYPE = 'application/x-www-form-urlencoded';
      
      this.hiddenInputMap = {};
      this.validationRules = {};
      this.initUI();
      
      if( TV.typeIsString  ( typeof parameters.title        ) ) this.setTitle        ( $.trim( parameters.title       ) );
      if( TV.typeIsString  ( typeof parameters.subtitle     ) ) this.setSubtitle     ( $.trim( parameters.subtitle    ) );
      if( TV.typeIsString  ( typeof parameters.description  ) ) this.setDescription  ( $.trim( parameters.description ) );
      if( TV.typeIsString  ( typeof parameters.confirmMsg   ) ) this.setConfirmMsg   ( $.trim( parameters.confirmMsg  ) );
      if( TV.typeIsNumber  ( typeof parameters.width        ) ) this.setWidth        ( parameters.width        		);
      if( TV.typeIsBoolean ( typeof parameters.border       ) ) this.setBorder       ( parameters.border      		);
      if( TV.typeIsBoolean ( typeof parameters.ajax         ) ) this.setAjax         ( parameters.ajax         		);
      if( TV.typeIsBoolean ( typeof parameters.cancelable   ) ) this.setCancelable   ( parameters.cancelable   		);
      if( TV.typeIsObject  ( typeof parameters.linkTo       ) ) this.linkTo          ( parameters.linkTo       		);
      if( TV.typeIsBoolean ( typeof parameters.showReqMarks ) ) this.setShowReqMarks ( parameters.showReqMarks 		);
      if( TV.typeIsBoolean ( typeof parameters.showReqHeader) ) this.setShowReqHeader( parameters.showReqHeader 	);
      if( TV.typeIsBoolean ( typeof parameters.csrfRequired ) ) this.setCsrfRequired ( parameters.csrfRequired   );
      
      // Form's attributes...
      this.setMethodAttribute  ( DEFAULT_METHOD );
      this.setActionAttribute  ( TV.typeIsString( typeof parameters.action ) ? parameters.action : DEFAULT_ACTION );
      this.setEnctypeAttribute ( DEFAULT_ENC_TYPE );
      
      // Avoid the form to be submitted the native way...
      this.form.onsubmit = function( event ) { return false; };
    };
    
    //-------------------------------------------------------------------------
    
    /** @type {boolean}                */ tsa.widgets.FormPanel.prototype.ajax            = true;
    /** @type {boolean}                */ tsa.widgets.FormPanel.prototype.cancelable      = true;
    /** @type {boolean}                */ tsa.widgets.FormPanel.prototype.showReqMarks    = true;
    /** @type {boolean}                */ tsa.widgets.FormPanel.prototype.showReqHeader   = true;
    /** @type {boolean}                */ tsa.widgets.FormPanel.prototype.csrfRequired    = true;
    /** @type {string}                 */ tsa.widgets.FormPanel.prototype.title           = '';
    /** @type {string}                 */ tsa.widgets.FormPanel.prototype.subtitle        = '';
    /** @type {string}                 */ tsa.widgets.FormPanel.prototype.description     = '';
    /** @type {string}                 */ tsa.widgets.FormPanel.prototype.confirmMsg      = '';
    /** @type {boolean}                */ tsa.widgets.FormPanel.prototype.border          = true;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.domNode         = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.panelDiv        = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.buttonDiv       = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.titleHeader     = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.bodyDiv         = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.form            = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.descParagraph   = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.descHR          = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.reqLegendSpan   = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.layoutTable     = null;
    /** @type {Element}                */ tsa.widgets.FormPanel.prototype.layoutTableBody = null;
    /** @type {Object.<string,string>} */ tsa.widgets.FormPanel.prototype.hiddenInputMap  = null;
    /** @type {tsa.widgets.FormPanel}  */ tsa.widgets.FormPanel.prototype.linkedTo        = null;
    /** @type {tsa.widgets.FormPanel}  */ tsa.widgets.FormPanel.prototype.linkedBy        = null;
    /** @type {Object.<string,?>}      */ tsa.widgets.FormPanel.prototype.validationRules = null;
    
    //-----------------------------------------------------------------------
    
    /** Get the main DOM Node (an HTML DIV Element).
     *  @return {Element} */
    
    tsa.widgets.FormPanel.prototype.getDomNode = function()
    {
      return this.domNode;
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Initialize the UI elements.
     * 
     *   this.domNode
     *       |
     *       +-- this.form
     *               |
     *               +-- this.panelDiv
     *               |       |
     *               |       +-- this.titleHeader
     *               |       +-- this.bodyDiv
     *               |               |
     *               |               +-- this.descParagraph
     *               |               +-- this.descHR
     *               |               +-- this.layoutTable
     *               |
     *               +-- this.buttonDiv
     * 
     * @private
     */
    tsa.widgets.FormPanel.prototype.initUI = function()
    {
      /** @type {Element} */ var colgroup = document.createElement( 'colgroup' );
      
      this.domNode         = document.createElement( 'div'   );
      this.panelDiv        = document.createElement( 'div'   );
      this.titleHeader     = document.createElement( 'h2'    );
      this.bodyDiv         = document.createElement( 'div'   );
      this.buttonDiv       = document.createElement( 'div'   );
      this.form            = document.createElement( 'form'  );
      this.descParagraph   = document.createElement( 'p'     );
      this.descHR          = document.createElement( 'div'   );
      this.layoutTable     = document.createElement( 'table' );
      this.layoutTableBody = document.createElement( 'tbody' );
      
      $( this.layoutTable ).append( colgroup             );
      $( this.layoutTable ).append( this.layoutTableBody );
      $( this.bodyDiv     ).append( this.descParagraph   );
      $( this.bodyDiv     ).append( this.descHR          );
      $( this.bodyDiv     ).append( this.layoutTable     );
      $( this.panelDiv    ).append( this.titleHeader     );
      $( this.panelDiv    ).append( this.bodyDiv         );
      $( this.form        ).append( this.panelDiv        );
      $( this.form        ).append( this.buttonDiv       );
      $( this.domNode     ).append( this.form            );
      
      $( this.panelDiv  ).addClass( 'ibm-container'      );
      $( this.buttonDiv ).addClass( 'ibm-buttons-row'    );
      $( this.bodyDiv   ).addClass( 'ibm-container-body' );
      $( this.descHR    ).addClass( 'ibm-rule'           );
      $( this.form      ).addClass( 'ibm-row-form'       );
      
      $( this.descParagraph ).hide();
      $( this.descHR        ).hide();
      $( this.titleHeader   ).hide();
      $( this.buttonDiv     ).hide();
      $( this.layoutTable   ).attr( 'role', 'presentation' );
      $( this.layoutTable   ).css( { 'width'           : '100%'
                                   , 'padding'         : '0px'
                                   , 'border-style'    : 'none'
                                   , 'border-collapse' : 'collapse' } );
      $( this.panelDiv      ).css( { 'margin-bottom'   : '14px' } );
      $( this.buttonDiv     ).css( { 'margin-bottom'   : '14px' } );
      
      $( this.descHR ).append( document.createElement( 'hr' ) );
      
      // Column widths:
      $.each( [ 30, 70 ], function( colIndex, colWidth )
      {
        /** @type {Element} */ var column = document.createElement( 'col' );
        
        $( column ).css( { 'width' : colWidth + '%' } );
        $( colgroup ).append( column );
      } );
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Add a linked FormPanel.
     * @param {tsa.widgets.FormPanel} formPanel The FormPanel to be linked by this FormPanel.
     */
    tsa.widgets.FormPanel.prototype.linkTo = function( formPanel )
    {
      if( TV.typeIsObject( typeof formPanel ) && TV.isInstanceOf( formPanel, tsa.widgets.FormPanel ) )
      {
        if( formPanel.linkedBy === null )
        {
          if( this.linkedTo === null )
          {
            formPanel.linkedBy = this;
            this.linkedTo = formPanel;
            $( this.getHeadLink().form ).append( $( this.linkedTo.form ).children() );
            
            // A global "required" legend will replace local legends if necessary...
            this.useGlobalRequiredFieldsLegendIfNeeded();
          }
          else
          {
            logger.error( 'Cannot link to FormPanel: another FormPanel has already been linked.' );
          }
        }
        else
        {
          logger.error( 'Cannot link to FormPanel: the given FormPanel has been already linked.' );
        }
      }
      else
      {
        logger.error( 'Unable to link: not a FormPanel.' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Get the first ancestor in the linked forms chain, in case this form is
     * linked; if the form is not linked, self is returned.
     * @return {tsa.widgets.FormPanel}
     */
    tsa.widgets.FormPanel.prototype.getHeadLink = function()
    {
      /** @type {tsa.widgets.FormPanel} */ var head = this;
      /** @type {number}                */ var max  = 32;
      
      while( ( head.linkedBy !== null ) && ( ( max-- ) >= 0 ) )
      {
        head = head.linkedBy;
      }
      
      return head;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Use a global Required Fields legend, if necessary (when this form is linked to/by another form).
     * If it is needed, local legends will be hidden.
     */
    tsa.widgets.FormPanel.prototype.useGlobalRequiredFieldsLegendIfNeeded = function()
    {
    	var setGlobalLegend = false;
    	if( ( this.getHeadLink().linkedTo !== null && $( this.getHeadLink().form ).find( '.tsa-required-legend' ).length > 0 ) )
    	{
    		setGlobalLegend = true;
    	}
    	
    	if( this.getHeadLink().linkedTo !== null )
    	{
    	   // Hide local legends...
    	   $( this.getHeadLink().form ).find( '.tsa-required-legend' ).css( { 'display' : 'none' } );
    	}  
        
    	// Create global legend if at least a local legend is found or there are required fields but the local legend is not displayed
    	if( setGlobalLegend )
        {
          /** @type {Element} */  var globalLegend         = tsa.widgets.FormPanel.buildRequiredFieldsLegend();
          /** @type {Element} */  var globalContainer      = document.createElement( 'p' );
          /** @type {string}  */  var globalContainerClass = 'tsa-required-legend-global';
          
          $( globalContainer ).append( globalLegend );
          $( globalContainer ).addClass( globalContainerClass );
          $( this.getHeadLink().form ).find( '.' + globalContainerClass ).remove();
          $( this.getHeadLink().form ).prepend( globalContainer );
        }
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Hide this form. Use the 'show' method to show it again.
     */
    tsa.widgets.FormPanel.prototype.hide = function()
    {
      $( this.getDomNode() ).hide();
      
      // In case it is linked...
      $( this.panelDiv  ).hide();
      $( this.buttonDiv ).hide();
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Show this form, if it was previously hidden. Use the 'hide' method to hide it again.
     */
    tsa.widgets.FormPanel.prototype.show = function()
    {
      $( this.getDomNode() ).show();
      
      // In case it is linked...
      $( this.panelDiv  ).show();
      $( this.buttonDiv ).show();
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Is this form hidden?
     * @return {boolean}
     */
    tsa.widgets.FormPanel.prototype.isHidden = function()
    {
      return ( this.getDomNode() .style.display === 'none' ) &&
             ( this.panelDiv     .style.display === 'none' ) &&
             ( this.buttonDiv    .style.display === 'none' );
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set title
     * @param {string} title Title.
     */
    tsa.widgets.FormPanel.prototype.setTitle = function( title )
    {
      if( TV.typeIsString( typeof title ) )
      {
        /** @type {string} */ var separator = ( ( this.subtitle.length > 0 ) ? ' :: ' : '' );
        
        $( this.titleHeader ).text( ( this.title = $.trim( title ) ) + separator + this.subtitle );
        
        $( this.titleHeader )[ ( ( this.title.length + this.subtitle.length ) > 0 ) ? 'show' : 'hide' ]();
        
        // Add arial-label to the form element for accessibility performance.
        $(this.form).attr("aria-label", title);
      }
      else
      {
        logger.error( 'Invalid title type (' + ( typeof title ) + ').' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set subtitle
     * @param {string} subtitle Subtitle.
     */
    tsa.widgets.FormPanel.prototype.setSubtitle = function( subtitle )
    {
      if( TV.typeIsString( typeof subtitle ) )
      {
        /** @type {string} */ var separator = ( ( this.title.length > 0 ) ? ' :: ' : '' );
        
        $( this.titleHeader ).text( this.title + separator + ( this.subtitle = $.trim( subtitle ) ) );
        
        $( this.titleHeader )[ ( ( this.title.length + this.subtitle.length ) > 0 ) ? 'show' : 'hide' ]();
      }
      else
      {
        logger.error( 'Invalid subtitle type (' + ( typeof subtitle ) + ').' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set description
     * @param {string} description Description.
     */
    tsa.widgets.FormPanel.prototype.setDescription = function( description )
    {
      if( TV.typeIsString( typeof description ) )
      {
        this.description = $.trim( description );
        
        /** @type {boolean} */ var localReqLegendSpanShown = ( this.reqLegendSpan               !== null   ) &&
                                                             ( this.reqLegendSpan.style.display !== 'none' );
        /** @type {boolean} */ var show = ( ( this.description.length > 0 ) || localReqLegendSpanShown );
        
        $( this.descParagraph ).text( this.description );
        
        if( localReqLegendSpanShown )
        {
          if( this.description.length > 0 )
          {
            $( this.descParagraph ).append( document.createElement( 'br' ) );
          }
          
          $( this.reqLegendSpan ).css( { 'display'    : 'inline-block'
                                       , 'margin-top' : ( ( this.description.length > 0 ) ? '1em' : '0em' ) } );
          $( this.descParagraph ).append( this.reqLegendSpan );
        }
        
        $( this.descParagraph )[ show ? 'show' : 'hide' ]();
        $( this.descHR        )[ show ? 'show' : 'hide' ]();
      }
      else
      {
        logger.error( 'Invalid description type (' + ( typeof description ) + ').' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set the submission confirmation message.
     * @param {string} confirmMsg Confirmation message.
     */
    tsa.widgets.FormPanel.prototype.setConfirmMsg = function( confirmMsg )
    {
      if( TV.typeIsString( typeof confirmMsg ) )
      {
        this.confirmMsg = $.trim( confirmMsg );
      }
      else
      {
        logger.error( 'Invalid submission confirmation message (' + ( typeof confirmMsg ) + ').' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set border flag.
     * @param {boolean} flag Border flag.
     */
    tsa.widgets.FormPanel.prototype.setBorder = function( flag )
    {
      if( TV.typeIsBoolean( typeof flag ) )
      {
        this.border = flag;
        
        $( this.bodyDiv ).css( { 'border-style'     : ( this.border ? '' : 'none' )
                               , 'background-image' : 'none' } );
      }
      else
      {
        logger.error( 'Invalid border flag type (' + ( typeof flag ) + ').' );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /** Set width.
     *  @param {number} width */
    
    tsa.widgets.FormPanel.prototype.setWidth = function( width )
    {
      if( TV.typeIsNumber( typeof width ) && !isNaN( width ) )
      {
        if( width > 0 )
        {
          this.width = ( ( width >= 160 ) ? width : 160 );
        }
        else
        {
          this.width = 0; // 100%
        }
        
        $( this.domNode ).css( { 'width' : ( ( this.width > 0 ) ? this.width + 'px' : '' ) } );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set ajax flag.
     * @param {boolean} flag Ajax flag.
     */
    tsa.widgets.FormPanel.prototype.setAjax = function( flag )
    {
      if( TV.typeIsBoolean( typeof flag ) )
      {
        this.ajax = flag;
      }
      else
      {
        logger.error( 'Invalid ajax flag type (' + ( typeof flag ) + ').' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set the cancelable flag: a cancel button will be shown during the AJAX request.
     * @param {boolean} flag Cancelable flag.
     */
    tsa.widgets.FormPanel.prototype.setCancelable = function( flag )
    {
      if( TV.typeIsBoolean( typeof flag ) )
      {
        this.cancelable = flag;
      }
      else
      {
        logger.error( 'Invalid cancelable flag type (' + ( typeof flag ) + ').' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Get the cancelable flag: a cancel button will be shown during the AJAX request.
     * @return {boolean} Cancelable flag.
     */
    tsa.widgets.FormPanel.prototype.isCancelable = function()
    {
      return this.cancelable;
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set the show required field marks flag.
     * @param {boolean} flag The flag.
     */
    tsa.widgets.FormPanel.prototype.setShowReqMarks = function( flag )
    {
      if( TV.typeIsBoolean( typeof flag ) )
      {
        this.showReqMarks = flag;
      }
      else
      {
        logger.error( 'Invalid show-required-field-marks flag type (' + ( typeof flag ) + ').' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set the show required header flag.
     * @param {boolean} flag The flag.
     */
    tsa.widgets.FormPanel.prototype.setShowReqHeader = function( flag )
    {
      if( TV.typeIsBoolean( typeof flag ) )
      {
        this.showReqHeader = flag;
      }
      else
      {
        logger.error( 'Invalid show-required-header flag type (' + ( typeof flag ) + ').' );
      }
    };
    
    /**
     * Set if secure token is required
     * @param {boolean} flag The flag.
     */
    tsa.widgets.FormPanel.prototype.setCsrfRequired = function( flag )
    {
      if( TV.typeIsBoolean( typeof flag ) )
      {
        this.csrfRequired = flag;
      }
      else
      {
        logger.error( 'Invalid secure token flag type (' + ( typeof flag ) + ').' );
      }
    };
    
    //-----------------------------------------------------------------------

    
    /**
     * Append a button.
     * Wrapper for <code>ButtonFactory.buildButton</code>
     * (refer to such method for documentation about the parameters).
     * @param {Object} parameters The parameters.
     * @return {Element} The button that has been added to the form.
     * 
     */
    tsa.widgets.FormPanel.prototype.appendButton = function( parameters )
    {
      /** @type {Element} */ var button = ButtonFactory.buildButton( parameters );
      
      if( button !== null )
      {
        /** @type {Element} */ var buttonContainer = null;
        
        if( $( this.buttonDiv ).children().length > 0 )
        {
          $( this.buttonDiv ).append( ButtonFactory.buildButtonSeparator() );
          buttonContainer = $( this.buttonDiv ).children().get(0);
        }
        else
        {
          buttonContainer = document.createElement( 'p' );
          $( this.buttonDiv ).append( buttonContainer );
          $( this.buttonDiv ).show();
        }
        
        // Appending...
        $( buttonContainer ).append( button );
      }
      
      return button;
    };
    
    //-----------------------------------------------------------------------
    
    /** Builds and adds a header row.
     *  @param {string=} title       Optional. The header title (bold style).
     *  @param {string=} description Optional. The description (regular style).
     *  @return {Element}            Reference to the HTML TR element that has been added. */
    
    tsa.widgets.FormPanel.prototype.appendHeaderRow = function( title, description )
    {
      /** @type {Element} */ var row            = document.createElement( 'tr'   );
      /** @type {Element} */ var cell           = document.createElement( 'td'   );
      /** @type {boolean} */ var titleAvailable = false;
      /** @type {boolean} */ var descAvailable  = false;
      
      if( !TV.typeIsString( typeof title       ) || ( title.length      === 0 ) ) title       = '';
      if( !TV.typeIsString( typeof description ) || ( title.description === 0 ) ) description = '';
      
      titleAvailable = ( ( title       = $.trim( title       ) ).length > 0 );
      descAvailable  = ( ( description = $.trim( description ) ).length > 0 );
      
      cell.colSpan = 2;
      
      $( cell ).css( { 'text-align'     : 'left'
                     , 'vertical-align' : 'top'
                     , 'padding'        : '5px'
                     , 'color'          : '#505050' } );
      
      if( titleAvailable )
      {
        /** @type {Element} */ var titleSpan = document.createElement( 'span' );
        
        $( titleSpan ).css( { 'font-weight' : 'bold' } );
        
        $( titleSpan ).text( title );
        $( cell ).append( titleSpan );
        
        if( descAvailable )
        {
          $( cell ).append( document.createElement( 'br' ) );
        }
      }
      
      if( descAvailable )
      {
        /** @type {Element} */ var descSpan = document.createElement( 'span' );
        
        $( descSpan ).text( description );
        $( cell ).append( descSpan );
      }
      
      $( row  ).append( cell );
      $( this.layoutTableBody ).append( row );
      
      return row;
    };
    
    //-----------------------------------------------------------------------
    
    /** Builds and adds a horizonal ruler
     *  @return {Element} Reference to the HTML TR element that has been added. */
    
    tsa.widgets.FormPanel.prototype.appendHR = function()
    {
      /** @type {Element} */ var row  = document.createElement( 'tr'  );
      /** @type {Element} */ var cell = document.createElement( 'td'  );
      /** @type {Element} */ var div  = document.createElement( 'div' );
      /** @type {Element} */ var hr   = document.createElement( 'hr'  );
      
      cell.colSpan = 2;
      
      $( cell ).css( { 'padding' : '5px 0px' } );
      
      $( div ).addClass( 'ibm-rule' );
      $( div ).append( hr );
      $( cell ).append( div );
      $( row ).append( cell );
      $( this.layoutTableBody ).append( row );
      
      return row;
    };
    
    //-----------------------------------------------------------------------
    
    /** Builds and adds a blank spacer
     *  @return {Element} Reference to the HTML TR element that has been added. */
    
    tsa.widgets.FormPanel.prototype.appendSpacer = function()
    {
      /** @type {Element} */ var row  = document.createElement( 'tr'  );
      /** @type {Element} */ var cell = document.createElement( 'td'  );
      /** @type {Element} */ var div  = document.createElement( 'div' );
      
      cell.colSpan = 2;
      
      $( cell ).css( { 'padding' : '5px 0px' } );
      $( div  ).css( { 'height'  : '1em'     } );
      
      $( cell ).append( div );
      $( row ).append( cell );
      $( this.layoutTableBody ).append( row );
      
      return row;
    };
    
    //-----------------------------------------------------------------------
    
    /** Builds and adds a joint row (single cell).
     *  @param {Node} node The DOM Node.
     *  @return {Element}  Reference to the HTML TR element that has been added. */
    
    tsa.widgets.FormPanel.prototype.appendJointRow = function( node )
    {
      /** @type {Element} */ var row  = document.createElement( 'tr' );
      /** @type {Element} */ var cell = document.createElement( 'td' );
      
      if( TV.typeIsObject( typeof node ) && ( node !== null ) && !TV.typeIsUndefined( node.nodeType ) )
      {
        cell.colSpan = 2;
        
        $( cell ).css( { 'text-align'     : 'left'
                       , 'vertical-align' : 'top'
                       , 'padding'        : '5px'
                       , 'color'          : '#000000' } );
        
        cell.appendChild( node );
        row.appendChild( cell );
      }
      
      this.layoutTableBody.appendChild( row );
      
      return row;
    };
    
    //-----------------------------------------------------------------------
    
    /** Builds and adds a field row.
     *  Parameters:
     *      label    : The text to be used to build the label (optional; default: no label).
     *      required : Is the field required? (optional; default: false).
     *      aRequire : Enable aria-required (optional, default : true ).
     *      element  : The HTML Input Element (or its parent).
     *      hint     : Help text to be displayed behind the element (optional; default: no hint).
     *  
     *  @param {{ label    : string
     *          , required : boolean
     *          , aRequire : boolean
     *          , element  : Element
     *          , hint     : string}} parameters
     *  @return {Element}             Reference to the HTML TR element that has been added. */
    
    tsa.widgets.FormPanel.prototype.appendFieldRow = function( parameters )
    {
      if( !TV.typeIsObject( typeof parameters ) || ( parameters === null ) ) parameters = {};
      
      if( !TV.typeIsString  ( typeof parameters.label    ) ) parameters.label    = '';
      if( !TV.typeIsBoolean ( typeof parameters.required ) ) parameters.required = false;
      if( !TV.typeIsBoolean ( typeof parameters.aRequire ) ) parameters.aRequire = true;
      if( !TV.typeIsObject  ( typeof parameters.element  ) ) parameters.element  = null;
      if( !TV.typeIsString  ( typeof parameters.hint     ) ) parameters.hint     = '';
      
      /** @type {string}  */ var labelSuffix = ':';
      /** @type {Element} */ var row         = document.createElement( 'tr' );
      /** @type {Element} */ var cellL       = document.createElement( 'td' );
      /** @type {Element} */ var cellR       = ( ( parameters.element !== null ) ? document.createElement( 'td' ) : null );
      /** @type {Element} */ var label       = document.createElement( 'label' );
      /** @type {Element} */ var labelChild  = document.createElement( 'em' );
      
      if( ( parameters.label = $.trim( parameters.label ) ).length === 0 )
      {
        parameters.label = tsa.NBSP;
        labelSuffix = '';
      }
      else if( parameters.label.charAt( parameters.label.length - 1 ) === ':' )
      {
        labelSuffix = '';
      }
      
      $( cellL ).css( { 'text-align'     : 'left'
                      , 'vertical-align' : 'top'
                      , 'padding'        : '5px' } );
      
      $( labelChild ).text( parameters.label + labelSuffix );
      $( label ).append( labelChild );
      
      if( parameters.required )
      {
    	  if( this.showReqMarks )
          {
          /** @type {Element} */ var requiredSpan = document.createElement( 'span' );
          
          if(parameters.aRequire)
          {
            $( requiredSpan ).addClass( 'ibm-required' );
          }
          $( requiredSpan ).text( ' *' );
          $( label ).append( requiredSpan );
          
          
          if( this.showReqHeader )
          {
        	  if( ( this.reqLegendSpan === null ) &&
        			  TV.typeIsObject( typeof tsa.i18n ) && ( tsa.i18n !== null ) &&
        			  TV.typeIsString( typeof tsa.i18n.labelRequiredLegend ) )
        	  {
        		  this.reqLegendSpan = tsa.widgets.FormPanel.buildRequiredFieldsLegend();
        		  $( this.reqLegendSpan ).css( { 'margin-top' : '1.0em'
        			  						   , 'display'    : ( this.linkedBy !== null ) ? 'none' : '' } );
            
        		  // Update description (the legend will be placed there):
        		  this.setDescription( this.description );
            
        		  // Define if a global legend should replace local legends...
        		  this.useGlobalRequiredFieldsLegendIfNeeded();
        	  }
          }
          
          else
          {
        	  // Define if a global legend should replace local legends...
    		  this.useGlobalRequiredFieldsLegendIfNeeded();
          }
        }
      }
      
      $( cellL ).append( label );
      $( row ).append( cellL );
      
      if( TV.typeIsObject( typeof parameters.element ) && ( parameters.element !== null ) &&
          TV.typeIsString( typeof parameters.element.tagName ) )
      {
        /** @type {string} */ var tagName = parameters.element.tagName.toLowerCase();
        
        $( parameters.element ).css( { 'margin' : '0px auto 0px 0px' } );
        
        $( cellR ).css( { 'text-align'     : 'left'
                        , 'vertical-align' : 'bottom'
                        , 'padding'        : '5px' } );
        
        if( parameters.required )
        {
          /** @type {jQuery} */ var $namedElements = $( parameters.element ).find( '[name]' ).addBack( '[name]' );
          
          if( $namedElements.length > 0 )
          {
            /** @type {string} */ var name = $namedElements.attr( 'name' );
            this.addValidationRule( name, 'required' );
          }
          if(parameters.aRequire)
          {
            $( parameters.element ).attr( 'aria-required', 'true' );
          }
        }
        
        // Associating the label with the element (the element must have an ID):
        if( ( ( tagName === 'select'   ) ||
              ( tagName === 'input'    ) ||
              ( tagName === 'textarea' ) ) &&
            TV.typeIsString( typeof parameters.element.id ) )
        {
          $( label ).attr( 'for', parameters.element.id );
        }
        
        // Appending the element:
        $( cellR ).append( parameters.element );
        
        // Hint:
        if( ( parameters.hint = $.trim( parameters.hint ) ).length > 0 )
        {
          /** @type {Element} */ var hintDiv = document.createElement( 'div' );
          
          $( hintDiv ).css( { 'color'     : '#666'
                            , 'padding'   : '2px'
                            , 'font-size' : '0.9em' } );
          $( hintDiv ).text( parameters.hint );
          $( cellR ).append( hintDiv );
          
          if( TV.typeIsString( typeof parameters.element.id ) )
          {
            /** @type {string} */ var idForHintDiv = 'hint_' + parameters.element.id;
            
            $( hintDiv ).attr( 'id', idForHintDiv );
            $( parameters.element ).attr( 'aria-describedby', idForHintDiv );
          }
        }
        
        $( row ).append( cellR );
      }
      else
      {
        $( cellL ).attr( 'colspan', '2' );
      }
      
      this.layoutTableBody.appendChild( row );
      
      return row;
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Add or change the value of a hidden input.
     * @param {string} name  Name.
     * @param {string} value Value; use null to remove the hidden input.
     */
    tsa.widgets.FormPanel.prototype.putHiddenInput = function( name, value )
    {
      if( TV.typeIsString( typeof name ) && ( ( name = $.trim( name ) ).length > 0 ) )
      {
        if( TV.typeIsUndefined( typeof value ) || ( TV.typeIsObject( typeof value ) && ( value === null ) ) )
        {
          if( !TV.typeIsUndefined( typeof this.hiddenInputMap[ name ] ) )
          {
            delete this.hiddenInputMap[ name ];
          }
        }
        else
        {
          this.hiddenInputMap[ name ] = ( TV.typeIsString( typeof value ) ? value : value + '' );
        }
      }
      else
      {
        logger.error( 'The name of the hidden input must be a non-empty string.' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Get the value of a hidden input, or null if the hidden input having the
     * given name does not exist.
     * @param  {string}  name  Name.
     * @return {?string} value Value, as string, or use null if the hidden input
     *                         having the given name does not exist.
     */
    tsa.widgets.FormPanel.prototype.getHiddenInput = function( name )
    {
      /** @type {?string} */ var value = null;
      
      if( TV.typeIsString( typeof name ) && ( ( name = $.trim( name ) ).length > 0 ) )
      {
        if( !TV.typeIsUndefined( typeof this.hiddenInputMap[ name ] ) )
        {
          value = this.hiddenInputMap[ name ];
        }
      }
      else
      {
        logger.error( 'The name of the hidden input must be a non-empty string.' );
      }
      
      return value;
    };
    
    //-----------------------------------------------------------------------
    /**
     * Is field disabled?
     * @return {boolean}
     */
    tsa.widgets.FormPanel.prototype.isFieldDisabled = function( name )
    {
      /** @type {boolean} */ var disabledFound = false;
      
      disabledFound = ( $( this.getHeadLink().form ).find( '[name=\'' + name + '\']' ).filter( ':disabled' ).length > 0 );
      
      return disabledFound;
    };
    
    //-----------------------------------------------------------------------
    
    /** Get all enabled fields' name-value pairs (as an object: {name1:value1,name2:value2,...,nameN:valueN}).
     *  WARNING -- The following fields will NOT be returned:
     *              a) Disabled fields.
     *              b) Any field if the whole form is hidden.
     *  @return {Object} */
    
    tsa.widgets.FormPanel.prototype.getValues = function()
    {
      /* Anonymous parameters (to be used only on secondary, recursive calls):
       * arguments[0] : Recursion level (limit is 24). Initial: 0.
       * arguments[1] : HTML Element to be processed now. Initial: this.panelDiv.
       * arguments[2] : The "values" map (Object). Initial: empty map (Object). 
       */
      
      /** @type {Object}  */ var values           = ( TV.typeIsObject( typeof arguments[2] ) ? /** @type {Object} */ ( arguments[2] ) : {} );
      /** @type {string}  */ var unrecognizedProp = '__unrecognized__';
      /** @type {number}  */ var recursiveLevel   = ( TV.typeIsNumber( typeof arguments[0] ) ? /** @type {number} */ ( arguments[0] ) : 0 );
      /** @type {Element} */ var element          = null;
      
      var putValue = function( name, value )
      {
        if( TV.typeIsString( typeof name ) )
        {
          if( TV.typeIsString( typeof value ) )
          {
            value = $.trim( value );
          }
          
          if( TV.typeIsUndefined( typeof values[ name ] ) )
          {
            values[ name ] = ( value );
          }
          else if( TV.isInstanceOf( values[ name ], Array ) )
          {
            values[ name ].push( value );
          }
          else
          {
            values[ name ] = [ values[ name ], value ];
          }
        }
      };
      var putValuePwd = function( name, value )
      {
        if( TV.typeIsString( typeof name ) )
        {
          if( TV.typeIsUndefined( typeof values[ name ] ) )
          {
            values[ name ] = ( value );
          }
          else
          {
            values[ name ] = [ values[ name ], value ];
          }
        }
      };
      
      if( ( recursiveLevel < 24 ) && !this.isHidden() )
      {
        if( recursiveLevel === 0 ) element = this.panelDiv;
        else                       element = ( TV.typeIsObject( typeof arguments[1] ) ? /** @type {Element} */ ( arguments[1] ) : null );
        
        if( ( element !== null ) && !TV.typeIsUndefined( typeof element.childNodes ) )
        {
          for( var i=0; i<element.childNodes.length; i++ )
          {
            /** @type {Node} */ var child = element.childNodes[i];
            
            if( !TV.typeIsUndefined( typeof child.tagName ) )
            {
              if( TV.typeIsString( typeof child.name ) && ( child.name.length > 0 ) && !$( child ).is(':disabled') )
              {
                switch( child.tagName.toLowerCase() )
                {
                  case 'input'    : if( ( child.type === 'text'     ) ||
                                        ( child.type === 'hidden'   ) )
                                    {
                                      putValue( child.name, child.value );
                                    }
                                    else if (( child.type === 'password' ))
                                    {
                                      // Password value will not be trimmed, user will decide if use whitespace
                                      putValuePwd( child.name, child.value );
                                      
                                    }
                                    else if( ( ( child.type === 'radio'    ) ||
                                               ( child.type === 'checkbox' ) ) && child.checked )
                                    {
                                      putValue( child.name, child.value );
                                    }
                                    else if( ( child.type === 'file' ) )
                                    {
                                      this.setEnctypeAttribute( 'multipart/form-data' );
                                      putValue( child.name, child );
                                    }
                                    break;
                  case 'select'   : if( child.options.length > 0 )
                                    {
                                      putValue( child.name, child.options[ child.selectedIndex ].value );
                                    }
                                    break;
                  case 'textarea' : putValue( child.name, child.value );
                                    break;
                  default         : // The way to obtain the value from this element is unknown:
                                    if( !TV.typeIsObject( typeof values[ unrecognizedProp ] ) )
                                    {
                                      values[ unrecognizedProp ] = {};
                                    }
                                    values[ unrecognizedProp ][ child.name ] = child.tagName.toLowerCase();
                }
              }
              
              if( child.childNodes.length > 0 )
              {
                this.getValues( recursiveLevel + 1, child, values );
              }
            }
          }
        }
      }
      
      if( recursiveLevel === 0 )
      {
        /** @type {string} */ var msg     = 'The following parameters have been retrieved from the form:';
        /** @type {string} */ var msg2    = 'The following parameters could not be retrieved from the form:';
        /** @type {number} */ var counter = 0;
        
        if( this.linkedTo !== null )
        {
          $.extend( values, this.linkedTo.getValues() );
        }
        
        for( /** @type {string} */ hiddenInputName in this.hiddenInputMap )
        {
          putValue( hiddenInputName, this.hiddenInputMap[ hiddenInputName ] );
        }
        
        if( TV.typeIsObject( typeof values[ unrecognizedProp ] ) )
        {
          for( /** @type {string} */ var unrecognizedPropName in values[ unrecognizedProp ] )
          {
            if( TV.typeIsUndefined( typeof values[ unrecognizedPropName ] ) )
            {
              /** @type {string} */ var unrecognizedPropTag = values[ unrecognizedProp ][ unrecognizedPropName ];
              
              msg2 += '\n' + ( ( ++counter < 10 ) ? '0' : '' ) + counter + '.- "' + unrecognizedPropName + '" (' + unrecognizedPropTag + ').'
            }
          }
          
          logger.info( msg2 );
          delete values[ unrecognizedProp ];
        }
        
        if( this.linkedBy === null )
        {
          counter = 0;
          
          for( /** @type {string} */ var name in values )
          {
            msg += '\n  ' + ( ( ++counter < 10 ) ? '0' : '' ) + counter + '.- "' + name + '".';
          }
          
          logger.info( msg );
        }
      }
      
      return values;
    };
    
    //-----------------------------------------------------------------------
    
    /** Get a value to the form's "method" attribute.
     *  @return {string} */
    
    tsa.widgets.FormPanel.prototype.getMethodAttribute = function()
    {
      return $( this.form ).attr( 'method' );
    };
    
    //-----------------------------------------------------------------------
    
    /** Get a value to the form's "enctype" attribute.
     *  @return {string} */
    
    tsa.widgets.FormPanel.prototype.getEnctypeAttribute = function()
    {
      return $( this.form ).attr( 'enctype' );
    };
    
    //-----------------------------------------------------------------------
    
    /** Get a value to the form's "action" attribute.
     *  @return {string} */
    
    tsa.widgets.FormPanel.prototype.getActionAttribute = function()
    {
      return $( this.form ).attr( 'action' );
    };
    
    //-----------------------------------------------------------------------
    
    /** Set a value to the form's "method" attribute.
     *  @param {string} value */
    
    tsa.widgets.FormPanel.prototype.setMethodAttribute = function( value )
    {
      $( this.form ).attr( 'method', value + '' );
    };
    
    //-----------------------------------------------------------------------
    
    /** Set a value to the form's "enctype" attribute.
     *  @param {string} value */
    
    tsa.widgets.FormPanel.prototype.setEnctypeAttribute = function( value )
    {
      $( this.form ).attr( 'enctype', value + '' );
    };
    
    //-----------------------------------------------------------------------
    
    /** Set a value to the form's "action" attribute.
     *  Use "controller_or_handler/method" for both ajax and non-ajax requests.
     *  @param {string} value */
    
    tsa.widgets.FormPanel.prototype.setActionAttribute = function( value )
    {
      $( this.form ).attr( 'action', value + '' );
    };
    
    //-----------------------------------------------------------------------
    /**
     * Common validation rule: required field.
     * @param  {?string} value Value.
     * @return {?string}       The reason this value is not valid, or null if it's valid.
     */
    tsa.widgets.FormPanel.prototype.validationRuleRequired = function( value )
    {
      return ( ( ( TV.typeIsString( typeof value ) ) && ( value.length > 0 ) ) ? null : tsa.i18n.labelRequiredFieldIsBlank );
    };
    
    //-----------------------------------------------------------------------
    /**
     * Common validation rule: ipv4 field.
     * @param  {?string} value Value.
     * @return {?string}       The reason this value is not valid, or null if it's valid.
     */
    tsa.widgets.FormPanel.validationRuleIpv4 = function( value )
    {
      return ( VT.isValidIPv4( value ) ? null : tsa.i18n.labelInvalidIPv4 );
    };
    
    //-----------------------------------------------------------------------
    /**
     * Add validation rule for a given field.
     * Note: password confirmation fields are automatically validated, they don't require
     *       validation rules. The only requirement is that the name of the confirmation
     *       field must have the '_confirm' suffix after the name of the base field.
     * @param {string} fieldName The name of the field (HTTP parameter name).
     * @param {(string|function(?string):?string} rule The rule; can be either a name of a common rule,
     *                                                 or a custom function rule. Common rules:
     *                                                 "required" : field is required (value can't be blank).
     *                                                 "ipv4"     : IPv4 required.
     */
    tsa.widgets.FormPanel.prototype.addValidationRule = function( name, rule )
    {
      if( TV.typeIsString( typeof name ) && ( ( name = $.trim( name ) ).length > 0 ) )
      {
        /** @type {function(string):boolean} */ var validationFunction = null;
        
        if( TV.typeIsString( typeof rule ) && ( rule.length > 0 ) )
        {
          /** @type {string} */ var methodName = 'validationRule' + rule.charAt(0).toUpperCase() + rule.substring(1);
          
          if( TV.typeIsFunction( typeof tsa.widgets.FormPanel[ methodName ] ) )
          {
            validationFunction = tsa.widgets.FormPanel[ methodName ];
          }
          else if( TV.typeIsFunction(typeof this[ methodName ]))
          {
            validationFunction = this[ methodName ];
          }
        }
        else if( TV.typeIsFunction( typeof rule ) && ( rule !== null ) )
        {
          validationFunction = rule;
        }
        else
        {
          logger.error( 'Validation rule can\'t be added: invalid rule.' );
        }
        
        if( validationFunction !== null )
        {
          if( TV.typeIsUndefined( typeof this.validationRules[ name ] ) )
          {
            this.validationRules[ name ] = [];
          }
          
          this.validationRules[ name ].push( validationFunction );
        }
      }
      else
      {
        logger.error( 'Validation rule can\'t be added: invalid field name.' );
      }
    };
    
    //-----------------------------------------------------------------------
    /**
     * Validate and display validation results.
     * @return {boolean}
     */
    tsa.widgets.FormPanel.prototype.validate = function()
    {
      /** @type {boolean} */ var linkedIsValid = ( ( this.linkedTo !== null ) ? this.linkedTo.validate() : true );
      /** @type {boolean} */ var thisIsValid   = true;
      
      if( !this.isHidden() )
      {
        /** @type {Object.<string,Array>} */ var actualValidationRules = null;
        /** @type {jQuery}                */ var $pwdInputs            = $( this.panelDiv ).find( '[type=\'password\']' ).not( ':disabled' );
        /** @type {string}                */ var pwdConfirmSuffix      = '_confirm';
        /** @type {string}                */ var pwdConfirmRegExp      = new RegExp( pwdConfirmSuffix + '$' );
        /** @type {Object}                */ var httpRequestParameters = this.getValues();
        
        // Remove previous validation error messages:
        $( this.panelDiv ).find( '.tsa-form-validation-error-msg'   ).remove();
        $( this.panelDiv ).find( '.tsa-form-validation-error-label' ).css( { 'color'   : '' } );
        $( this.panelDiv ).find( '.tsa-form-validation-error-field' ).css( { 'outline' : '' } );
        $( this.panelDiv ).find( '.tsa-form-validation-error-label' ).removeClass( 'tsa-form-validation-error-label' );
        $( this.panelDiv ).find( '.tsa-form-validation-error-field' ).removeClass( 'tsa-form-validation-error-field' );
        
        // Password confirmation fields are automatically validated; adding the rules...
        $pwdInputs.each( $.proxy( function( pwdInputIndex, pwdInput )
        {
          /** @type {string}  */ var pwdInputName   = pwdInput.name;
          /** @type {boolean} */ var isConfirmation = pwdInputName.match( pwdConfirmRegExp );
          
          if( isConfirmation )
          {
            /** @type {string}  */ var basePwdInputName  = pwdInput.name.substring( 0, pwdInput.name.lastIndexOf( pwdConfirmSuffix ) );
            /** @type {?string} */ var pwdInputValue     = ( TV.typeIsString( typeof httpRequestParameters[ pwdInputName     ] ) ? httpRequestParameters[ pwdInputName     ] : null );
            /** @type {?string} */ var basePwdInputValue = ( TV.typeIsString( typeof httpRequestParameters[ basePwdInputName ] ) ? httpRequestParameters[ basePwdInputName ] : null );
            
            if( TV.typeIsString( typeof basePwdInputValue ) &&
                TV.typeIsString( typeof pwdInputValue     ) &&
                ( basePwdInputValue !== pwdInputValue ) )
            {
              /** @type {Array.<function(?string):?string>} */ var rules = null;
              
              // Make a deep copy of validation rules...
              if( actualValidationRules === null )
              {
                actualValidationRules = ( ( this.validationRules !== null ) ? $.extend( true, {}, this.validationRules ) : {} );
              }
              
              if( TV.typeIsUndefined( typeof actualValidationRules[ pwdInputName ] ) )
              {
                actualValidationRules[ pwdInputName ] = [];
              }
              
              rules = actualValidationRules[ pwdInputName ];
              rules.push( function( value ) { return tsa.i18n.labelConfirmationPasswordMustMatch } );
            }
          }
        }
        , this ) );
        
        // No automatic rules added? use the original rules...
        if( actualValidationRules === null )
        {
          actualValidationRules = this.validationRules;
        }
        
        // Are there any validation rules?
        if( actualValidationRules !== null )
        {
          // Iterating the rules...
          for( /** @type {string} */ var name in actualValidationRules )
          {
            // Disabled fields are ignored...
            if( !this.isFieldDisabled( name ) )
            {
              /** @type {?}       */ var raw    = ( TV.typeIsUndefined( typeof httpRequestParameters[ name ] ) ? null : httpRequestParameters[ name ] );
              /** @type {?string} */ var value  = null;
              /** @type {boolean} */ var valid  = true;
              /** @type {?string} */ var reason = null;
              
              if( TV.typeIsString( typeof raw ) )
              {
                value = $.trim( raw );
              }
              else if( TV.typeIsObject( typeof raw ) && ( raw !== null ) )
              {
                if( TV.isInstanceOf( raw, Array ) )
                {
                  /** @type {Array.<string>} */ var multivalue = [];
                  $.each( raw, function( entryIndex, entry ) { multivalue.push( $.trim( '' + entry ) ) } );
                  value = '' + multivalue;
                }
                else
                {
                  value = $.trim( '' + ( TV.typeIsUndefined( typeof raw.value ) ? raw : raw.value ) );
                }
              }
              
              $.each( actualValidationRules[ name ], $.proxy( function( i, validationFunction )
              {
                if( validationFunction === tsa.widgets.FormPanel.prototype.validationRuleRequired )
                {
                  valid = ( ( reason = validationFunction.call( null, value ) ) === null );
                }
                else
                {
                  valid = ( value        === null ) ||
                          ( value.length === 0    ) || ( ( reason = validationFunction.call( null, value ) ) === null );
                }
                
                return valid;
              }
              , this ) );
              
              if( !valid )
              {
                logger.warn( 'Invalid field detected. Name: "' + name + '". Reason: "' + reason + '".' );
                this.showFieldError( name, reason );
                thisIsValid = false;
              }
            }
          }
        }
      }
      
      return ( thisIsValid && linkedIsValid );
    };
    //-----------------------------------------------------------------------
    /**
     * Remove validation
     * @return {boolean}
     */
    tsa.widgets.FormPanel.prototype.removeValidation = function(name)
    {
      delete this.validationRules[name];
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Show error message for a specific field. The field is searched in the linked form as well.
     * @param {string} name Field name.
     * @param {string} msg  Error message.
     */
    tsa.widgets.FormPanel.prototype.showFieldError = function( name, msg )
    {
      if( TV.typeIsString( typeof name ) && TV.typeIsString( typeof msg ) )
      {
        try
        {
          /** @type {Element} */ var errorSpan = $( '<span class="tsa-form-validation-error-msg" style="color: red"></span>' ).get(0);
          /** @type {Element} */ var errorEm   = $( '<em></em>' ).get(0);
          /** @type {jQuery}  */ var $fields   = $( this.getHeadLink().form ).find( '[name=\'' + name + '\']' );
          
          if( TV.typeIsObject( typeof $fields ) && ( $fields !== null ) && ( $fields.length > 0 ) )
          {
            /** @type {Element} */ var fieldCell = $( $fields.get(0) ).closest( 'td' ).get(0);
            /** @type {Element} */ var labelCell = $( fieldCell ).closest( 'tr' ).find( 'td' ).get(0);
            /** @type {Element} */ var label     = $( labelCell ).find( 'label' ).get(0);
            
            $fields.addClass( 'tsa-form-validation-error-field' );
            $fields.css( { 'outline' : '1px solid red' } );
            
            $( label ).addClass( 'tsa-form-validation-error-label' );
            $( label ).css( { 'color' : 'red' } );
            
            $( errorEm ).text( msg );
            $( errorSpan ).append( errorEm, document.createElement( 'br' ) );
            $( fieldCell ).prepend( errorSpan );
            tsa.clearSysMessages();
            tsa.printSysMessage( new tsa.cmd.SystemMessage( { level   : 'error'
                                                            , message : tsa.i18n.labelFormHasErrors } ) );
          }
          else
          {
            logger.error( 'Could not show field error: field not found (name: "' + name + '", message: "' + msg + '").' );
          }
        }
        catch( error )
        {
          logger.error( 'Could not show field error: unexpected error (name: "' + name + '", message: "' + msg + '").' );
          logger.error( error );
          // Ignore. We did what we can.
        }
      }
      else
      {
        logger.error( 'Could not show field error: invalid arguments.' );
      }
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Ask the user for submission confirmation (only if a submission confirmation message is set).
     * Returns true if the user confirms the submission of the form, or if confirmation is not required
     * because a confirmation message was not set.
     * @return {boolean}
     */
    tsa.widgets.FormPanel.prototype.confirm = function()
    {
      /** @type {boolean} */ var result = false;
      
      if( TV.typeIsString( typeof this.confirmMsg ) && ( this.confirmMsg.length > 0 ) )
      {
        if( result = window.confirm( this.confirmMsg ) )
        {
          logger.info( 'The user has confirmed the form submission; it will proceed.' );
        }
        else
        {
          logger.info( 'The user has cancelled the form submission.' );
        }
      }
      else
      {
        result = true;
        logger.info( 'Form submission doesn\'t require confirmation from the user; it will proceed.' );
      }
      
      return result;
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Submit the form.
     */
    tsa.widgets.FormPanel.prototype.submit = function()
    {
      if( this.validate() )
      {
        if( this.confirm() )
        {
          if( this.linkedBy === null )
          {
            if( this.ajax === true )
            {
              /** @type {string}  */ var action                = this.getActionAttribute();
              /** @type {Object}  */ var httpRequestParameters = this.getValues();
              /** @type {boolean} */ var forceAjax             = !TV.typeIsUndefined( typeof httpRequestParameters[ 'ajax' ] );
              
              if( !TV.typeIsString( typeof action ) ||
                  ( ( action.indexOf( '/ajax'   ) < 0 ) &&
                    ( action.indexOf( '/get'    ) < 0 ) &&
                    ( action.indexOf( '/put'    ) < 0 ) &&
                    ( action.indexOf( '/post'   ) < 0 ) &&
                    ( action.indexOf( '/delete' ) < 0 ) && !forceAjax ) )
              {
                logger.warn( 'Ajax form has an action that the server will not process as ajax (' + action + ').' );
              }
              
              logger.info( 'Ajax form will execute a command (action: "' + action + '").' );
              
              /** @type {tsa.cmd.Command} */ var cmd = null;
              
              cmd = new Command( { action     : action
                                 , ajax       : true
                                 , cancelable : this.isCancelable()
                                 , parameters : httpRequestParameters
                                 , secureToken : this.csrfRequired } );
              
              // Prepare event forwarding...
              $( cmd ).on( 'executionSucceeded' , $.proxy( function( event ) { $( this ).trigger( { type : 'executionSucceeded' , response : event.response } ); }, this ) );
              $( cmd ).on( 'executionFailed'    , $.proxy( function( event ) { $( this ).trigger( { type : 'executionFailed'    , response : event.response } ); }, this ) );
              $( cmd ).on( 'executionFinished'  , $.proxy( function( event ) { $( this ).trigger( { type : 'executionFinished'  , response : event.response } ); }, this ) );
              
              // Executing...
              cmd.execute();
            }
            else
            {
              window.setTimeout( $.proxy( function() { this.form.submit(); }, this ), 100 );
              tsa.blockUI();
            }
          }
          else
          {
            logger.error( 'Cannot submit a FormPanel linked by another. Submit the first FormPanel in the chain.' );
          }
        }
        else
        {
          logger.warn( 'Form will not be submitted: confirmation denied.' );
        }
      }
      else
      {
        logger.warn( 'Form can\'t be submitted: it contains invalid data.' );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Build the Required Fields legend HTML Element.
     * @return {Element}
     */
    tsa.widgets.FormPanel.buildRequiredFieldsLegend = function()
    {
      /** @type {Array.<string>} */ var legendChunks = tsa.i18n.labelRequiredLegend.split( '*' );
      /** @type {Element}        */ var legendSpan   = document.createElement( 'span' );
      
      $( legendSpan ).addClass( 'tsa-required-legend' );
      
      if( legendChunks.length === 2 )
      {
        $( legendSpan ).append( document.createTextNode( legendChunks[0] )
                              , $( '<span class="ibm-required"> * </span>' ).get(0)
                              , document.createTextNode( legendChunks[1] ) );
      }
      else
      {
        $( legendSpan ).text( tsa.i18n.labelRequiredLegend );
      }
      
      return legendSpan;
    };
    
    //-------------------------------------------------------------------------
    
    return tsa.widgets.FormPanel;
  }
);