/*
 * IBM Technical Support Appliance Version 2 Release 6 Modification 0
 *
 * (C) Copyright IBM Corp. 2019
 */

//-----------------------------------------------------------------------------

define(
  [ 'jquery'
  , 'tsa/TypeVerificator'
  , 'tsa/cmd/Command'
  ],
  
  function( $
          , TV
          , Command )
  {
    //-------------------------------------------------------------------------
    
    var logger = tsa.getLogger( 'TablePanel' );
    
    //-------------------------------------------------------------------------
    
    /** @constructor
     *  Parameters:
     *    title         The title (optional).
     *    subtitle      The subtitle (optional).
     *    description   The description, shown below the title (optional; default: none).
     *    summary       The summary (table attribute)
     *    emptyMsg      The message the table will show if empty. Optional.
     *    width         The width, in pixels (optional; default: 100%).
     *    ajaxAction    The ajax action to load entries (by using a Command).
     *                  Optional. Use appendRow to manually append rows to this table.
     *                  The following parameters will be sent to the server:
     *                    page            Number of page. First page is 1.
     *                    entriesPerPage  Number of entries per page. Default: 20.
     *                    orderBy         Index or ID (if available) of the column that will be
     *                                    sorted.
     *                    ascending       <code>true</code> will be used for ascending, while
     *                                    <code>false</code> for descending.
     *                  The response from the server must contain the following properties:
     *                    entries         An array of entry data (objects).
     *                    total           Total number of entries available in the server.
     *                  Each entry data object conatined in the server's response will be
     *                  processed by the entryHandler property.
     *    entryHandler  Function that processes an object containing entry data, as received
     *                  from the server, an that must return an array of either HTML Elements
     *                  or strings to be added to this table (each element of the array will
     *                  be put in a cell of a row), by using the appendRow method; refer to such
     *                  method for information about the supported array element types.
     *    pageSize      The entries per page (only honored if data is retrieved using AJAX).
     *    columns       Array of columns' data (an object per column):
     *                    id       : The ID of the column. Optional. The column index will be used
     *                               if an ID is not provided, although providing one is recommended.
     *                    label    : The label of the column. Optional. If no labels are
     *                               provided, the table header will be hidden.
     *                    width    : The width of the column (css notation, example: '15%').
     *                               Optional. Default: auto.
     *                    sortable : Is the column sortable? Optional. Default: true.
     *                    initSort : If provided, use either 'asc' or 'desc', meaning this is
     *                               the column that is initially sorted. Optional.     *                               
     *    csrfRequired  Generate and insert CSRF token to the ajaxAction(optional, default true);
     *  
     *  @param { { title        : (string|undefined)
     *           , subtitle     : (string|undefined)
     *           , description  : (string|undefined)
     *           , summary      : (string|undefined)
     *           , emptyMsg     : (string|undefined)
     *           , width        : (number|undefined)
     *           , ajaxAction   : (string|undefined)
     *           , entryHandler : (Function|undefined)
     *           , pageSize     : (number|undefined)
     *           , loadPage     : (Function|undefined)
     *           , columns      : Array.<{ id       : string
     *                                   , label    : string
     *                                   , width    : string
     *                                   , sortable : boolean
     *                                   , initSort : string }> } } parameters The parameters object. */
    
    tsa.widgets.TablePanel = function( parameters )
    {
      if( !TV.typeIsObject( typeof parameters ) || ( parameters === null ) ) parameters = {};
      
      if( TV.typeIsString   ( typeof parameters.title        ) ) this.setTitle       ( parameters.title        );
      if( TV.typeIsString   ( typeof parameters.subtitle     ) ) this.setSubtitle    ( parameters.subtitle     );
      if( TV.typeIsString   ( typeof parameters.description  ) ) this.setDescription ( parameters.description  );
      if( TV.typeIsString   ( typeof parameters.summary      ) ) this.setSummary     ( parameters.summary      );
      if( TV.typeIsString   ( typeof parameters.emptyMsg     ) ) this.setEmptyMsg    ( parameters.emptyMsg     );
      if( TV.typeIsNumber   ( typeof parameters.width        ) ) this.setWidth       ( parameters.width        );
      if( TV.typeIsNumber   ( typeof parameters.pageSize     ) ) this.setPageSize    ( parameters.pageSize     );
      if( TV.typeIsString   ( typeof parameters.ajaxAction   ) ) this.ajaxAction     = parameters.ajaxAction    ;
      if( TV.typeIsFunction ( typeof parameters.entryHandler ) ) this.entryHandler   = parameters.entryHandler  ;
      if( TV.typeIsObject   ( typeof parameters.columns      ) ) this.columns        = parameters.columns       ;
      if( TV.typeIsBoolean  ( typeof parameters.csrfRequired ) ) this.csrfRequired   = parameters.csrfRequired  ;
      
      this.initUIElement();
      
      if( this.isAjaxDriven() )
      {
        this.goToPage( 1 );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /** Main HTML Element (DIV).
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TablePanel.prototype.domNode = null;
    
    /** The HTML Table Element.
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TablePanel.prototype.table = null;
    
    /** The header (HTML CAPTION Element).
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TablePanel.prototype.headerCaption = null;
    
    /** The table header (HTML THEAD Element).
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TablePanel.prototype.tableHeader = null;
    
    /** The layout table body (HTML TBODY Element).
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TablePanel.prototype.tableBody = null;
    
    /** The navigation section.
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TablePanel.prototype.navSection = null;
    
    /** The empty message element.
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TablePanel.prototype.emptyMsgElement = null;
    
    /** The title.
     *  @private
     *  @type {string}
     */
    tsa.widgets.TablePanel.prototype.title = '';
    
    /** The subtitle.
     *  @private
     *  @type {string}
     */
    tsa.widgets.TablePanel.prototype.subtitle = '';
    
    /** The description.
     *  @private
     *  @type {string}
     */
    tsa.widgets.TablePanel.prototype.description = '';
    
    /** The summary.
     *  @private
     *  @type {string}
     */
    tsa.widgets.TablePanel.prototype.summary = '';
    
    /** The empty message.
     *  @private
     *  @type {string}
     */
    tsa.widgets.TablePanel.prototype.emptyMsg = tsa.i18n.labelTablePanelNoEntries;
    
    /** The width.
     *  @private
     *  @type {number}
     */
    tsa.widgets.TablePanel.prototype.width = 0;
    
    /**
     * The AJAX action, to be used in a command for retrieving
     * entries from server. The entries will then be processed
     * by <code>entryHandler</code>.
     * @private
     * @type {string}
     */
    tsa.widgets.TablePanel.prototype.ajaxAction = '';
    
    /**
     * The method that will convert entry data into UI Elements to be
     * put into HTML Table Cells.
     * @private
     * @type {function(Object):Array.<(string|Element)>}
     */
    tsa.widgets.TablePanel.prototype.entryHandler = null;
    
    /** The columns' metadata.
     *  @private
     *  @type {Array.<{ id       : string
     *                , label    : string
     *                , width    : string
     *                , sortable : boolean
     *                , initSort : string }>} */
    
    tsa.widgets.TablePanel.prototype.columns = null;
    /** CSRF token required
     *  @private
     *  @type {boolean}>} */
    
    tsa.widgets.TablePanel.prototype.csrfRequired = true;
    /**
     * Current page.
     * @private
     * @type {number}
     */
    tsa.widgets.TablePanel.prototype.currentPage = 1;
    
    /**
     * Entries per page.
     * @private
     * @type {number}
     */
    tsa.widgets.TablePanel.prototype.pageSize = 20;
    
    /**
     * Total number of entries.
     * @private
     * @type {number}
     */
    tsa.widgets.TablePanel.prototype.numberOfEntries = 0;
    
    /**
     * Number of the column for which the sort will be requested.
     * First column is number 1, second column is number 2, etc.
     * Negative number will be used for reverse (descending) order.
     * Zero will be used if no order will be requested.
     * @private
     * @type {number}
     */
    tsa.widgets.TablePanel.prototype.sortByColumn = 0;
    
    /** The description section (HTML Element).
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TablePanel.prototype.descDiv = null;
    
    //-------------------------------------------------------------------------
    
    /** @return {Element} */
    
    tsa.widgets.TablePanel.prototype.getDomNode = function()
    {
      return this.domNode;
    };
    
    //-------------------------------------------------------------------------
    
    tsa.widgets.TablePanel.prototype.isAjaxDriven = function()
    {
      return ( ( this.ajaxAction.length > 0 ) && ( this.entryHandler !== null ) );
    };
    
    //-------------------------------------------------------------------------
    
    tsa.widgets.TablePanel.prototype.initUIElement = function()
    {
      /** @type {Array.<Object>}  */ var columns    = this.columns;
      /** @type {boolean}         */ var sortable   = false;
      /** @type {number}          */ var hideHeader = true;
      
      this.domNode         = document.createElement( 'div'     );
      this.table           = document.createElement( 'table'   );
      this.headerCaption   = document.createElement( 'caption' );
      this.descDiv         = document.createElement( 'div'     );
      this.tableHeader     = document.createElement( 'thead'   );
      this.tableBody       = document.createElement( 'tbody'   );
      this.emptyMsgElement = document.createElement( 'div'     );
      
      $( this.domNode ).append( this.table );
      $( this.domNode ).append( this.emptyMsgElement );
      
      $( this.domNode         ).addClass( 'ibm-container' );
      $( this.table           ).addClass( 'ibm-data-table' );
      $( this.domNode         ).css( { 'margin-bottom'   : '1.2em' } );
      $( this.table           ).attr( "border","0" );
      $( this.table           ).attr( "cellpadding","0" );
      $( this.table           ).attr( "cellspacing","0" );
      
      $( this.emptyMsgElement ).css( { 'padding'         : '0.8em'
                                     , 'font-family'     : 'sans-serif'
                                     , 'text-align'      : 'left'
                                     , 'font-size'       : '0.8em'
                                     , 'font-style'      : 'italic'
                                     , 'border-width'    : '1px'
                                     , 'border-style'    : 'solid'
                                     , 'border-color'    : '#ccc' } );
      $( this.descDiv         ).css( { 'margin-top'      : '0.3em'
                                     , 'padding-top'     : '0.3em'
                                     , 'border-top'      : '1px solid #cccccc'
                                     , 'display'         : 'none' } );
      
      if( TV.typeIsObject( typeof columns ) && TV.isInstanceOf( columns, Array ) && ( columns.length > 0 ) )
      {
        /** @type {Element} */ var colGrp = document.createElement( 'colgroup' );
        
        $( this.table ).append( colGrp );
        /** @type {Element} */ var trElement  = document.createElement( 'tr' );
        $.each( columns, $.proxy( function( columnIndex, column )
        {
          /** @type {Element} */ var col   = document.createElement( 'col' );
          /** @type {Element} */ var cell  = document.createElement( 'th' );
          /** @type {string}  */ var label = '';
          
          $( cell ).attr( 'scope', 'col' );
          $( cell ).css( { 'whitespace' : 'nowrap' } );
          
          $( colGrp ).append( col );
          $( trElement ).append( cell );
          
          // Setting column width:
          if( !TV.typeIsUndefined( typeof column.width ) )
          {
            if( TV.typeIsString( typeof column.width ) )
            {
              $( col ).css( { 'width' : column.width } );
            }
            else if( TV.typeIsNumber( typeof column.width ) && !isNaN( column.width ) && ( column.width > 0 ) )
            {
              $( col ).css( { 'width' : ( column.width + 'px' ) } );
            }
            else
            {
              logger.warn( 'Ignoring invalid column width: [' + column.width + '] (' + ( typeof column.width ) + ').' );
            }
          }
          
          // Sanitizing the label:
          if( !TV.typeIsUndefined( typeof column.label ) )
          {
            label = ( TV.typeIsString( typeof column.label ) ? column.label : column.label + '' );
            hideHeader = ( hideHeader && ( label.length === 0 ) );
          }
          else
          {
            label = tsa.NBSP;
          }
          
          // Is this column sortable? (only for AJAX)
          if( ( !TV.typeIsBoolean( typeof column.sortable ) || column.sortable ) && this.isAjaxDriven() )
          {
            /** @type {Element} */ var link      = document.createElement( 'a'    );
            /** @type {Element} */ var labelSpan = document.createElement( 'span' );
            /** @type {Element} */ var iconSpan  = document.createElement( 'span' );
            
            // Add the sortable class to the cell:
            $( cell ).addClass( 'ibm-sort' );
            
            if( TV.typeIsString( typeof column.initSort ) )
            {
              /** @type {boolean} */ var descending = ( column.initSort === 'desc' );
              
              $( cell ).addClass( descending ? 'ibm-sort-down' : 'ibm-sort-up' );
              this.sortByColumn = ( columnIndex + 1 ) * ( descending ? -1 : 1 );
            }
            
            // Add the sortable class to the table:
            if( !sortable )
            {
              $( this.table ).addClass( 'ibm-sortable-table' );
            }
            
            // Setting the sortable flag:
            sortable = true;
            
            // Setting the label:
            $( link ).attr( 'href', '#' );
            $( iconSpan ).addClass( 'ibm-icon' );
            $( iconSpan ).append( document.createTextNode( tsa.NBSP ) );
            $( labelSpan ).text( label );
            $( link ).append( labelSpan );
            $( link ).append( iconSpan );
            $( cell ).append( link );
            
            // Sort function:
            $( link ).on( 'click', $.proxy( function( event )
            {
              /** @type {boolean} */ var wasAscending = $( cell ).hasClass( 'ibm-sort-up' );
              
              this.setSortByColumn( columnIndex, !wasAscending );
              
              $( cell ).parent().children().removeClass( 'ibm-sort-up'   );
              $( cell ).parent().children().removeClass( 'ibm-sort-down' );
              
              $( cell ).addClass( wasAscending ? 'ibm-sort-down' : 'ibm-sort-up' );
              
              try
              {
                event.stopPropagation();
              }
              catch( error ) { /* Ignore. */ }
              return false;
            }
            , this ) );
          }
          else
          {
            // Just set the label:
            $( cell ).text( label );
          }
        }
        , this ) );
        $( this.tableHeader ).append( trElement ); 
      }
      else
      {
        logger.warn( 'Invalid column metadata.' );
      }
      
      $( this.table ).append( this.headerCaption );
      $( this.table ).append( this.tableHeader   );
      $( this.table ).append( this.tableBody     );
      
      // Hide header? (meaning no labels were provided)
      if( hideHeader )
      {
        $( this.tableHeader ).hide();
      }
      
      // These methods need to be called by first time...
      this.onWidthChange( null );
      this.onTitleChange( null );
      this.onDescriptionChange( null );
      this.onSummaryChange( null );
      this.onEmptyMsgChange( null );
      
      // Attaching event handlers...
      $( this ).on( 'widthChange'       , $.proxy( this.onWidthChange       , this ) );
      $( this ).on( 'titleChange'       , $.proxy( this.onTitleChange       , this ) );
      $( this ).on( 'descriptionChange' , $.proxy( this.onDescriptionChange , this ) );
      $( this ).on( 'summaryChange'     , $.proxy( this.onSummaryChange     , this ) );
      $( this ).on( 'emptyMsgChange'    , $.proxy( this.onEmptyMsgChange    , this ) );
    };
    
    //-------------------------------------------------------------------------
    
    /** @param {Event} event */
    
    tsa.widgets.TablePanel.prototype.onWidthChange = function( event )
    {
      $( this.domNode ).css( { 'width' : ( ( this.width > 0 ) ? this.width + 'px' : '' ) } );
      $( this.table   ).css( { 'width' : ( ( this.width > 0 ) ? this.width + 'px' : '' ) } );
    };
    
    //-------------------------------------------------------------------------
    
    /** @param {Event} event */
    
    tsa.widgets.TablePanel.prototype.onTitleChange = function( event )
    {
      /** @type {Element} */ var em = document.createElement( 'em' );
      
      $( em ).text( this.getTitle() + ( ( this.getSubtitle().length > 0 ) ? ' - ' + this.getSubtitle() : '' ) );
      
      $( this.headerCaption ).empty();
      $( this.headerCaption ).append( em );
      $( this.headerCaption ).append( this.descDiv );
    };
    
    //-------------------------------------------------------------------------
    
    /** @param {Event} event */
    
    tsa.widgets.TablePanel.prototype.onDescriptionChange = function( event )
    {
      if( this.descDiv !== null )
      {
        /** @type {boolean} */ var show = ( this.description.length > 0 );
        
        $( this.descDiv ).text( this.description );
        $( this.descDiv )[ show ? 'show' : 'hide' ]();
      }
    };
    
    //-------------------------------------------------------------------------
    
    /** @param {Event} event */
    
    tsa.widgets.TablePanel.prototype.onSummaryChange = function( event )
    {
      if( ( this.table !== null ) && ( this.summary.length !== 0 ) )
      {
        $( this.table ).attr( 'summary', this.summary );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /** @param {Event} event */
    
    tsa.widgets.TablePanel.prototype.onEmptyMsgChange = function( event )
    {
      if( this.emptyMsgElement !== null )
      {
        $( this.emptyMsgElement ).text( this.emptyMsg );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /** Set title.
     *  @param {string} title */
    
    tsa.widgets.TablePanel.prototype.setTitle = function( title )
    {
      this.title = '' + title;
      
      $( this ).trigger( { 'type' : 'titleChange' } );
    };
    
    //-------------------------------------------------------------------------
    
    /** @return {string} Title. */
    
    tsa.widgets.TablePanel.prototype.getTitle = function()
    {
      return this.title;
    };
    
    //-------------------------------------------------------------------------
    
    /** Set subtitle.
     *  @param {string} subtitle */
    
    tsa.widgets.TablePanel.prototype.setSubtitle = function( subtitle )
    {
      this.subtitle = '' + subtitle;
      
      $( this ).trigger( { 'type' : 'titleChange' } );
    };
    
    //-------------------------------------------------------------------------
    
    /** @return {string} Subtitle. */
    
    tsa.widgets.TablePanel.prototype.getSubtitle = function()
    {
      return this.subtitle;
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Set description
     * @param {string} description Description.
     */
    tsa.widgets.TablePanel.prototype.setDescription = function( description )
    {
      this.description = '' + $.trim( description );
      
      $( this ).trigger( { 'type' : 'descriptionChange' } );
    };
    
    //-----------------------------------------------------------------------
    
    /**
     * Get description
     * @return {string} Description.
     */
    tsa.widgets.TablePanel.prototype.getDescription = function()
    {
      return this.description;
    };
    
    //-------------------------------------------------------------------------
    
    /** Set summary (table attribute).
     *  @param {string} summary */
    
    tsa.widgets.TablePanel.prototype.setSummary = function( summary )
    {
      this.summary = '' + summary;
      
      $( this ).trigger( { 'type' : 'summaryChange' } );
    };
    
    //-------------------------------------------------------------------------
    
    /** @return {string} Summary. */
    
    tsa.widgets.TablePanel.prototype.getSummary = function()
    {
      return this.summary;
    };
    
    //-------------------------------------------------------------------------
    
    /** Set empty message.
     *  @param {string} emptyMsg */
    
    tsa.widgets.TablePanel.prototype.setEmptyMsg = function( emptyMsg )
    {
      this.emptyMsg = '' + emptyMsg;
      
      $( this ).trigger( { 'type' : 'emptyMsgChange' } );
    };
    
    //-------------------------------------------------------------------------
    
    /** @return {string} Empty message. */
    
    tsa.widgets.TablePanel.prototype.getEmptyMsg = function()
    {
      return this.emptyMsg;
    };
    
    //-------------------------------------------------------------------------
    
    /** Set width.
     *  @param {number} width */
    
    tsa.widgets.TablePanel.prototype.setWidth = function( width )
    {
      
      if( TV.typeIsNumber( typeof width ) && !isNaN( width ) )
      {
        if( width > 0 )
        {
          this.width = width;
        }
        else
        {
          this.width = 0; // 100%
        }
        
        $( this ).trigger( { 'type' : 'widthChange' } );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /** @return {number} Width. */
    
    tsa.widgets.TablePanel.prototype.getWidth = function()
    {
      return this.width;
    };
    
    //-----------------------------------------------------------------------
    
    /** Builds and adds a row.
     *  @param {Array.<(string|Element)>} elements The HTML Elements to be put at the row (each one in its own cell).
     *                                             Any non-element entry will be converted to text.
     *  @return {Element}                          Reference to the HTML TR (row) element that has been added. */
    
    tsa.widgets.TablePanel.prototype.appendRow = function( elements )
    {
      /** @type {Element} */ var row = document.createElement( 'tr' );
      
      if( TV.typeIsObject( typeof elements ) && ( elements !== null ) && TV.isInstanceOf( elements, Array ) )
      {
        $.each( elements, $.proxy( function( i, element )
        {
          /** @type {Element} */ var cell = document.createElement( 'td' );
          
          $( cell ).css( { 'word-wrap' : 'break-word' } );
          
          if( TV.typeIsObject( typeof element ) && ( element !== null ) && !TV.typeIsUndefined( element.nodeType ) )
          {
            $( cell ).append( element );
          }
          else if( TV.typeIsString( typeof element ) )
          {
            $( cell ).text( element );
          }
          else
          {
            $( cell ).text( element + '' );
          }
          
          $( row ).append( cell );
        }
        , this ) );
      }
      else
      {
        logger.warn( 'Cannot add row: invalid elements.' );
      }
      
      $( this.tableBody ).append( row );
      $( this.emptyMsgElement ).hide();
      
      return row;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * @return {number} Entries per page.
     */
    tsa.widgets.TablePanel.prototype.getPageSize = function()
    {
      return this.pageSize;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * @return {number} Current page number. First page is 1.
     */
    tsa.widgets.TablePanel.prototype.getCurrentPage = function()
    {
      return this.currentPage;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * @return {number} The total number of entries as reported by the server.
     */
    tsa.widgets.TablePanel.prototype.getNumberOfEntries = function()
    {
      return this.numberOfEntries;
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * @param {number}   columnIndex Index of the column for which the sort will be requested.
     * @param {boolean=} ascending   Optional. Use <code>false</code> for descending. Default: true.
     */
    tsa.widgets.TablePanel.prototype.setSortByColumn = function( columnIndex, ascending )
    {
      if( TV.typeIsNumber( typeof columnIndex ) && !isNaN( columnIndex ) && ( columnIndex >= 0 ) )
      {
        /** @type {boolean} */ var ascending    = !TV.typeIsBoolean( typeof ascending ) || ascending;
        /** @type {number}  */ var sortByColumn = ( columnIndex + 1 ) * ( ascending ? 1 : -1 );
        /** @type {boolean} */ var refresh      = ( this.sortByColumn !== sortByColumn );
        
        this.sortByColumn = sortByColumn;
        
        if( refresh )
        {
          this.goToPage( 1 );
        }
      }
      else
      {
        logger.error( 'Invalid column index (' + columnIndex + '; ' + ( typeof columnIndex ) + ').' );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * @param {number} pageSize The page size.
     */
    tsa.widgets.TablePanel.prototype.setPageSize = function( pageSize )
    {
      if( TV.typeIsNumber( typeof pageSize ) && !isNaN( pageSize ) && ( pageSize > 0 ) )
      {
        /** @type {boolean} */ var refresh = ( pageSize !== this.pageSize );
        
        this.pageSize = pageSize;
        
        if( refresh )
        {
          this.goToPage( 1 );
        }
      }
      else
      {
        logger.error( 'Invalid page size (' + pageSize + '; ' + ( typeof pageSize ) + ').' );
      }
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Go to page.
     * Requires valid <code>ajaxAction</code> and <code>entryHandler</code>.
     * @param {number} page Page number.
     */
    tsa.widgets.TablePanel.prototype.goToPage = function( page )
    {
      if( this.isAjaxDriven() )
      {
        if( page >= 1 )
        {
          /** @type {tsa.cmd.Command} */ var cmd = null;
          /** @type {Object}          */ var cmdParams = { page           : page
                                                         , entriesPerPage : this.getPageSize()};
          
          if( this.sortByColumn !== 0 )
          {
            /** @type {number}  */ var colIndex   = Math.abs( this.sortByColumn ) - 1;
            /** @type {boolean} */ var validIndex = ( colIndex < this.columns.length );
            
            if( validIndex )
            {
              /** @type {boolean} */ var colHasId = TV.typeIsString( typeof this.columns[ colIndex ].id );
              
              cmdParams.orderBy   = ( colHasId ? this.columns[ colIndex ].id : '' + colIndex );
              cmdParams.ascending = ( this.sortByColumn > 0 );
            }
            else
            {
              logger.error( 'Illegal sort index will be ignored (' + colIndex + ').' );
            }
          }
          
          cmd = new Command( { action              : this.ajaxAction
                             , parameters          : cmdParams
                             , clearSystemMessages : false } );
          
          $( this.tableBody ).empty();
          $( this.emptyMsgElement ).show();
          $( this.navSection ).remove();
          
          $( cmd ).on( 'executionFailed', $.proxy( function( event )
          {
            logger.error( 'The AJAX request couldn\'t be fulfilled.' );
          }
          , this ) );
          
          $( cmd ).on( 'executionSucceeded', $.proxy( function( event )
          {
            /** @type {Array.<Array.<(string|Element)>>} */ var rowsContents = null;
            /** @type {tsa.cmd.AjaxResponse}             */ var response     = event.response;
            /** @type {Object}                           */ var payload      = ( ( response !== null ) ? response.payload : null );
            
            this.currentPage = page;
            
            if( TV.typeIsObject( typeof payload         ) && ( payload         !== null ) &&
                TV.typeIsObject( typeof payload.entries ) && ( payload.entries !== null ) &&
                TV.isInstanceOf( payload.entries, Array ) )
            {
              rowsContents = [];
              
              if( TV.typeIsNumber( typeof payload.total ) && !isNaN( payload.total ) &&
                  ( payload.total >= payload.entries.length ) )
              {
                this.numberOfEntries = ( TV.typeIsNumber( typeof payload.total )
                                         ? payload.total
                                         : payload.entries.length );
              }
              else
              {
                logger.warn( 'Invalid number of total entries reported by server.' );
                this.numberOfEntries = payload.entries.length;
              }
              
              try
              {
                $.each( payload.entries, $.proxy( function( entryIndex, entry )
                {
                  rowsContents.push( this.entryHandler( entry ) );
                }
                , this ) );
              }
              catch( error )
              {
                logger.error( error );
                logger.error( 'The provided entry handler failed at processing the entry at index ' + rowsContents.length + '.' );
              }
            }
            else
            {
              logger.error( 'The response received from server does not contain an "entries" property.' );
            }
            
            if( ( rowsContents !== null ) && ( rowsContents.length > 0 ) )
            {
              try
              {
                $.each( rowsContents, $.proxy( function( rowIndex, contents )
                {
                  this.appendRow( contents );
                }
                , this ) );
                
                // Add pagination controls:
                this.navSection = this.buildNavigationSection();
                $( this.domNode ).append( this.navSection );
              }
              catch( error )
              {
                logger.error( error );
                logger.error( 'The entry handler returned content that could not be added.' );
              }
            }
          }
          , this ) );
          
          if(!this.csrfRequired){
            cmd.secureToken = false;
          }

            cmd.execute();

        }
        else
        {
          logger.error( 'Invalid page: ' + page + '.' );
        }
      }
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Build the navigation section.
     * @return {Element}
     */
    tsa.widgets.TablePanel.prototype.buildNavigationSection = function()
    {
      /** @type {Element} */ var main          = document.createElement( 'div'  );
      /** @type {Element} */ var paragraph     = document.createElement( 'p'    );
      /** @type {Element} */ var pageNumCtrl   = document.createElement( 'span' );
      /** @type {Element} */ var pageSizeCtrl  = document.createElement( 'span' );
      /** @type {Element} */ var pageSizeLinks = document.createElement( 'span' );
      
      /** @type {Array.<number>} */ var pageSizes = [ 20, 50, 100 ];
      /** @type {number}         */ var from      = 1 + ( this.getCurrentPage() - 1 ) * this.getPageSize();
      /** @type {number}         */ var to        = from + this.getPageSize() - 1;
      /** @type {number}         */ var total     = this.getNumberOfEntries();
      
      if( to > total ) to = total;
      
      $( paragraph ).attr( 'role', 'navigation' );
      $( paragraph ).attr( 'aria-label', 'Navigation paragraph' );
      
      $( main          ).addClass( 'ibm-container-body'         );
      $( paragraph     ).addClass( 'ibm-table-navigation'       );
      $( pageNumCtrl   ).addClass( 'ibm-primary-navigation'     );
      $( pageSizeCtrl  ).addClass( 'ibm-secondary-navigation'   );
      $( pageSizeLinks ).addClass( 'ibm-table-navigation-links' );
      
      $( paragraph ).append( pageNumCtrl  );
      $( paragraph ).append( pageSizeCtrl );
      $( main      ).append( paragraph    );
      
      // The FROM - TO descriptor:
      $( pageNumCtrl ).html( '<strong>' + from + ' - '+ to + '</strong> ' + tsa.i18n.labelPaginationOf
                           + ' <strong>' + total + '</strong> ' + tsa.i18n.labelPaginationEntries );
      
      // The NEXT and PREVIOUS links:
      if( ( from > 1 ) || ( to < total ) )
      {
        /** @type {Element} */ var pageNavLinks = document.createElement( 'span' );
        
        $( pageNavLinks ).addClass( 'ibm-table-navigation-links' );
        $( pageNumCtrl ).append( pageNavLinks );
        
        if( from > 1 )
        {
          /** @type {Element} */ var navPrevLink = document.createElement( 'a' );
          
          $( navPrevLink ).addClass( 'ibm-back-em-link' );
          $( navPrevLink ).text( tsa.i18n.labelPaginationPrevious );
          $( navPrevLink ).attr( 'href', '#' );
          $( pageNavLinks ).append( document.createTextNode( ' | ' ) );
          $( pageNavLinks ).append( navPrevLink );
          
          $( navPrevLink ).on( 'click', $.proxy( function( event )
          {
            this.goToPage( this.getCurrentPage() - 1 );
            try
            {
              event.stopPropagation();
            } catch( error ) { /* Ignore. */ }
            return false;
          }
          , this ) );
        }
        
        if( to < total )
        {
          /** @type {Element} */ var navNextLink = document.createElement( 'a' );
          
          $( navNextLink ).addClass( 'ibm-forward-em-link' );
          $( navNextLink ).text( tsa.i18n.labelPaginationNext );
          $( navNextLink ).attr( 'href', '#' );
          $( pageNavLinks ).append( document.createTextNode( ' | ' ) );
          $( pageNavLinks ).append( navNextLink );
          
          $( navNextLink ).on( 'click', $.proxy( function( event )
          {
            this.goToPage( this.getCurrentPage() + 1 );
            try
            {
              event.stopPropagation();
            } catch( error ) { /* Ignore. */ }
            return false;
          }
          , this ) );
        }
      }
      
      // The page size links:
      $( pageSizeCtrl ).text( tsa.i18n.labelPaginationEntriesPerPage + ': ' );
      $( pageSizeCtrl ).append( pageSizeLinks );
      
      $.each( pageSizes, $.proxy( function( index, size )
      {
        /** @type {Element} */ var pageSizeLink = document.createElement( 'a' );
        
        if( index > 0 ) $( pageSizeLinks ).append( document.createTextNode( ' | ' ) );
        
        $( pageSizeLink ).text( size );
        $( pageSizeLink ).attr( 'href', '#' );
        $( pageSizeLinks ).append( pageSizeLink );
        
        $( pageSizeLink ).on( 'click', $.proxy( function( event )
        {
          this.setPageSize( size );
          try
          {
            event.stopPropagation();
          } catch( error ) { /* Ignore. */ }
          return false;
        }
        , this ) );
      }
      , this ) );
      
      return main;
    };
    
    //-------------------------------------------------------------------------
    
    return tsa.widgets.TablePanel;
  }
);