/*
 * IBM Technical Support Appliance Version 2 Release 3 Modification 0
 *
 * (C) Copyright IBM Corp. 2018
 */

//-----------------------------------------------------------------------------

define(
  [ 'jquery'
  , 'tsa/TypeVerificator'
  , 'tsa/cmd/ShowCommand'
  ],
  
  function( $
          , TV
          , ShowCommand )
  {
    //-------------------------------------------------------------------------
    
    var logger = tsa.getLogger( 'TabsPanel' );
    
    //-------------------------------------------------------------------------
    
    /** @constructor
     *  Parameters:
     *    tabs  Array of objects, each one containing information about a particular tab:
     *          title (string)      : The title of the tab.
     *          action (string)     : Url to redirect.
     *          body  (HTMLElement) : The HTML Element to be displayed when this tab is selected.
     *                                A paragraph (<p>) or a div (<div>) are recommended.
     *  
     *  @param { tabs : Array.<{ title : string
     *                         , body  : HTMLElement}> } parameters The parameters object. */
    
    tsa.widgets.TabsPanel = function( parameters )
    {
      if( !TV.typeIsObject( typeof parameters ) || ( parameters === null ) ) parameters = {};
      
      if( TV.typeIsObject( typeof parameters.tabs ) && TV.isInstanceOf( parameters.tabs, Array ) )
      {
        this.tabData = parameters.tabs;
      }
      else
      {
        logger.error( 'Invalid tab data provided (an array is expected).' );
      }
      
      this.initUIElement();
    };
    
    //-------------------------------------------------------------------------
    
    /** Main HTML Element (DIV).
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TabsPanel.prototype.domNode = null;
    
    /** The tab data.
     *  @private
     *  @type {Array.<{title:string, body:HTMLElement}>}
     */
    tsa.widgets.TabsPanel.prototype.tabData = null;
    
    /** The tab list.
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TabsPanel.prototype.tabList = null;
    
    /** The body.
     *  @private
     *  @type {Element}
     */
    tsa.widgets.TabsPanel.prototype.body = null;
    
    //-------------------------------------------------------------------------
    /**
     * Append element to body.
     * @param {element} tabIndex Tab index.
     */
    tsa.widgets.TabsPanel.prototype.appendElement = function( element )
    {
      if(TV.typeIsObject( typeof element )){
        $(this.body).append(element);
      }
      else
      {
        console.log("Not able to append element to TabsPanel body");
      }
      	
    };
    
    //-------------------------------------------------------------------------
    /** @return {Element} */
    
    tsa.widgets.TabsPanel.prototype.getDomNode = function()
    {
      return this.domNode;
    };
    
    //-------------------------------------------------------------------------
    
    tsa.widgets.TabsPanel.prototype.initUIElement = function()
    {
      /** @type {Element} */ var tabSection = document.createElement( 'div' );
      
      this.domNode = document.createElement( 'div' );
      this.tabList = document.createElement( 'ul' );
      this.body    = document.createElement( 'div' );
      
      $( this.domNode ).addClass( 'ibm-container' );
      $( this.domNode ).addClass( 'ibm-graphic-tabs' );
      $( tabSection   ).addClass( 'ibm-tab-section' );
      $( this.tabList ).addClass( 'ibm-tabs' );
      $( this.body    ).addClass( 'ibm-container-body' );
      $( this.body    ).addClass( 'ibm-inner-data-table-last' );
      
      $( tabSection   ).append( this.tabList );
      $( this.domNode ).append( tabSection );
      $( this.domNode ).append( this.body );
      
      if( this.tabData !== null )
      {
        $.each( this.tabData, $.proxy( function( tabIndex, tabDataEntry )
        {
          /** @type {boolean} */ var isFirst     = ( tabIndex === 0 );
          /** @type {boolean} */ var isLast      = ( tabIndex === ( this.tabData.length - 1 ) );
          /** @type {Element} */ var tabListItem = document.createElement( 'li' );
          /** @type {Element} */ var tabLink     = document.createElement( 'a' );
          /** @type {Element} */ var tabBody     = document.createElement( 'div' );
          /** @type {string}  */ var title       = ( TV.typeIsString( typeof tabDataEntry.title ) ? tabDataEntry.title : '-' );
          /** @type {Element} */ var subbody     = ( ( TV.typeIsObject( typeof tabDataEntry.body  ) && ( tabDataEntry.body !== null ) ) ? tabDataEntry.body : ' ' );
          
          $( tabListItem  ).append( tabLink );
          $( this.tabList ).append( tabListItem );
          
          if( isFirst || isLast )
          {
            $( tabListItem ).addClass( 'ibm-' + ( isFirst ? 'first' : 'last' ) + '-tab' );
          }
          
          try
          {
            
            $( this.body ).append( subbody );
            
            
          }
          catch( error )
          {
            logger.error( 'Body for tab at index ' + ( tabIndex ) + ' could not be added.' );
          }
          
          $( tabLink ).text( tabDataEntry.title + '' );
          
          if(  TV.typeIsString( typeof tabDataEntry.action )){
            $( tabLink ).on( 'click',function(){
            (new ShowCommand( {
              action : tabDataEntry.action
                , clearSystemMessages : false} 
             )).execute();
            });
          }
          else
          {
          $( tabLink ).on( 'click', $.proxy( function( event )
          {
            this.goToTab( tabIndex );
            
            $( this ).trigger( { type     : 'tabChanged'
                               , tabIndex : tabIndex } );
            
            try
            {
              event.stopPropagation();
            }
            catch( error ) { /* Ignore. */ }
            
            return false;
          }
          , this ) );
          }
          
          tabLink.href = '#';
        }
        , this ) );
      }
      
      this.goToTab( 0 );
    };
    
    //-------------------------------------------------------------------------
    
    /**
     * Go to tab.
     * @param {number} tabIndex Tab index.
     */
    tsa.widgets.TabsPanel.prototype.goToTab = function( tabIndex )
    {
      $( this.tabList ).children().each( function( childIndex, child )
      {
        $( child )[ ( childIndex === tabIndex ) ? 'addClass' : 'removeClass' ]( 'ibm-highlight-tab' );
      } );
      
      $( this.body ).children().each( function( childIndex, child )
      {
        $( child )[ ( childIndex === tabIndex ) ? 'show' : 'hide' ]();
      } );
    };
    
    //-------------------------------------------------------------------------
    
    return tsa.widgets.TabsPanel;
  }
);