#! /bin/bash

shopt -s extglob extquote
shopt -s nocasematch

# Set script name and full path.
TSACONFIG_SH=$(basename "$0")
TSACONFIG_FULL_SH="$0"
export TSACONFIG_SH TSACONFIG_FULL_SH

oneline_usage="$TSACONFIG_SH"

usage()
{
cat <<-EndUsage
        Usage: $oneline_usage
        Try '$TSACONFIG_SH -h' for more information.
EndUsage
    exit 1
}

help()
{
cat <<-EndOptionsHelp
          Usage: $oneline_usage

          This script is used for initial network configuration of this appliance

          Options:
            -h
                Display a help message
                
            -i
               Show current network configuration

EndOptionsHelp
}

info()
{
    ip=$(ifconfig eth0 | grep 'inet addr:' | cut -d: -f2 | awk '{print $1}')
    iptype=$(cat /etc/sysconfig/network-scripts/ifcfg-eth0 | grep 'BOOTPROTO' | cut -d= -f2)
    hostname=$(hostname)
    
cat <<-EndOptionsInfo
        IP TYPE:${iptype}
        IP Address:${ip}
        
        TSA console url https://${ip}/appliance
        
        eth0 configuration:
        `ifconfig eth0`

EndOptionsInfo
}

dieWithHelp()
{
    case "$1" in
        help)       help;;
    esac
    shift

    die "$@"
}
die()
{
    echo "$*"
    exit 1
}

if [ $EUID != 0 ]; then
    die "This script should be run either as root or by sudo "
fi

action="interactive"
# == PROCESS OPTIONS ==
while getopts "hi" Option
do
  case $Option in
    h )
        OPTIND=2
        action="help"
        ;;
    i )
        OPTIND=2
        action="info"
        ;;    
  esac
done
shift $(($OPTIND - 1))


getinput() {
    echo -n $1
    read input
    eval "$2='${input}'"
}

getDNS() {

    domain=""
    DNS=""
    
    getinput "Enter network domain of system for DNS usage(optional):" domain
    DNS="DOMAIN:${domain}"
        
    dns1=''
    getinput "Enter DNS 1(optional): " dns1
    
    if [ -n "${dns1}" ]; then
                    
        DNS="${DNS}\nDNS1:${dns1}" 
                    
        dns2=''
        getinput "Enter DNS 2(optional): " dns2
    
        if [ -n "${dns2}" ]; then
    
            DNS="${DNS}\nDNS2:${dns2}" 
                        
            dns3=''
            getinput "Enter DNS 3(optional): " dns3
    
            if [ -n "${dns3}" ]; then
                DNS="${DNS}\nDNS3:${dns3}" 
            fi
        fi
    fi

    eval "$1='${DNS}'"
}

savenetworkconfig() {
    echo -e "${1}" > /tmp/ipconfig.properties

    if [ -f /tmp/ipconfig.properties ]; then
        java -cp /opt/ibm/servicestation/config/. Startup 1>> /tmp/network.log 2>> /tmp/network.err
        /etc/init.d/network restart 1>> /tmp/network.log 2>> /tmp/network.err
        
        dnstype=$(cat /tmp/ipconfig.properties | grep 'IPTYPE' | cut -d: -f2)
        
        if [[ "${dnstype}" = "static" ]]; then
            echo -e "NAMESERVICETYPE=STATIC\n" > /etc/ibmtsa.conf
        else
            echo -e "NAMESERVICETYPE=DHCP\n" > /etc/ibmtsa.conf
        fi
        
    fi
}

updateUser() {
    
    ip=$(ifconfig eth0 | grep 'inet addr:' | cut -d: -f2 | awk '{print $1}')
    if [ "${ip}" = "" ]; then
        ip=$(ifconfig eth0 | grep 'inet ' | tr -s ' ' | cut -d ' ' -f 3 | awk '{print $1}')
    fi
    hostname=$(hostname)
cat <<-EndOptionsUserUpdate

Network configuration is successful.
Access TSA console at https://${ip}/appliance

eth0 configuration:
`ifconfig eth0`

Hostname is ${hostname}

Rebooting system in 15 seconds for new network configuration to take effect
EndOptionsUserUpdate

sleep 15
reboot
}

showMenu() {
    clear
    printf "\n----- TSA Config Menu -----"
    printf "\n1) Setup network configuration"
    printf "\n2) Change tsausr password"
    printf "\n3) Set Appliance certificate to default"
    printf "\n4) Exit"
    printf "\n\nChoose an option: "
    read option
    if [ "${option}" == "1" ]; then
        getUserInput
    elif [ "${option}" == "2" ]; then
        clear
        passwd tsausr
        printf "\nReturning to menu in 5 seconds..."
        sleep 5
        showMenu
    elif [ "${option}" == "3" ]; then
    	setDefaultCertificate
    elif [ "${option}" == "4" ]; then
        exit
    else
        showMenu
    fi
}

setDefaultCertificate() {
	clear
	printf "\nConfirm setting appliance certificate to default certificate\n[y|n]:"
    read confirm
    

    if [ "${confirm}" = "y" ]; then
        printf "Setting appliance certificate to default Certificate\n"	
		
		def_cert_path="\t      location=\"/opt/ibm/servicestation/keystore/ibmtsa0.jks\""
		server_file="/opt/ibm/servicestation/liberty/usr/servers/tsa/server.xml"
		def_pass="\t      password=\"{xor}Nj0yLz4sLChvLTs=\""
		loc_path="location=.*"
		pass="password=.*"
		tmp_server_file="/tmp/tmpCert.xml"

		awk -v loc_path="$loc_path" -v path="$def_cert_path" ' NR==1,/location=.*/{sub(loc_path, path,$1)}1' $server_file > $tmp_server_file
    
    if [ $? -eq 0 ]; then
				awk -v pass="$pass" -v def_pass="$def_pass" ' NR==1,/password=.*/{sub(pass, def_pass,$1)}1'  $tmp_server_file > $server_file
				printf "Rebooting system in 5 seconds to set up appliance default certificate\n"
				sleep 5
				reboot		
	fi
	else
		showMenu
	fi
}

getUserInput() {
    clear
    iptype=''
    getinput "Enter IPTYPE={static|dhcp}: " iptype 
    [ -z  "${iptype}" ] && iptype="dhcp"

    hostname=''
    getinput "Enter Hostname(default=ibmtsa): " hostname
    [ -z  "${hostname}" ] && hostname="ibmtsa"

    if [[ "${iptype}" = "static" ]]; then
        ipaddr=''
        getinput "Enter IP Address: " ipaddr

        netmask=''
        getinput "Enter Netmask: " netmask

        gateway=''
        getinput "Enter Gateway Address: " gateway

        DNS=""
        getDNS DNS 

        ipconfig="IPTYPE:${iptype}\nHOSTNAME:${hostname}\nIPADDR:${ipaddr}\nNETMASK:${netmask}\nGATEWAY:${gateway}\n${DNS}"
    else
        DNS=""
        getDNS DNS
        ipconfig="IPTYPE:${iptype}\nHOSTNAME:${hostname}\n${DNS}"
    fi

    printf "\nConfirm network configuration\n${ipconfig}\n[y|n]:"
    read confirm

    if [ "${confirm}" = "y" ]; then
        printf "Saving network configuration\n"
        if [[ "${iptype}" == "dhcp" ]]; then
            #Reset GATEWAY previously set for static ip
            ipconfig="${ipconfig}\nGATEWAY:\n"
        fi
        savenetworkconfig "${ipconfig}"
        updateUser
    else
        showMenu
    fi
}

## Run required action.
case $action in
"interactive" )
        showMenu
    ;;

"help" )
        help
    ;;
    
"info" )
        info
    ;;

* )
    usage;;
esac
